@@ -62,15 +62,15 @@ import org.junit.Test;
 
 import org.apache.celeborn.common.CelebornConf;
 import org.apache.celeborn.plugin.flink.buffer.BufferPacker;
-import org.apache.celeborn.plugin.flink.buffer.SortBuffer;
+import org.apache.celeborn.plugin.flink.buffer.DataBuffer;
 import org.apache.celeborn.plugin.flink.readclient.FlinkShuffleClientImpl;
 import org.apache.celeborn.plugin.flink.utils.BufferUtils;
 
 public class RemoteShuffleResultPartitionSuiteJ {
   private final int networkBufferSize = 32 * 1024;
-  private BufferCompressor bufferCompressor = new BufferCompressor(networkBufferSize, "lz4");
-  private RemoteShuffleOutputGate remoteShuffleOutputGate = mock(RemoteShuffleOutputGate.class);
-  private final String compressCodec = "LZ4";
+  private final BufferCompressor bufferCompressor = new BufferCompressor(networkBufferSize, "lz4");
+  private final RemoteShuffleOutputGate remoteShuffleOutputGate =
+      mock(RemoteShuffleOutputGate.class);
   private final CelebornConf conf = new CelebornConf();
   BufferDecompressor bufferDecompressor = new BufferDecompressor(networkBufferSize, "LZ4");
 
@@ -129,10 +129,10 @@ public class RemoteShuffleResultPartitionSuiteJ {
     doNothing().when(remoteShuffleOutputGate).regionStart(anyBoolean());
     doNothing().when(remoteShuffleOutputGate).regionFinish();
     when(remoteShuffleOutputGate.getBufferPool()).thenReturn(bufferPool.get(1).get());
-    SortBuffer sortBuffer = remoteShuffleResultPartition.getDelegation().getUnicastSortBuffer();
+    DataBuffer dataBuffer = remoteShuffleResultPartition.getDelegation().getUnicastDataBuffer();
     ByteBuffer byteBuffer = ByteBuffer.wrap(new byte[] {1, 2, 3});
-    sortBuffer.append(byteBuffer, 0, Buffer.DataType.DATA_BUFFER);
-    remoteShuffleResultPartition.getDelegation().flushSortBuffer(sortBuffer, true);
+    dataBuffer.append(byteBuffer, 0, Buffer.DataType.DATA_BUFFER);
+    remoteShuffleResultPartition.getDelegation().flushDataBuffer(dataBuffer, true);
   }
 
   private List<SupplierWithException<BufferPool, IOException>> createBufferPoolFactory() {
@@ -551,7 +551,7 @@ public class RemoteShuffleResultPartitionSuiteJ {
     }
   }
 
-  private RemoteShuffleDescriptor getShuffleDescriptor() throws Exception {
+  private RemoteShuffleDescriptor getShuffleDescriptor() {
     Random random = new Random();
     byte[] bytes = new byte[16];
     random.nextBytes(bytes);
