@@ -68,7 +68,6 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
   private final ShuffleWriteMetricsReporter writeMetrics;
   private final int shuffleId;
   private final int mapId;
-  private final int encodedAttemptId;
   private final TaskContext taskContext;
   private final ShuffleClient shuffleClient;
   private final int numMappers;
@@ -84,7 +83,7 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
   private int[] sendOffsets;
 
   private final LongAdder[] mapStatusLengths;
-  private final long[] tmpRecords;
+  protected long tmpRecordsWritten = 0;
 
   private final SendBufferPool sendBufferPool;
 
@@ -113,7 +112,6 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
     this.mapId = taskContext.partitionId();
     this.dep = handle.dependency();
     this.shuffleId = shuffleId;
-    this.encodedAttemptId = SparkCommonUtils.getEncodedAttemptNumber(taskContext);
     SerializerInstance serializer = dep.serializer().newInstance();
     this.partitioner = dep.partitioner();
     this.writeMetrics = metrics;
@@ -130,7 +128,6 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
     for (int i = 0; i < numPartitions; i++) {
       mapStatusLengths[i] = new LongAdder();
     }
-    tmpRecords = new long[numPartitions];
 
     PUSH_BUFFER_INIT_SIZE = conf.clientPushBufferInitialSize();
     PUSH_BUFFER_MAX_SIZE = conf.clientPushBufferMaxSize();
@@ -145,7 +142,7 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
           new DataPusher(
               shuffleId,
               mapId,
-              encodedAttemptId,
+              taskContext.attemptNumber(),
               taskContext.taskAttemptId(),
               numMappers,
               numPartitions,
@@ -161,7 +158,6 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
 
   @Override
   public void write(scala.collection.Iterator<Product2<K, V>> records) throws IOException {
-    boolean needCleanupPusher = true;
     try {
       if (canUseFastWrite()) {
         fastWrite0(records);
@@ -175,13 +171,8 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
         write0(records);
       }
       close();
-      needCleanupPusher = false;
     } catch (InterruptedException e) {
       TaskInterruptedHelper.throwTaskKillException();
-    } finally {
-      if (needCleanupPusher) {
-        cleanupPusher();
-      }
     }
   }
 
@@ -235,14 +226,10 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
             rowSize);
         sendOffsets[partitionId] = offset + serializedRecordSize;
       }
-      incRecordsWritten(partitionId);
+      tmpRecordsWritten++;
     }
   }
 
-  protected void incRecordsWritten(int partitionId) {
-    tmpRecords[partitionId] += 1;
-  }
-
   private void write0(scala.collection.Iterator iterator) throws IOException, InterruptedException {
     final scala.collection.Iterator<Product2<K, ?>> records = iterator;
 
@@ -266,7 +253,7 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
         System.arraycopy(serBuffer.getBuf(), 0, buffer, offset, serializedRecordSize);
         sendOffsets[partitionId] = offset + serializedRecordSize;
       }
-      incRecordsWritten(partitionId);
+      tmpRecordsWritten++;
     }
   }
 
@@ -286,7 +273,7 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
         shuffleClient.pushData(
             shuffleId,
             mapId,
-            encodedAttemptId,
+            taskContext.attemptNumber(),
             partitionId,
             buffer,
             0,
@@ -313,7 +300,7 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
 
     if ((buffer.length - offset) < serializedRecordSize) {
       flushSendBuffer(partitionId, buffer, offset);
-      updateMapStatus();
+      updateRecordsWrittenMetrics();
       offset = 0;
     }
     return offset;
@@ -350,7 +337,7 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
         shuffleClient.mergeData(
             shuffleId,
             mapId,
-            encodedAttemptId,
+            taskContext.attemptNumber(),
             partitionId,
             buffer,
             offset,
@@ -361,29 +348,19 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
     writeMetrics.incBytesWritten(bytesWritten);
   }
 
-  private void cleanupPusher() throws IOException {
-    try {
-      dataPusher.waitOnTermination();
-      sendBufferPool.returnPushTaskQueue(dataPusher.getAndResetIdleQueue());
-    } catch (InterruptedException e) {
-      TaskInterruptedHelper.throwTaskKillException();
-    }
-  }
-
   private void close() throws IOException, InterruptedException {
     // here we wait for all the in-flight batches to return which sent by dataPusher thread
     long pushMergedDataTime = System.nanoTime();
     dataPusher.waitOnTermination();
-    sendBufferPool.returnPushTaskQueue(dataPusher.getAndResetIdleQueue());
-    shuffleClient.prepareForMergeData(shuffleId, mapId, encodedAttemptId);
+    sendBufferPool.returnPushTaskQueue(dataPusher.getIdleQueue());
+    shuffleClient.prepareForMergeData(shuffleId, mapId, taskContext.attemptNumber());
     closeWrite();
-    shuffleClient.pushMergedData(shuffleId, mapId, encodedAttemptId);
+    shuffleClient.pushMergedData(shuffleId, mapId, taskContext.attemptNumber());
     writeMetrics.incWriteTime(System.nanoTime() - pushMergedDataTime);
-
-    updateMapStatus();
+    updateRecordsWrittenMetrics();
 
     long waitStartTime = System.nanoTime();
-    shuffleClient.mapperEnd(shuffleId, mapId, encodedAttemptId, numMappers);
+    shuffleClient.mapperEnd(shuffleId, mapId, taskContext.attemptNumber(), numMappers);
     writeMetrics.incWriteTime(System.nanoTime() - waitStartTime);
 
     BlockManagerId bmId = SparkEnv.get().blockManager().shuffleServerId();
@@ -392,13 +369,9 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
             bmId, SparkUtils.unwrap(mapStatusLengths), taskContext.taskAttemptId());
   }
 
-  private void updateMapStatus() {
-    long recordsWritten = 0;
-    for (int i = 0; i < partitioner.numPartitions(); i++) {
-      recordsWritten += tmpRecords[i];
-      tmpRecords[i] = 0;
-    }
-    writeMetrics.incRecordsWritten(recordsWritten);
+  private void updateRecordsWrittenMetrics() {
+    writeMetrics.incRecordsWritten(tmpRecordsWritten);
+    tmpRecordsWritten = 0;
   }
 
   @Override
@@ -420,7 +393,7 @@ public class HashBasedShuffleWriter<K, V, C> extends ShuffleWriter<K, V> {
         }
       }
     } finally {
-      shuffleClient.cleanup(shuffleId, mapId, encodedAttemptId);
+      shuffleClient.cleanup(shuffleId, mapId, taskContext.attemptNumber());
     }
   }
 
