@@ -18,9 +18,11 @@
 package org.apache.spark.shuffle.celeborn;
 
 import java.io.IOException;
+import java.util.Objects;
 import java.util.concurrent.ConcurrentHashMap;
 
 import org.apache.spark.*;
+import org.apache.spark.internal.config.package$;
 import org.apache.spark.launcher.SparkLauncher;
 import org.apache.spark.rdd.DeterministicLevel;
 import org.apache.spark.shuffle.*;
@@ -94,7 +96,17 @@ public class SparkShuffleManager implements ShuffleManager {
               + "use Celeborn as Remote Shuffle Service to avoid performance degradation.",
           SQLConf.LOCAL_SHUFFLE_READER_ENABLED().key());
     }
-    SparkCommonUtils.validateAttemptConfig(conf);
+    if ((Boolean) conf.get(package$.MODULE$.DYN_ALLOCATION_SHUFFLE_TRACKING_ENABLED())) {
+      String key = package$.MODULE$.DYN_ALLOCATION_SHUFFLE_TRACKING_ENABLED().key();
+      Boolean defaultValue =
+          (Boolean) package$.MODULE$.DYN_ALLOCATION_SHUFFLE_TRACKING_ENABLED().defaultValue().get();
+      logger.warn(
+          "Detected {} (default is {}) is enabled, "
+              + "it's highly recommended to disable it when use Celeborn as Remote Shuffle Service "
+              + "to avoid performance degradation.",
+          key,
+          defaultValue);
+    }
     this.conf = conf;
     this.isDriver = isDriver;
     this.celebornConf = SparkUtils.fromSparkConf(conf);
@@ -179,7 +191,7 @@ public class SparkShuffleManager implements ShuffleManager {
 
   @Override
   public boolean unregisterShuffle(int appShuffleId) {
-    if (sortShuffleIds.contains(appShuffleId)) {
+    if (sortShuffleIds.remove(appShuffleId)) {
       return sortShuffleManager().unregisterShuffle(appShuffleId);
     }
     // For Spark driver side trigger unregister shuffle.
@@ -234,7 +246,29 @@ public class SparkShuffleManager implements ShuffleManager {
         int shuffleId = SparkUtils.celebornShuffleId(shuffleClient, h, context, true);
         shuffleIdTracker.track(h.shuffleId(), shuffleId);
 
-        if (ShuffleMode.SORT.equals(celebornConf.shuffleWriterMode())) {
+        ShuffleMode shuffleMode = celebornConf.shuffleWriterMode();
+        if (celebornConf.dynamicWriteModeEnabled()) {
+          int partitionCount = h.dependency().partitioner().numPartitions();
+          if (partitionCount > celebornConf.dynamicWriteModePartitionNumThreshold()) {
+            logger.info(
+                "Shuffle {} write mode is changed to SORT because "
+                    + "partition count {} is greater than threshold {}",
+                shuffleId,
+                partitionCount,
+                celebornConf.dynamicWriteModePartitionNumThreshold());
+            shuffleMode = ShuffleMode.SORT;
+          } else {
+            logger.info(
+                "Shuffle {} write mode is changed to HASH because "
+                    + "partition count {} is less than threshold {}",
+                shuffleId,
+                partitionCount,
+                celebornConf.dynamicWriteModePartitionNumThreshold());
+            shuffleMode = ShuffleMode.HASH;
+          }
+        }
+
+        if (ShuffleMode.SORT.equals(shuffleMode)) {
           return new SortBasedShuffleWriter<>(
               shuffleId,
               h.dependency(),
@@ -244,10 +278,11 @@ public class SparkShuffleManager implements ShuffleManager {
               shuffleClient,
               metrics,
               SendBufferPool.get(cores, sendBufferPoolCheckInterval, sendBufferPoolExpireTimeout));
-        } else if (ShuffleMode.HASH.equals(celebornConf.shuffleWriterMode())) {
+        } else if (ShuffleMode.HASH.equals(shuffleMode)) {
           SendBufferPool pool =
               SendBufferPool.get(cores, sendBufferPoolCheckInterval, sendBufferPoolExpireTimeout);
           if (COLUMNAR_SHUFFLE_CLASSES_PRESENT && celebornConf.columnarShuffleEnabled()) {
+            logger.info("Creating columnar hash shuffle writer for shuffle {}", shuffleId);
             return SparkUtils.createColumnarHashBasedShuffleWriter(
                 shuffleId, h, context, celebornConf, shuffleClient, metrics, pool);
           } else {
@@ -259,7 +294,7 @@ public class SparkShuffleManager implements ShuffleManager {
               "Unrecognized shuffle write mode!" + celebornConf.shuffleWriterMode());
         }
       } else {
-        sortShuffleIds.add(handle.shuffleId());
+        checkUserClassPathFirst(handle);
         return sortShuffleManager().getWriter(handle, mapId, context, metrics);
       }
     } catch (IOException e) {
@@ -281,6 +316,7 @@ public class SparkShuffleManager implements ShuffleManager {
       return getCelebornShuffleReader(
           handle, startPartition, endPartition, startMapIndex, endMapIndex, context, metrics);
     }
+    checkUserClassPathFirst(handle);
     return SparkUtils.getReader(
         sortShuffleManager(),
         handle,
@@ -304,6 +340,7 @@ public class SparkShuffleManager implements ShuffleManager {
       return getCelebornShuffleReader(
           handle, startPartition, endPartition, 0, Integer.MAX_VALUE, context, metrics);
     }
+    checkUserClassPathFirst(handle);
     return SparkUtils.getReader(
         sortShuffleManager(),
         handle,
@@ -328,6 +365,7 @@ public class SparkShuffleManager implements ShuffleManager {
       return getCelebornShuffleReader(
           handle, startPartition, endPartition, startMapIndex, endMapIndex, context, metrics);
     }
+    checkUserClassPathFirst(handle);
     return SparkUtils.getReader(
         sortShuffleManager(),
         handle,
@@ -386,6 +424,22 @@ public class SparkShuffleManager implements ShuffleManager {
     }
   }
 
+  private void checkUserClassPathFirst(ShuffleHandle handle) {
+    if ((Boolean) conf.get(package$.MODULE$.EXECUTOR_USER_CLASS_PATH_FIRST())
+        && !Objects.equals(
+            handle.getClass().getClassLoader(), CelebornShuffleHandle.class.getClassLoader())) {
+      String key = package$.MODULE$.EXECUTOR_USER_CLASS_PATH_FIRST().key();
+      Boolean defaultValue =
+          (Boolean) package$.MODULE$.EXECUTOR_USER_CLASS_PATH_FIRST().defaultValue().get();
+      logger.warn(
+          "Detected {} (default is {}) is enabled, "
+              + "it's highly recommended to disable it when use Celeborn as Remote Shuffle Service "
+              + "to avoid falling back to vanilla Spark SortShuffleManager for ShuffleManger defined in user jar.",
+          key,
+          defaultValue);
+    }
+  }
+
   // for testing
   public LifecycleManager getLifecycleManager() {
     return this.lifecycleManager;
