@@ -29,6 +29,7 @@ import scala.Tuple2;
 import scala.reflect.ClassTag$;
 
 import com.google.common.annotations.VisibleForTesting;
+import com.google.common.collect.Lists;
 import io.netty.buffer.CompositeByteBuf;
 import io.netty.buffer.Unpooled;
 import org.apache.commons.lang3.StringUtils;
@@ -46,9 +47,12 @@ import org.apache.celeborn.common.network.TransportContext;
 import org.apache.celeborn.common.network.buffer.NettyManagedBuffer;
 import org.apache.celeborn.common.network.client.RpcResponseCallback;
 import org.apache.celeborn.common.network.client.TransportClient;
+import org.apache.celeborn.common.network.client.TransportClientBootstrap;
 import org.apache.celeborn.common.network.client.TransportClientFactory;
 import org.apache.celeborn.common.network.protocol.PushData;
 import org.apache.celeborn.common.network.protocol.PushMergedData;
+import org.apache.celeborn.common.network.sasl.SaslClientBootstrap;
+import org.apache.celeborn.common.network.sasl.SaslCredentials;
 import org.apache.celeborn.common.network.server.BaseMessageHandler;
 import org.apache.celeborn.common.network.util.TransportConf;
 import org.apache.celeborn.common.protocol.*;
@@ -84,6 +88,7 @@ public class ShuffleClientImpl extends ShuffleClient {
 
   protected RpcEndpointRef lifecycleManagerRef;
 
+  private TransportContext transportContext;
   protected TransportClientFactory dataClientFactory;
 
   protected final int BATCH_HEADER_SIZE = 4 * 4;
@@ -115,10 +120,11 @@ public class ShuffleClientImpl extends ShuffleClient {
 
   private final ExecutorService pushDataRetryPool;
 
-  private final ExecutorService partitionSplitPool;
   private final Map<Integer, Set<Integer>> splitting = JavaUtils.newConcurrentHashMap();
 
   protected final String appUniqueId;
+  private final boolean authEnabled;
+  private final TransportConf dataTransportConf;
 
   private final ThreadLocal<Compressor> compressorThreadLocal =
       new ThreadLocal<Compressor>() {
@@ -178,31 +184,60 @@ public class ShuffleClientImpl extends ShuffleClient {
     } else {
       pushDataTimeout = conf.pushDataTimeoutMs();
     }
+    authEnabled = conf.authEnabledOnClient();
 
     // init rpc env
-    rpcEnv = RpcEnv.create(RpcNameConstants.SHUFFLE_CLIENT_SYS, Utils.localHostName(conf), 0, conf);
+    rpcEnv =
+        RpcEnv.create(
+            RpcNameConstants.SHUFFLE_CLIENT_SYS,
+            TransportModuleConstants.RPC_APP_CLIENT_MODULE,
+            Utils.localHostName(conf),
+            0,
+            conf,
+            scala.None$.empty());
 
     String module = TransportModuleConstants.DATA_MODULE;
-    TransportConf dataTransportConf =
+    dataTransportConf =
         Utils.fromCelebornConf(conf, module, conf.getInt("celeborn." + module + ".io.threads", 8));
-    TransportContext context =
-        new TransportContext(
-            dataTransportConf, new BaseMessageHandler(), conf.clientCloseIdleConnections());
-    dataClientFactory = context.createClientFactory();
-
+    initDataClientFactoryIfNeeded();
     int pushDataRetryThreads = conf.clientPushRetryThreads();
     pushDataRetryPool =
         ThreadUtils.newDaemonCachedThreadPool("celeborn-retry-sender", pushDataRetryThreads, 60);
 
-    int pushSplitPartitionThreads = conf.clientPushSplitPartitionThreads();
-    partitionSplitPool =
-        ThreadUtils.newDaemonCachedThreadPool(
-            "celeborn-shuffle-split", pushSplitPartitionThreads, 60);
     reviveManager = new ReviveManager(this, conf);
 
     logger.info("Created ShuffleClientImpl, appUniqueId: {}", appUniqueId);
   }
 
+  private void initDataClientFactoryIfNeeded() {
+    if (dataClientFactory != null) {
+      return;
+    }
+    this.transportContext =
+        new TransportContext(
+            dataTransportConf, new BaseMessageHandler(), conf.clientCloseIdleConnections());
+    if (!authEnabled) {
+      logger.info("Initializing data client factory for {}.", appUniqueId);
+      dataClientFactory = transportContext.createClientFactory();
+    } else if (lifecycleManagerRef != null) {
+      PbApplicationMetaRequest pbApplicationMetaRequest =
+          PbApplicationMetaRequest.newBuilder().setAppId(appUniqueId).build();
+      PbApplicationMeta pbApplicationMeta =
+          lifecycleManagerRef.askSync(
+              pbApplicationMetaRequest,
+              conf.clientRpcRegisterShuffleAskTimeout(),
+              ClassTag$.MODULE$.apply(PbApplicationMeta.class));
+      logger.info("Initializing data client factory for secured {}.", appUniqueId);
+      List<TransportClientBootstrap> bootstraps = Lists.newArrayList();
+      bootstraps.add(
+          new SaslClientBootstrap(
+              dataTransportConf,
+              appUniqueId,
+              new SaslCredentials(appUniqueId, pbApplicationMeta.getSecret())));
+      dataClientFactory = transportContext.createClientFactory(bootstraps);
+    }
+  }
+
   private boolean isPushTargetWorkerExcluded(
       PartitionLocation location, RpcResponseCallback wrappedCallback) {
     // If pushExcludeWorkerOnFailureEnabled = false, pushExcludedWorkers should be empty.
@@ -284,6 +319,7 @@ public class ShuffleClientImpl extends ShuffleClient {
       try {
         if (!isPushTargetWorkerExcluded(newLoc, pushDataRpcResponseCallback)) {
           if (!testRetryRevive || remainReviveTimes < 1) {
+            assert dataClientFactory != null;
             TransportClient client =
                 dataClientFactory.createClient(newLoc.getHost(), newLoc.getPushPort(), partitionId);
             NettyManagedBuffer newBuffer = new NettyManagedBuffer(Unpooled.wrappedBuffer(body));
@@ -569,14 +605,15 @@ public class ShuffleClientImpl extends ShuffleClient {
         StatusCode respStatus = Utils.toStatusCode(response.getStatus());
         if (StatusCode.SUCCESS.equals(respStatus)) {
           ConcurrentHashMap<Integer, PartitionLocation> result = JavaUtils.newConcurrentHashMap();
-          for (int i = 0; i < response.getPartitionLocationsList().size(); i++) {
-            PartitionLocation partitionLoc =
-                PbSerDeUtils.fromPbPartitionLocation(response.getPartitionLocationsList().get(i));
-            pushExcludedWorkers.remove(partitionLoc.hostAndPushPort());
-            if (partitionLoc.hasPeer()) {
-              pushExcludedWorkers.remove(partitionLoc.getPeer().hostAndPushPort());
+          Tuple2<List<PartitionLocation>, List<PartitionLocation>> locations =
+              PbSerDeUtils.fromPbPackedPartitionLocationsPair(
+                  response.getPackedPartitionLocationsPair());
+          for (PartitionLocation location : locations._1) {
+            pushExcludedWorkers.remove(location.hostAndPushPort());
+            if (location.hasPeer()) {
+              pushExcludedWorkers.remove(location.getPeer().hostAndPushPort());
             }
-            result.put(partitionLoc.getId(), partitionLoc);
+            result.put(location.getId(), location);
           }
           return result;
         } else if (StatusCode.SLOT_NOT_AVAILABLE.equals(respStatus)) {
@@ -848,7 +885,7 @@ public class ShuffleClientImpl extends ShuffleClient {
           partitionId,
           -1,
           null,
-          StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE)) {
+          StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE_PRIMARY)) {
         throw new CelebornIOException(
             String.format("Revive for shuffle %s partition %d failed.", shuffleId, partitionId));
       }
@@ -964,7 +1001,19 @@ public class ShuffleClientImpl extends ShuffleClient {
                       attemptId,
                       partitionId,
                       nextBatchId);
-                  splitPartition(shuffleId, partitionId, latest);
+                  if (!newerPartitionLocationExists(
+                      reducePartitionMap.get(shuffleId), partitionId, latest.getEpoch(), false)) {
+                    ReviveRequest reviveRequest =
+                        new ReviveRequest(
+                            shuffleId,
+                            mapId,
+                            attemptId,
+                            partitionId,
+                            latest.getEpoch(),
+                            latest,
+                            StatusCode.SOFT_SPLIT);
+                    reviveManager.addRequest(reviveRequest);
+                  }
                   pushState.onSuccess(latest.hostAndPushPort());
                   pushState.removeBatch(nextBatchId, latest.hostAndPushPort());
                   callback.onSuccess(response);
@@ -1106,13 +1155,14 @@ public class ShuffleClientImpl extends ShuffleClient {
       try {
         if (!isPushTargetWorkerExcluded(loc, wrappedCallback)) {
           if (!testRetryRevive) {
+            assert dataClientFactory != null;
             TransportClient client =
                 dataClientFactory.createClient(loc.getHost(), loc.getPushPort(), partitionId);
             client.pushData(pushData, pushDataTimeout, wrappedCallback);
           } else {
             wrappedCallback.onFailure(
                 new CelebornIOException(
-                    StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE,
+                    StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE_PRIMARY,
                     new RuntimeException("Mock push data first time failed.")));
           }
         }
@@ -1151,33 +1201,6 @@ public class ShuffleClientImpl extends ShuffleClient {
     return body.length;
   }
 
-  private void splitPartition(int shuffleId, int partitionId, PartitionLocation loc) {
-    Set<Integer> splittingSet =
-        splitting.computeIfAbsent(shuffleId, integer -> ConcurrentHashMap.newKeySet());
-    //noinspection SynchronizationOnLocalVariableOrMethodParameter
-    synchronized (splittingSet) {
-      if (splittingSet.contains(partitionId)) {
-        logger.debug(
-            "Splitting for shuffle {} partition {}, skip split request.", shuffleId, partitionId);
-        return;
-      }
-      splittingSet.add(partitionId);
-    }
-
-    ConcurrentHashMap<Integer, PartitionLocation> currentShuffleLocs =
-        reducePartitionMap.get(shuffleId);
-
-    ShuffleClientHelper.sendShuffleSplitAsync(
-        lifecycleManagerRef,
-        conf,
-        PartitionSplit$.MODULE$.apply(shuffleId, partitionId, loc.getEpoch(), loc),
-        partitionSplitPool,
-        splittingSet,
-        partitionId,
-        shuffleId,
-        currentShuffleLocs);
-  }
-
   @Override
   public int pushData(
       int shuffleId,
@@ -1268,9 +1291,9 @@ public class ShuffleClientImpl extends ShuffleClient {
       PushState pushState,
       int remainReviveTimes) {
     String hostPort = addressPair.getLeft();
-    final String[] splits = hostPort.split(":");
-    final String host = splits[0];
-    final int port = Integer.parseInt(splits[1]);
+    String[] hostPortArr = Utils.parseColonSeparatedHostPorts(hostPort, 1);
+    final String host = hostPortArr[0];
+    final int port = Integer.parseInt(hostPortArr[1]);
 
     int groupedBatchId = pushState.nextBatchId();
     pushState.addBatch(groupedBatchId, hostPort);
@@ -1484,12 +1507,13 @@ public class ShuffleClientImpl extends ShuffleClient {
     try {
       if (!isPushTargetWorkerExcluded(batches.get(0).loc, wrappedCallback)) {
         if (!testRetryRevive || remainReviveTimes < 1) {
+          assert dataClientFactory != null;
           TransportClient client = dataClientFactory.createClient(host, port);
           client.pushMergedData(mergedData, pushDataTimeout, wrappedCallback);
         } else {
           wrappedCallback.onFailure(
               new CelebornIOException(
-                  StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE,
+                  StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE_PRIMARY,
                   new RuntimeException("Mock push merge data failed.")));
         }
       }
@@ -1621,18 +1645,10 @@ public class ShuffleClientImpl extends ShuffleClient {
     }
   }
 
-  protected ReduceFileGroups updateFileGroup(int shuffleId, int partitionId)
+  public ReduceFileGroups updateFileGroup(int shuffleId, int partitionId)
       throws CelebornIOException {
     Tuple2<ReduceFileGroups, String> fileGroupTuple =
-        reduceFileGroupsMap.compute(
-            shuffleId,
-            (id, existsTuple) -> {
-              if (existsTuple == null || existsTuple._1 == null) {
-                return loadFileGroupInternal(shuffleId);
-              } else {
-                return existsTuple;
-              }
-            });
+        reduceFileGroupsMap.computeIfAbsent(shuffleId, (id) -> loadFileGroupInternal(shuffleId));
     if (fileGroupTuple._1 == null) {
       throw new CelebornIOException(
           loadFileGroupException(shuffleId, partitionId, (fileGroupTuple._2)));
@@ -1658,26 +1674,40 @@ public class ShuffleClientImpl extends ShuffleClient {
       int startMapIndex,
       int endMapIndex,
       ExceptionMaker exceptionMaker,
+      ArrayList<PartitionLocation> locations,
+      ArrayList<PbStreamHandler> streamHandlers,
+      int[] mapAttempts,
       MetricsCallback metricsCallback)
       throws IOException {
-    if (shuffleId == Utils$.MODULE$.UNKNOWN_APP_SHUFFLE_ID()) {
+    if (partitionId == Utils$.MODULE$.UNKNOWN_APP_SHUFFLE_ID()) {
       logger.warn("Shuffle data is empty for shuffle {}: UNKNOWN_APP_SHUFFLE_ID.", shuffleId);
       return CelebornInputStream.empty();
     }
-    ReduceFileGroups fileGroups = updateFileGroup(shuffleId, partitionId);
 
-    if (fileGroups.partitionGroups.isEmpty()
-        || !fileGroups.partitionGroups.containsKey(partitionId)) {
+    // When `mapAttempts` is not null, it's guaranteed that the code path comes from
+    // CelebornShuffleReader, which means `updateFileGroup` is already called and
+    // batch open stream has been tried
+    if (mapAttempts == null) {
+      ReduceFileGroups fileGroups = updateFileGroup(shuffleId, partitionId);
+      mapAttempts = fileGroups.mapAttempts;
+      if (fileGroups.partitionGroups.containsKey(partitionId)) {
+        locations = new ArrayList(fileGroups.partitionGroups.get(partitionId));
+      }
+    }
+
+    if (locations == null || locations.size() == 0) {
       logger.warn("Shuffle data is empty for shuffle {} partition {}.", shuffleId, partitionId);
       return CelebornInputStream.empty();
     } else {
       String shuffleKey = Utils.makeShuffleKey(appUniqueId, shuffleId);
+      assert dataClientFactory != null;
       return CelebornInputStream.create(
           conf,
           dataClientFactory,
           shuffleKey,
-          fileGroups.partitionGroups.get(partitionId).toArray(new PartitionLocation[0]),
-          fileGroups.mapAttempts,
+          locations,
+          streamHandlers,
+          mapAttempts,
           attemptNumber,
           startMapIndex,
           endMapIndex,
@@ -1707,12 +1737,12 @@ public class ShuffleClientImpl extends ShuffleClient {
     if (null != dataClientFactory) {
       dataClientFactory.close();
     }
+    if (null != transportContext) {
+      transportContext.close();
+    }
     if (null != pushDataRetryPool) {
       pushDataRetryPool.shutdown();
     }
-    if (null != partitionSplitPool) {
-      partitionSplitPool.shutdown();
-    }
     if (null != lifecycleManagerRef) {
       lifecycleManagerRef = null;
     }
@@ -1727,11 +1757,13 @@ public class ShuffleClientImpl extends ShuffleClient {
   public void setupLifecycleManagerRef(String host, int port) {
     lifecycleManagerRef =
         rpcEnv.setupEndpointRef(new RpcAddress(host, port), RpcNameConstants.LIFECYCLE_MANAGER_EP);
+    initDataClientFactoryIfNeeded();
   }
 
   @Override
   public void setupLifecycleManagerRef(RpcEndpointRef endpointRef) {
     lifecycleManagerRef = endpointRef;
+    initDataClientFactoryIfNeeded();
   }
 
   @Override
@@ -1753,7 +1785,9 @@ public class ShuffleClientImpl extends ShuffleClient {
     StatusCode cause;
     if (message == null) {
       logger.error("Push data throw unexpected exception");
-      cause = StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE;
+      cause = StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE_PRIMARY;
+    } else if (message.startsWith(StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE_REPLICA.name())) {
+      cause = StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE_REPLICA;
     } else if (message.startsWith(StatusCode.PUSH_DATA_WRITE_FAIL_REPLICA.name())) {
       cause = StatusCode.PUSH_DATA_WRITE_FAIL_REPLICA;
     } else if (message.startsWith(StatusCode.PUSH_DATA_WRITE_FAIL_PRIMARY.name())) {
@@ -1776,21 +1810,15 @@ public class ShuffleClientImpl extends ShuffleClient {
       cause = StatusCode.PUSH_DATA_PRIMARY_WORKER_EXCLUDED;
     } else if (message.startsWith(StatusCode.PUSH_DATA_REPLICA_WORKER_EXCLUDED.name())) {
       cause = StatusCode.PUSH_DATA_REPLICA_WORKER_EXCLUDED;
-    } else if (connectFail(message)) {
+    } else if (ExceptionUtils.connectFail(message)) {
       // Throw when push to primary worker connection causeException.
       cause = StatusCode.PUSH_DATA_CONNECTION_EXCEPTION_PRIMARY;
     } else {
-      cause = StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE;
+      cause = StatusCode.PUSH_DATA_FAIL_NON_CRITICAL_CAUSE_PRIMARY;
     }
     return cause;
   }
 
-  private boolean connectFail(String message) {
-    return (message.startsWith("Connection from ") && message.endsWith(" closed"))
-        || message.equals("Connection reset by peer")
-        || message.startsWith("Failed to send RPC ");
-  }
-
   @VisibleForTesting
   public TransportClientFactory getDataClientFactory() {
     return dataClientFactory;
