@@ -25,10 +25,10 @@ import java.util.concurrent.TimeUnit;
 import java.util.concurrent.TimeoutException;
 import java.util.concurrent.atomic.LongAdder;
 
-import com.github.luben.zstd.ZstdException;
+import scala.Tuple2;
+
 import com.google.common.util.concurrent.Uninterruptibles;
 import io.netty.buffer.ByteBuf;
-import net.jpountz.lz4.LZ4Exception;
 import org.roaringbitmap.RoaringBitmap;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
@@ -38,10 +38,7 @@ import org.apache.celeborn.client.compress.Decompressor;
 import org.apache.celeborn.common.CelebornConf;
 import org.apache.celeborn.common.exception.CelebornIOException;
 import org.apache.celeborn.common.network.client.TransportClientFactory;
-import org.apache.celeborn.common.protocol.CompressionCodec;
-import org.apache.celeborn.common.protocol.PartitionLocation;
-import org.apache.celeborn.common.protocol.StorageInfo;
-import org.apache.celeborn.common.protocol.TransportModuleConstants;
+import org.apache.celeborn.common.protocol.*;
 import org.apache.celeborn.common.unsafe.Platform;
 import org.apache.celeborn.common.util.ExceptionMaker;
 import org.apache.celeborn.common.util.Utils;
@@ -53,7 +50,8 @@ public abstract class CelebornInputStream extends InputStream {
       CelebornConf conf,
       TransportClientFactory clientFactory,
       String shuffleKey,
-      PartitionLocation[] locations,
+      ArrayList<PartitionLocation> locations,
+      ArrayList<PbStreamHandler> streamHandlers,
       int[] attempts,
       int attemptNumber,
       int startMapIndex,
@@ -66,7 +64,7 @@ public abstract class CelebornInputStream extends InputStream {
       ExceptionMaker exceptionMaker,
       MetricsCallback metricsCallback)
       throws IOException {
-    if (locations == null || locations.length == 0) {
+    if (locations == null || locations.size() == 0) {
       return emptyInputStream;
     } else {
       return new CelebornInputStreamImpl(
@@ -74,6 +72,7 @@ public abstract class CelebornInputStream extends InputStream {
           clientFactory,
           shuffleKey,
           locations,
+          streamHandlers,
           attempts,
           attemptNumber,
           startMapIndex,
@@ -125,7 +124,8 @@ public abstract class CelebornInputStream extends InputStream {
     private final CelebornConf conf;
     private final TransportClientFactory clientFactory;
     private final String shuffleKey;
-    private PartitionLocation[] locations;
+    private ArrayList<PartitionLocation> locations;
+    private ArrayList<PbStreamHandler> streamHandlers;
     private int[] attempts;
     private final int attemptNumber;
     private final int startMapIndex;
@@ -175,7 +175,8 @@ public abstract class CelebornInputStream extends InputStream {
         CelebornConf conf,
         TransportClientFactory clientFactory,
         String shuffleKey,
-        PartitionLocation[] locations,
+        ArrayList<PartitionLocation> locations,
+        ArrayList<PbStreamHandler> streamHandlers,
         int[] attempts,
         int attemptNumber,
         int startMapIndex,
@@ -191,7 +192,10 @@ public abstract class CelebornInputStream extends InputStream {
       this.conf = conf;
       this.clientFactory = clientFactory;
       this.shuffleKey = shuffleKey;
-      this.locations = (PartitionLocation[]) Utils.randomizeInPlace(locations, RAND);
+      this.locations = locations;
+      if (streamHandlers != null && streamHandlers.size() == locations.size()) {
+        this.streamHandlers = streamHandlers;
+      }
       this.attempts = attempts;
       this.attemptNumber = attemptNumber;
       this.startMapIndex = startMapIndex;
@@ -241,24 +245,25 @@ public abstract class CelebornInputStream extends InputStream {
       return true;
     }
 
-    private PartitionLocation nextReadableLocation() {
-      int locationCount = locations.length;
+    private Tuple2<PartitionLocation, PbStreamHandler> nextReadableLocation() {
+      int locationCount = locations.size();
       if (fileIndex >= locationCount) {
         return null;
       }
-      PartitionLocation currentLocation = locations[fileIndex];
+      PartitionLocation currentLocation = locations.get(fileIndex);
       while (skipLocation(startMapIndex, endMapIndex, currentLocation)) {
         skipCount.increment();
         fileIndex++;
         if (fileIndex == locationCount) {
           return null;
         }
-        currentLocation = locations[fileIndex];
+        currentLocation = locations.get(fileIndex);
       }
 
       fetchChunkRetryCnt = 0;
 
-      return currentLocation;
+      return new Tuple2(
+          currentLocation, streamHandlers == null ? null : streamHandlers.get(fileIndex));
     }
 
     private void moveToNextReader(boolean fetchChunk) throws IOException {
@@ -266,11 +271,11 @@ public abstract class CelebornInputStream extends InputStream {
         currentReader.close();
         currentReader = null;
       }
-      PartitionLocation currentLocation = nextReadableLocation();
+      Tuple2<PartitionLocation, PbStreamHandler> currentLocation = nextReadableLocation();
       if (currentLocation == null) {
         return;
       }
-      currentReader = createReaderWithRetry(currentLocation);
+      currentReader = createReaderWithRetry(currentLocation._1, currentLocation._2);
       fileIndex++;
       while (!currentReader.hasNext()) {
         currentReader.close();
@@ -279,7 +284,7 @@ public abstract class CelebornInputStream extends InputStream {
         if (currentLocation == null) {
           return;
         }
-        currentReader = createReaderWithRetry(currentLocation);
+        currentReader = createReaderWithRetry(currentLocation._1, currentLocation._2);
         fileIndex++;
       }
       if (fetchChunk) {
@@ -331,14 +336,15 @@ public abstract class CelebornInputStream extends InputStream {
       return connectException || rpcTimeout || fetchChunkTimeout;
     }
 
-    private PartitionReader createReaderWithRetry(PartitionLocation location) throws IOException {
+    private PartitionReader createReaderWithRetry(
+        PartitionLocation location, PbStreamHandler pbStreamHandler) throws IOException {
       Exception lastException = null;
       while (fetchChunkRetryCnt < fetchChunkMaxRetry) {
         try {
           if (isExcluded(location)) {
             throw new CelebornIOException("Fetch data from excluded worker! " + location);
           }
-          return createReader(location, fetchChunkRetryCnt, fetchChunkMaxRetry);
+          return createReader(location, pbStreamHandler, fetchChunkRetryCnt, fetchChunkMaxRetry);
         } catch (Exception e) {
           lastException = e;
           excludeFailedLocation(location, e);
@@ -403,7 +409,7 @@ public abstract class CelebornInputStream extends InputStream {
               if (fetchChunkRetryCnt % 2 == 0) {
                 Uninterruptibles.sleepUninterruptibly(retryWaitMs, TimeUnit.MILLISECONDS);
               }
-              currentReader = createReaderWithRetry(currentReader.getLocation().getPeer());
+              currentReader = createReaderWithRetry(currentReader.getLocation().getPeer(), null);
             } else {
               logger.warn(
                   "Fetch chunk failed {}/{} times for location {}",
@@ -412,7 +418,7 @@ public abstract class CelebornInputStream extends InputStream {
                   currentReader.getLocation(),
                   e);
               Uninterruptibles.sleepUninterruptibly(retryWaitMs, TimeUnit.MILLISECONDS);
-              currentReader = createReaderWithRetry(currentReader.getLocation());
+              currentReader = createReaderWithRetry(currentReader.getLocation(), null);
             }
           }
         }
@@ -421,11 +427,14 @@ public abstract class CelebornInputStream extends InputStream {
     }
 
     private PartitionReader createReader(
-        PartitionLocation location, int fetchChunkRetryCnt, int fetchChunkMaxRetry)
+        PartitionLocation location,
+        PbStreamHandler pbStreamHandler,
+        int fetchChunkRetryCnt,
+        int fetchChunkMaxRetry)
         throws IOException, InterruptedException {
       if (!location.hasPeer()) {
         logger.debug("Partition {} has only one partition replica.", location);
-      } else if (attemptNumber % 2 == 1) {
+      } else if (pbStreamHandler == null && attemptNumber % 2 == 1) {
         location = location.getPeer();
         logger.debug("Read peer {} for attempt {}.", location, attemptNumber);
       }
@@ -435,7 +444,10 @@ public abstract class CelebornInputStream extends InputStream {
       switch (storageInfo.getType()) {
         case HDD:
         case SSD:
-          if (enabledReadLocalShuffle && location.getWorker().host().equals(localHostAddress)) {
+        case MEMORY:
+          if (enabledReadLocalShuffle
+              && location.getHost().equals(localHostAddress)
+              && storageInfo.getType() != StorageInfo.Type.MEMORY) {
             logger.debug("Read local shuffle file {}", localHostAddress);
             containLocalRead = true;
             return new LocalPartitionReader(
@@ -445,6 +457,7 @@ public abstract class CelebornInputStream extends InputStream {
                 conf,
                 shuffleKey,
                 location,
+                pbStreamHandler,
                 clientFactory,
                 startMapIndex,
                 endMapIndex,
@@ -512,7 +525,7 @@ public abstract class CelebornInputStream extends InputStream {
     @Override
     public synchronized void close() {
       if (!closed) {
-        int locationsCount = locations.length;
+        int locationsCount = locations.size();
         logger.debug(
             "AppShuffleId {}, shuffleId {}, partitionId {}, total location count {}, read {}, skip {}",
             appShuffleId,
@@ -555,7 +568,7 @@ public abstract class CelebornInputStream extends InputStream {
       if (currentReader.hasNext()) {
         currentChunk = getNextChunk();
         return true;
-      } else if (fileIndex < locations.length) {
+      } else if (fileIndex < locations.size()) {
         moveToNextReader(true);
         return currentReader != null;
       }
@@ -646,20 +659,8 @@ public abstract class CelebornInputStream extends InputStream {
         }
 
         return hasData;
-      } catch (LZ4Exception | ZstdException | IOException e) {
-        logger.error(
-            "Failed to fill buffer from chunk. AppShuffleId {}, shuffleId {}, partitionId {}, location {}",
-            appShuffleId,
-            shuffleId,
-            partitionId,
-            currentReader.getLocation(),
-            e);
-        IOException ioe;
-        if (e instanceof IOException) {
-          ioe = (IOException) e;
-        } else {
-          ioe = new IOException(e);
-        }
+      } catch (IOException e) {
+        IOException ioe = e;
         if (exceptionMaker != null) {
           if (shuffleClient.reportShuffleFetchFailure(appShuffleId, shuffleId)) {
             /*
@@ -674,21 +675,12 @@ public abstract class CelebornInputStream extends InputStream {
           }
         }
         throw ioe;
-      } catch (Exception e) {
-        logger.error(
-            "Failed to fill buffer from chunk. AppShuffleId {}, shuffleId {}, partitionId {}, location {}",
-            appShuffleId,
-            shuffleId,
-            partitionId,
-            currentReader.getLocation(),
-            e);
-        throw e;
       }
     }
 
     @Override
     public int totalPartitionsToRead() {
-      return locations.length;
+      return locations.size();
     }
 
     @Override
