@@ -26,7 +26,7 @@ import scala.collection.mutable
 
 import com.google.common.cache.{Cache, CacheBuilder}
 
-import org.apache.celeborn.client.{ClientUtils, ShuffleCommittedInfo, WorkerStatusTracker}
+import org.apache.celeborn.client.{ShuffleCommittedInfo, WorkerStatusTracker}
 import org.apache.celeborn.client.CommitManager.CommittedPartitionInfo
 import org.apache.celeborn.client.LifecycleManager.{ShuffleAllocatedWorkers, ShuffleFailedWorkers}
 import org.apache.celeborn.common.CelebornConf
@@ -245,10 +245,10 @@ class ReducePartitionCommitHandler(
       if (attempts(mapId) < 0) {
         attempts(mapId) = attemptId
         // Mapper with this attemptId finished, also check all other mapper finished or not.
-        (true, ClientUtils.areAllMapperAttemptsFinished(attempts))
+        (true, !attempts.exists(_ < 0))
       } else {
         // Mapper with another attemptId finished, skip this request
-        (false, false)
+        (false, !attempts.exists(_ < 0))
       }
     }
   }
@@ -275,7 +275,8 @@ class ReducePartitionCommitHandler(
         GetReducerFileGroupResponse(
           StatusCode.SHUFFLE_DATA_LOST,
           JavaUtils.newConcurrentHashMap(),
-          Array.empty))
+          Array.empty,
+          new util.HashSet[Integer]()))
     } else {
       // LocalNettyRpcCallContext is for the UTs
       if (context.isInstanceOf[LocalNettyRpcCallContext]) {
