@@ -20,16 +20,23 @@ package org.apache.celeborn.common.protocol;
 public class RpcNameConstants {
   // For Master
   public static String MASTER_SYS = "Master";
+  public static String MASTER_INTERNAL_SYS = "MasterInternal";
+
   // Master Endpoint Name
   public static String MASTER_EP = "MasterEndpoint";
+  public static String MASTER_INTERNAL_EP = "MasterInternalEndpoint";
 
   // For Worker
   public static String WORKER_SYS = "Worker";
+  public static String WORKER_INTERNAL_SYS = "WorkerInternal";
   // Worker Endpoint Name
   public static String WORKER_EP = "WorkerEndpoint";
+  public static String WORKER_INTERNAL_EP = "WorkerInternalEndpoint";
 
   // For LifecycleManager
   public static String LIFECYCLE_MANAGER_SYS = "LifecycleManager";
+  public static String LIFECYCLE_MANAGER_MASTER_SYS = "LifecycleManagerMasterSys";
+  public static String LIFECYCLE_MANAGER_WORKER_SYS = "LifecycleManagerWorkerSys";
   public static String LIFECYCLE_MANAGER_EP = "LifecycleManagerEndpoint";
 
   // For Shuffle Client
