@@ -26,11 +26,12 @@ import scala.collection.JavaConverters._
 import com.google.protobuf.InvalidProtocolBufferException
 
 import org.apache.celeborn.common.identity.UserIdentifier
-import org.apache.celeborn.common.meta.{AppDiskUsage, AppDiskUsageSnapShot, DiskInfo, FileInfo, WorkerInfo}
+import org.apache.celeborn.common.meta.{AppDiskUsage, AppDiskUsageSnapShot, ApplicationMeta, DiskFileInfo, DiskInfo, MapFileMeta, ReduceFileMeta, WorkerEventInfo, WorkerInfo, WorkerStatus}
 import org.apache.celeborn.common.protocol._
 import org.apache.celeborn.common.protocol.PartitionLocation.Mode
 import org.apache.celeborn.common.protocol.message.ControlMessages.WorkerResource
 import org.apache.celeborn.common.quota.ResourceConsumption
+import org.apache.celeborn.common.util.{CollectionUtils => localCollectionUtils}
 
 object PbSerDeUtils {
 
@@ -71,6 +72,7 @@ object PbSerDeUtils {
       pbDiskInfo.getAvgFetchTime,
       pbDiskInfo.getUsedSlots)
       .setStatus(Utils.toDiskStatus(pbDiskInfo.getStatus))
+      .setTotalSpace(pbDiskInfo.getTotalSpace)
     diskInfo.setStorageType(StorageInfo.typesMap.get(pbDiskInfo.getStorageType))
     diskInfo
   }
@@ -84,40 +86,54 @@ object PbSerDeUtils {
       .setUsedSlots(diskInfo.activeSlots)
       .setStatus(diskInfo.status.getValue)
       .setStorageType(diskInfo.storageType.getValue)
+      .setTotalSpace(diskInfo.totalSpace)
       .build
 
-  def fromPbFileInfo(pbFileInfo: PbFileInfo): FileInfo =
+  def fromPbFileInfo(pbFileInfo: PbFileInfo): DiskFileInfo =
     fromPbFileInfo(pbFileInfo, fromPbUserIdentifier(pbFileInfo.getUserIdentifier))
 
-  def fromPbFileInfo(pbFileInfo: PbFileInfo, userIdentifier: UserIdentifier) =
-    new FileInfo(
-      pbFileInfo.getFilePath,
-      pbFileInfo.getChunkOffsetsList,
+  def fromPbFileInfo(pbFileInfo: PbFileInfo, userIdentifier: UserIdentifier) = {
+    val meta = Utils.toPartitionType(pbFileInfo.getPartitionType) match {
+      case PartitionType.REDUCE =>
+        new ReduceFileMeta(pbFileInfo.getChunkOffsetsList)
+      case PartitionType.MAP =>
+        new MapFileMeta(pbFileInfo.getBufferSize, pbFileInfo.getNumSubpartitions)
+      case PartitionType.MAPGROUP =>
+        throw new NotImplementedError("Map group is not implemented")
+    }
+    new DiskFileInfo(
       userIdentifier,
-      Utils.toPartitionType(pbFileInfo.getPartitionType),
-      pbFileInfo.getBufferSize,
-      pbFileInfo.getNumSubpartitions,
-      pbFileInfo.getBytesFlushed,
-      pbFileInfo.getPartitionSplitEnabled)
+      pbFileInfo.getPartitionSplitEnabled,
+      meta,
+      pbFileInfo.getFilePath,
+      pbFileInfo.getBytesFlushed)
+  }
 
-  def toPbFileInfo(fileInfo: FileInfo): PbFileInfo =
-    PbFileInfo.newBuilder
+  def toPbFileInfo(fileInfo: DiskFileInfo): PbFileInfo = {
+    val builder = PbFileInfo.newBuilder
       .setFilePath(fileInfo.getFilePath)
-      .addAllChunkOffsets(fileInfo.getChunkOffsets)
       .setUserIdentifier(toPbUserIdentifier(fileInfo.getUserIdentifier))
-      .setPartitionType(fileInfo.getPartitionType.getValue)
-      .setBufferSize(fileInfo.getBufferSize)
-      .setNumSubpartitions(fileInfo.getNumSubpartitions)
       .setBytesFlushed(fileInfo.getFileLength)
       .setPartitionSplitEnabled(fileInfo.isPartitionSplitEnabled)
-      .build
+    if (fileInfo.getFileMeta.isInstanceOf[MapFileMeta]) {
+      val mapFileMeta = fileInfo.getFileMeta.asInstanceOf[MapFileMeta]
+      builder.setPartitionType(PartitionType.MAP.getValue)
+      builder.setBufferSize(mapFileMeta.getBufferSize)
+      builder.setNumSubpartitions(mapFileMeta.getNumSubpartitions)
+    } else {
+      val reduceFileMeta = fileInfo.getFileMeta.asInstanceOf[ReduceFileMeta]
+      builder.setPartitionType(PartitionType.REDUCE.getValue)
+      builder.addAllChunkOffsets(reduceFileMeta.getChunkOffsets)
+    }
+    builder.build
+  }
 
   @throws[InvalidProtocolBufferException]
   def fromPbFileInfoMap(
       data: Array[Byte],
-      cache: ConcurrentHashMap[String, UserIdentifier]): ConcurrentHashMap[String, FileInfo] = {
+      cache: ConcurrentHashMap[String, UserIdentifier]): ConcurrentHashMap[String, DiskFileInfo] = {
     val pbFileInfoMap = PbFileInfoMap.parseFrom(data)
-    val fileInfoMap = JavaUtils.newConcurrentHashMap[String, FileInfo]
+    val fileInfoMap = JavaUtils.newConcurrentHashMap[String, DiskFileInfo]
     pbFileInfoMap.getValuesMap.entrySet().asScala.foreach { entry =>
       val fileName = entry.getKey
       val pbFileInfo = entry.getValue
@@ -135,7 +151,7 @@ object PbSerDeUtils {
     fileInfoMap
   }
 
-  def toPbFileInfoMap(fileInfoMap: ConcurrentHashMap[String, FileInfo]): Array[Byte] = {
+  def toPbFileInfoMap(fileInfoMap: ConcurrentHashMap[String, DiskFileInfo]): Array[Byte] = {
     val pbFileInfoMap = JavaUtils.newConcurrentHashMap[String, PbFileInfo]()
     fileInfoMap.entrySet().asScala.foreach { entry =>
       pbFileInfoMap.put(entry.getKey, toPbFileInfo(entry.getValue))
@@ -217,27 +233,32 @@ object PbSerDeUtils {
       pbWorkerInfo.getPushPort,
       pbWorkerInfo.getFetchPort,
       pbWorkerInfo.getReplicatePort,
+      pbWorkerInfo.getInternalPort,
       disks,
       userResourceConsumption)
   }
 
   def toPbWorkerInfo(
       workerInfo: WorkerInfo,
-      eliminateUserResourceConsumption: Boolean): PbWorkerInfo = {
-    val diskInfos = workerInfo.diskInfos.values
-    val pbDiskInfos = new util.ArrayList[PbDiskInfo]()
-    diskInfos.asScala.foreach(diskInfo => pbDiskInfos.add(PbSerDeUtils.toPbDiskInfo(diskInfo)))
+      eliminateUserResourceConsumption: Boolean,
+      eliminateDiskInfo: Boolean): PbWorkerInfo = {
     val builder = PbWorkerInfo.newBuilder
       .setHost(workerInfo.host)
       .setRpcPort(workerInfo.rpcPort)
       .setFetchPort(workerInfo.fetchPort)
       .setPushPort(workerInfo.pushPort)
       .setReplicatePort(workerInfo.replicatePort)
-      .addAllDisks(pbDiskInfos)
+      .setInternalPort(workerInfo.internalPort)
     if (!eliminateUserResourceConsumption) {
       builder.putAllUserResourceConsumption(
         PbSerDeUtils.toPbUserResourceConsumption(workerInfo.userResourceConsumption))
     }
+    if (!eliminateDiskInfo) {
+      val diskInfos = workerInfo.diskInfos.values
+      val pbDiskInfos = new util.ArrayList[PbDiskInfo]()
+      diskInfos.asScala.foreach(diskInfo => pbDiskInfos.add(PbSerDeUtils.toPbDiskInfo(diskInfo)))
+      builder.addAllDisks(pbDiskInfos)
+    }
     builder.build
   }
 
@@ -321,10 +342,8 @@ object PbSerDeUtils {
   def fromPbWorkerResource(pbWorkerResource: util.Map[String, PbWorkerResource]): WorkerResource = {
     val slots = new WorkerResource()
     pbWorkerResource.asScala.foreach { case (uniqueId, pbWorkerResource) =>
-      val Array(host, rpcPort, pushPort, fetchPort, replicatePort) = uniqueId.split(":")
       val networkLocation = pbWorkerResource.getNetworkLocation
-      val workerInfo =
-        new WorkerInfo(host, rpcPort.toInt, pushPort.toInt, fetchPort.toInt, replicatePort.toInt)
+      val workerInfo = WorkerInfo.fromUniqueId(uniqueId)
       workerInfo.networkLocation = networkLocation
       val primaryPartitionLocation = new util.ArrayList[PartitionLocation](pbWorkerResource
         .getPrimaryPartitionsList.asScala.map(PbSerDeUtils.fromPbPartitionLocation).asJava)
@@ -397,23 +416,25 @@ object PbSerDeUtils {
       manuallyExcludedWorkers: java.util.Set[WorkerInfo],
       workerLostEvent: java.util.Set[WorkerInfo],
       appHeartbeatTime: java.util.Map[String, java.lang.Long],
-      workers: java.util.List[WorkerInfo],
+      workers: java.util.Set[WorkerInfo],
       partitionTotalWritten: java.lang.Long,
       partitionTotalFileCount: java.lang.Long,
       appDiskUsageMetricSnapshots: Array[AppDiskUsageSnapShot],
       currentAppDiskUsageMetricsSnapshot: AppDiskUsageSnapShot,
       lostWorkers: ConcurrentHashMap[WorkerInfo, java.lang.Long],
-      shutdownWorkers: java.util.Set[WorkerInfo]): PbSnapshotMetaInfo = {
+      shutdownWorkers: java.util.Set[WorkerInfo],
+      workerEventInfos: ConcurrentHashMap[WorkerInfo, WorkerEventInfo],
+      applicationMetas: ConcurrentHashMap[String, ApplicationMeta]): PbSnapshotMetaInfo = {
     val builder = PbSnapshotMetaInfo.newBuilder()
       .setEstimatedPartitionSize(estimatedPartitionSize)
       .addAllRegisteredShuffle(registeredShuffle)
       .addAllHostnameSet(hostnameSet)
-      .addAllExcludedWorkers(excludedWorkers.asScala.map(toPbWorkerInfo(_, true)).asJava)
+      .addAllExcludedWorkers(excludedWorkers.asScala.map(toPbWorkerInfo(_, true, false)).asJava)
       .addAllManuallyExcludedWorkers(manuallyExcludedWorkers.asScala
-        .map(toPbWorkerInfo(_, true)).asJava)
-      .addAllWorkerLostEvents(workerLostEvent.asScala.map(toPbWorkerInfo(_, true)).asJava)
+        .map(toPbWorkerInfo(_, true, false)).asJava)
+      .addAllWorkerLostEvents(workerLostEvent.asScala.map(toPbWorkerInfo(_, true, false)).asJava)
       .putAllAppHeartbeatTime(appHeartbeatTime)
-      .addAllWorkers(workers.asScala.map(toPbWorkerInfo(_, true)).asJava)
+      .addAllWorkers(workers.asScala.map(toPbWorkerInfo(_, true, false)).asJava)
       .setPartitionTotalWritten(partitionTotalWritten)
       .setPartitionTotalFileCount(partitionTotalFileCount)
       // appDiskUsageMetricSnapshots can have null values,
@@ -423,11 +444,221 @@ object PbSerDeUtils {
       .putAllLostWorkers(lostWorkers.asScala.map {
         case (worker: WorkerInfo, time: java.lang.Long) => (worker.toUniqueId(), time)
       }.asJava)
-      .addAllShutdownWorkers(shutdownWorkers.asScala.map(toPbWorkerInfo(_, true)).asJava)
+      .addAllShutdownWorkers(shutdownWorkers.asScala.map(toPbWorkerInfo(_, true, false)).asJava)
+      .putAllWorkerEventInfos(workerEventInfos.asScala.map {
+        case (worker, workerEventInfo) =>
+          (worker.toUniqueId(), PbSerDeUtils.toPbWorkerEventInfo(workerEventInfo))
+      }.asJava)
     if (currentAppDiskUsageMetricsSnapshot != null) {
       builder.setCurrentAppDiskUsageMetricsSnapshot(
         toPbAppDiskUsageSnapshot(currentAppDiskUsageMetricsSnapshot))
     }
+    val pbApplicationMetas = applicationMetas.asScala.map {
+      case (appId, applicationMeta) => (appId, toPbApplicationMeta(applicationMeta))
+    }.asJava
+    if (localCollectionUtils.isNotEmpty(pbApplicationMetas)) {
+      builder.putAllApplicationMetas(pbApplicationMetas)
+    }
     builder.build()
   }
+
+  def toPbApplicationMeta(meta: ApplicationMeta): PbApplicationMeta = {
+    PbApplicationMeta.newBuilder()
+      .setAppId(meta.appId)
+      .setSecret(meta.secret).build()
+  }
+
+  def fromPbApplicationMeta(pbApplicationMeta: PbApplicationMeta): ApplicationMeta = {
+    new ApplicationMeta(pbApplicationMeta.getAppId, pbApplicationMeta.getSecret)
+  }
+
+  def toPbWorkerStatus(workerStatus: WorkerStatus): PbWorkerStatus = {
+    PbWorkerStatus.newBuilder()
+      .setState(workerStatus.getState)
+      .setStateStartTime(workerStatus.getStateStartTime)
+      .build()
+  }
+
+  def fromPbWorkerStatus(pbWorkerStatus: PbWorkerStatus): WorkerStatus = {
+    new WorkerStatus(pbWorkerStatus.getState.getNumber, pbWorkerStatus.getStateStartTime)
+  }
+
+  def toPbWorkerEventInfo(workerEventInfo: WorkerEventInfo): PbWorkerEventInfo = {
+    PbWorkerEventInfo.newBuilder()
+      .setEventStartTime(workerEventInfo.getEventStartTime)
+      .setWorkerEventType(workerEventInfo.getEventType)
+      .build()
+  }
+
+  def fromPbWorkerEventInfo(pbWorkerEventInfo: PbWorkerEventInfo): WorkerEventInfo = {
+    new WorkerEventInfo(
+      pbWorkerEventInfo.getWorkerEventType.getNumber,
+      pbWorkerEventInfo.getEventStartTime())
+  }
+
+  private def toPackedPartitionLocation(
+      pbPackedLocationsBuilder: PbPackedPartitionLocations.Builder,
+      workerIdIndex: Map[String, Int],
+      mountPointsIndex: Map[String, Int],
+      location: PartitionLocation): PbPackedPartitionLocations.Builder = {
+    pbPackedLocationsBuilder.addIds(location.getId)
+    pbPackedLocationsBuilder.addEpoches(location.getEpoch)
+    pbPackedLocationsBuilder.addWorkerIds(workerIdIndex(location.getWorker.toUniqueId()))
+    pbPackedLocationsBuilder.addMapIdBitMap(
+      Utils.roaringBitmapToByteString(location.getMapIdBitMap))
+    pbPackedLocationsBuilder.addTypes(location.getStorageInfo.getType.getValue)
+    pbPackedLocationsBuilder.addMountPoints(
+      mountPointsIndex(location.getStorageInfo.getMountPoint))
+    pbPackedLocationsBuilder.addFinalResult(location.getStorageInfo.isFinalResult)
+    if (location.getStorageInfo.getFilePath != null && location.getStorageInfo.getFilePath.nonEmpty) {
+      pbPackedLocationsBuilder.addFilePaths(location.getStorageInfo.getFilePath
+        .substring(location.getStorageInfo.getMountPoint.length))
+    } else {
+      pbPackedLocationsBuilder.addFilePaths("")
+    }
+    pbPackedLocationsBuilder.addAvailableStorageTypes(location.getStorageInfo.availableStorageTypes)
+    pbPackedLocationsBuilder.addModes(location.getMode.mode())
+  }
+
+  def toPbPackedPartitionLocationsPair(inputLocations: List[PartitionLocation])
+      : PbPackedPartitionLocationsPair = {
+    val packedLocationPairsBuilder = PbPackedPartitionLocationsPair.newBuilder()
+    val packedLocationsBuilder = PbPackedPartitionLocations.newBuilder()
+
+    val implicateLocations = inputLocations.map(_.getPeer).filterNot(_ == null)
+
+    val allLocations = (inputLocations ++ implicateLocations)
+    val workerIdList = new util.ArrayList[String](
+      allLocations.map(_.getWorker.toUniqueId()).toSet.asJava)
+    val workerIdIndex = workerIdList.asScala.zipWithIndex.toMap
+    val mountPointsList = new util.ArrayList[String](
+      allLocations.map(
+        _.getStorageInfo.getMountPoint).toSet.asJava)
+    val mountPointsIndex = mountPointsList.asScala.zipWithIndex.toMap
+
+    packedLocationsBuilder.addAllWorkerIdsSet(workerIdList)
+    packedLocationsBuilder.addAllMountPointsSet(mountPointsList)
+
+    val locationIndexes = allLocations.zipWithIndex.toMap
+
+    for (location <- allLocations) {
+      toPackedPartitionLocation(
+        packedLocationsBuilder,
+        workerIdIndex,
+        mountPointsIndex,
+        location)
+      if (location.getPeer != null) {
+        packedLocationPairsBuilder.addPeerIndexes(
+          locationIndexes(location.getPeer))
+      } else {
+        packedLocationPairsBuilder.addPeerIndexes(Integer.MAX_VALUE)
+      }
+    }
+
+    packedLocationPairsBuilder.setInputLocationSize(inputLocations.size)
+    packedLocationPairsBuilder.setLocations(packedLocationsBuilder.build()).build()
+  }
+
+  def fromPbPackedPartitionLocationsPair(pbPartitionLocationsPair: PbPackedPartitionLocationsPair)
+      : (util.List[PartitionLocation], util.List[PartitionLocation]) = {
+    val primaryLocations = new util.ArrayList[PartitionLocation]()
+    val replicateLocations = new util.ArrayList[PartitionLocation]()
+    val pbPackedPartitionLocations = pbPartitionLocationsPair.getLocations
+    val inputLocationSize = pbPartitionLocationsPair.getInputLocationSize
+    val idList = pbPackedPartitionLocations.getIdsList
+    val locationCount = idList.size()
+    var index = 0
+
+    val locations = new util.ArrayList[PartitionLocation]()
+    while (index < locationCount) {
+      val loc =
+        fromPackedPartitionLocations(pbPackedPartitionLocations, index)
+      if (index < inputLocationSize) {
+        if (loc.getMode == Mode.PRIMARY) {
+          primaryLocations.add(loc)
+        } else {
+          replicateLocations.add(loc)
+        }
+      }
+      locations.add(loc)
+      index = index + 1
+    }
+
+    index = 0
+    while (index < locationCount) {
+      val replicateIndex = pbPartitionLocationsPair.getPeerIndexes(index)
+      if (replicateIndex != Integer.MAX_VALUE) {
+        locations.get(index).setPeer(locations.get(replicateIndex))
+      }
+      index = index + 1
+    }
+
+    (primaryLocations, replicateLocations)
+  }
+
+  private def fromPackedPartitionLocations(
+      pbPackedPartitionLocations: PbPackedPartitionLocations,
+      index: Int): PartitionLocation = {
+    val workerIdParts = pbPackedPartitionLocations.getWorkerIdsSet(
+      pbPackedPartitionLocations.getWorkerIds(index)).split(":").map(_.trim)
+    var filePath = pbPackedPartitionLocations.getFilePaths(index)
+    if (filePath != "") {
+      filePath = pbPackedPartitionLocations.getMountPointsSet(
+        pbPackedPartitionLocations.getMountPoints(index)) +
+        pbPackedPartitionLocations.getFilePaths(index)
+    }
+
+    val mode =
+      if (pbPackedPartitionLocations.getModes(index) == Mode.PRIMARY.mode()) {
+        Mode.PRIMARY
+      } else {
+        Mode.REPLICA
+      }
+
+    new PartitionLocation(
+      pbPackedPartitionLocations.getIds(index),
+      pbPackedPartitionLocations.getEpoches(index),
+      workerIdParts(0),
+      workerIdParts(1).toInt,
+      workerIdParts(2).toInt,
+      workerIdParts(3).toInt,
+      workerIdParts(4).toInt,
+      mode,
+      null,
+      new StorageInfo(
+        StorageInfo.typesMap.get(pbPackedPartitionLocations.getTypes(index)),
+        pbPackedPartitionLocations.getMountPointsSet(
+          pbPackedPartitionLocations.getMountPoints(index)),
+        pbPackedPartitionLocations.getFinalResult(index),
+        filePath,
+        pbPackedPartitionLocations.getAvailableStorageTypes(index)),
+      Utils.byteStringToRoaringBitmap(pbPackedPartitionLocations.getMapIdBitMap(index)))
+  }
+
+  def fromPbPackedWorkerResource(pbWorkerResource: util.Map[String, PbPackedWorkerResource])
+      : WorkerResource = {
+    val slots = new WorkerResource()
+    pbWorkerResource.asScala.foreach { case (uniqueId, pbPackedWorkerResource) =>
+      val networkLocation = pbPackedWorkerResource.getNetworkLocation
+      val workerInfo = WorkerInfo.fromUniqueId(uniqueId)
+      workerInfo.networkLocation = networkLocation
+      val (primaryLocations, replicateLocations) =
+        fromPbPackedPartitionLocationsPair(pbPackedWorkerResource.getLocationPairs)
+      slots.put(workerInfo, (primaryLocations, replicateLocations))
+    }
+    slots
+  }
+
+  def toPbPackedWorkerResource(workerResource: WorkerResource)
+      : util.Map[String, PbPackedWorkerResource] = {
+    workerResource.asScala.map { case (workerInfo, (primaryLocations, replicaLocations)) =>
+      val pbWorkerResource = PbPackedWorkerResource.newBuilder()
+        .setLocationPairs(toPbPackedPartitionLocationsPair(
+          primaryLocations.asScala.toList ++ replicaLocations.asScala.toList))
+        .setNetworkLocation(workerInfo.networkLocation)
+        .build()
+      workerInfo.toUniqueId() -> pbWorkerResource
+    }.asJava
+  }
+
 }
