@@ -22,7 +22,6 @@ import java.io.File;
 import java.io.FileInputStream;
 import java.io.IOException;
 import java.nio.file.Files;
-import java.util.ArrayList;
 import java.util.List;
 import java.util.Map;
 import java.util.Optional;
@@ -35,6 +34,8 @@ import java.util.stream.Collectors;
 
 import scala.Option;
 
+import org.apache.hadoop.net.NetworkTopology;
+import org.apache.hadoop.net.Node;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
@@ -42,16 +43,21 @@ import org.apache.celeborn.common.CelebornConf;
 import org.apache.celeborn.common.identity.UserIdentifier;
 import org.apache.celeborn.common.meta.AppDiskUsageMetric;
 import org.apache.celeborn.common.meta.AppDiskUsageSnapShot;
+import org.apache.celeborn.common.meta.ApplicationMeta;
 import org.apache.celeborn.common.meta.DiskInfo;
 import org.apache.celeborn.common.meta.DiskStatus;
+import org.apache.celeborn.common.meta.WorkerEventInfo;
 import org.apache.celeborn.common.meta.WorkerInfo;
+import org.apache.celeborn.common.meta.WorkerStatus;
+import org.apache.celeborn.common.network.CelebornRackResolver;
 import org.apache.celeborn.common.protocol.PbSnapshotMetaInfo;
+import org.apache.celeborn.common.protocol.PbWorkerStatus;
 import org.apache.celeborn.common.quota.ResourceConsumption;
 import org.apache.celeborn.common.rpc.RpcEnv;
 import org.apache.celeborn.common.util.JavaUtils;
 import org.apache.celeborn.common.util.PbSerDeUtils;
 import org.apache.celeborn.common.util.Utils;
-import org.apache.celeborn.service.deploy.master.network.CelebornRackResolver;
+import org.apache.celeborn.common.util.WorkerStatusUtils;
 
 public abstract class AbstractMetaManager implements IMetadataHandler {
   private static final Logger LOG = LoggerFactory.getLogger(AbstractMetaManager.class);
@@ -59,8 +65,11 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
   // Metadata for master service
   public final Set<String> registeredShuffle = ConcurrentHashMap.newKeySet();
   public final Set<String> hostnameSet = ConcurrentHashMap.newKeySet();
-  public final ArrayList<WorkerInfo> workers = new ArrayList<>();
+  public final Set<WorkerInfo> workers = ConcurrentHashMap.newKeySet();
+
   public final ConcurrentHashMap<WorkerInfo, Long> lostWorkers = JavaUtils.newConcurrentHashMap();
+  public final ConcurrentHashMap<WorkerInfo, WorkerEventInfo> workerEventInfos =
+      JavaUtils.newConcurrentHashMap();
   public final ConcurrentHashMap<String, Long> appHeartbeatTime = JavaUtils.newConcurrentHashMap();
   public final Set<WorkerInfo> excludedWorkers = ConcurrentHashMap.newKeySet();
   public final Set<WorkerInfo> manuallyExcludedWorkers = ConcurrentHashMap.newKeySet();
@@ -77,6 +86,9 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
   public final LongAdder partitionTotalFileCount = new LongAdder();
   public AppDiskUsageMetric appDiskUsageMetric = null;
 
+  public final ConcurrentHashMap<String, ApplicationMeta> applicationMetas =
+      JavaUtils.newConcurrentHashMap();
+
   public void updateRequestSlotsMeta(
       String shuffleKey, String hostName, Map<String, Map<String, Integer>> workerWithAllocations) {
     registeredShuffle.add(shuffleKey);
@@ -110,6 +122,7 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
   public void updateAppLostMeta(String appId) {
     registeredShuffle.removeIf(shuffleKey -> shuffleKey.startsWith(appId));
     appHeartbeatTime.remove(appId);
+    applicationMetas.remove(appId);
   }
 
   public void updateWorkerExcludeMeta(
@@ -148,6 +161,7 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
         if (lostWorkers.containsKey(workerInfo)) {
           lostWorkers.remove(workerInfo);
           shutdownWorkers.remove(workerInfo);
+          workerEventInfos.remove(workerInfo);
         }
       }
     }
@@ -163,10 +177,11 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
       Map<UserIdentifier, ResourceConsumption> userResourceConsumption,
       Map<String, Long> estimatedAppDiskUsage,
       long time,
+      WorkerStatus workerStatus,
       boolean highWorkload) {
     WorkerInfo worker =
         new WorkerInfo(
-            host, rpcPort, pushPort, fetchPort, replicatePort, disks, userResourceConsumption);
+            host, rpcPort, pushPort, fetchPort, replicatePort, -1, disks, userResourceConsumption);
     AtomicLong availableSlots = new AtomicLong();
     LOG.debug("update worker {}:{} heartbeat {}", host, rpcPort, disks);
     synchronized (workers) {
@@ -177,8 +192,19 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
             info.updateThenGetUserResourceConsumption(userResourceConsumption);
             availableSlots.set(info.totalAvailableSlots());
             info.lastHeartbeat_$eq(time);
+            info.setWorkerStatus(workerStatus);
           });
     }
+
+    WorkerEventInfo workerEventInfo = workerEventInfos.get(worker);
+    if (workerEventInfo != null
+        && WorkerStatusUtils.meetFinalState(workerEventInfo, workerStatus)) {
+      workerEventInfos.remove(worker);
+      if (workerStatus.getState() == PbWorkerStatus.State.Normal) {
+        shutdownWorkers.remove(worker);
+      }
+    }
+
     appDiskUsageMetric.update(estimatedAppDiskUsage);
     // If using HDFSONLY mode, workers with empty disks should not be put into excluded worker list.
     long healthyDiskNum =
@@ -199,13 +225,28 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
       int pushPort,
       int fetchPort,
       int replicatePort,
+      int internalPort,
+      String networkLocation,
       Map<String, DiskInfo> disks,
       Map<UserIdentifier, ResourceConsumption> userResourceConsumption) {
     WorkerInfo workerInfo =
         new WorkerInfo(
-            host, rpcPort, pushPort, fetchPort, replicatePort, disks, userResourceConsumption);
+            host,
+            rpcPort,
+            pushPort,
+            fetchPort,
+            replicatePort,
+            internalPort,
+            disks,
+            userResourceConsumption);
     workerInfo.lastHeartbeat_$eq(System.currentTimeMillis());
-    workerInfo.networkLocation_$eq(rackResolver.resolve(host).getNetworkLocation());
+    if (networkLocation != null
+        && !networkLocation.isEmpty()
+        && !NetworkTopology.DEFAULT_RACK.equals(networkLocation)) {
+      workerInfo.networkLocation_$eq(networkLocation);
+    } else {
+      workerInfo.networkLocation_$eq(rackResolver.resolve(host).getNetworkLocation());
+    }
     workerInfo.updateDiskMaxSlots(estimatedPartitionSize);
     synchronized (workers) {
       if (!workers.contains(workerInfo)) {
@@ -214,6 +255,7 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
       shutdownWorkers.remove(workerInfo);
       lostWorkers.remove(workerInfo);
       excludedWorkers.remove(workerInfo);
+      workerEventInfos.remove(workerInfo);
     }
   }
 
@@ -239,7 +281,9 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
                 appDiskUsageMetric.snapShots(),
                 appDiskUsageMetric.currentSnapShot().get(),
                 lostWorkers,
-                shutdownWorkers)
+                shutdownWorkers,
+                workerEventInfos,
+                applicationMetas)
             .toByteArray();
     Files.write(file.toPath(), snapshotBytes);
   }
@@ -275,31 +319,47 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
 
       registeredShuffle.forEach(
           shuffleKey -> {
-            String appId = Utils.splitShuffleKey(shuffleKey)._1;
+            String appId = shuffleKey.split("-")[0];
             if (!appHeartbeatTime.containsKey(appId)) {
               appHeartbeatTime.put(appId, System.currentTimeMillis());
             }
           });
 
-      workers.addAll(
+      Set<WorkerInfo> workerInfoSet =
           snapshotMetaInfo.getWorkersList().stream()
               .map(PbSerDeUtils::fromPbWorkerInfo)
-              .collect(Collectors.toSet())
-              .stream()
-              .map(
+              .collect(Collectors.toSet());
+      List<String> workerHostList =
+          workerInfoSet.stream()
+              .filter(w -> NetworkTopology.DEFAULT_RACK.equals(w.networkLocation()))
+              .map(WorkerInfo::host)
+              .collect(Collectors.toList());
+      scala.collection.immutable.Map<String, Node> resolveMap =
+          rackResolver.resolveToMap(workerHostList);
+      workers.addAll(
+          workerInfoSet.stream()
+              .peek(
                   workerInfo -> {
                     // Reset worker's network location with current master's configuration.
-                    workerInfo.networkLocation_$eq(
-                        rackResolver.resolve(workerInfo.host()).getNetworkLocation());
-                    return workerInfo;
+                    if (NetworkTopology.DEFAULT_RACK.equals(workerInfo.networkLocation())) {
+                      workerInfo.networkLocation_$eq(
+                          resolveMap.get(workerInfo.host()).get().getNetworkLocation());
+                    }
                   })
               .collect(Collectors.toSet()));
 
       snapshotMetaInfo
           .getLostWorkersMap()
+          .forEach((key, value) -> lostWorkers.put(WorkerInfo.fromUniqueId(key), value));
+
+      snapshotMetaInfo
+          .getWorkerEventInfosMap()
           .entrySet()
           .forEach(
-              entry -> lostWorkers.put(WorkerInfo.fromUniqueId(entry.getKey()), entry.getValue()));
+              entry ->
+                  workerEventInfos.put(
+                      WorkerInfo.fromUniqueId(entry.getKey()),
+                      PbSerDeUtils.fromPbWorkerEventInfo(entry.getValue())));
 
       shutdownWorkers.addAll(
           snapshotMetaInfo.getShutdownWorkersList().stream()
@@ -316,6 +376,11 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
           new AtomicReference<AppDiskUsageSnapShot>(
               PbSerDeUtils.fromPbAppDiskUsageSnapshot(
                   snapshotMetaInfo.getCurrentAppDiskUsageMetricsSnapshot())));
+
+      snapshotMetaInfo
+          .getApplicationMetasMap()
+          .forEach(
+              (key, value) -> applicationMetas.put(key, PbSerDeUtils.fromPbApplicationMeta(value)));
     } catch (Exception e) {
       throw new IOException(e);
     }
@@ -342,6 +407,8 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
     workerLostEvents.clear();
     partitionTotalWritten.reset();
     partitionTotalFileCount.reset();
+    workerEventInfos.clear();
+    applicationMetas.clear();
   }
 
   public void updateMetaByReportWorkerUnavailable(List<WorkerInfo> failedWorkers) {
@@ -350,6 +417,25 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
     }
   }
 
+  public void updateWorkerEventMeta(int workerEventTypeValue, List<WorkerInfo> workerInfoList) {
+    long eventTime = System.currentTimeMillis();
+    ResourceProtos.WorkerEventType eventType =
+        ResourceProtos.WorkerEventType.forNumber(workerEventTypeValue);
+    synchronized (this.workers) {
+      for (WorkerInfo workerInfo : workerInfoList) {
+        WorkerEventInfo workerEventInfo = workerEventInfos.get(workerInfo);
+        LOG.info("Received worker event: {} for worker: {}", eventType, workerInfo.toUniqueId());
+        if (workerEventInfo == null || !workerEventInfo.isSameEvent(eventType.getNumber())) {
+          if (eventType == ResourceProtos.WorkerEventType.None) {
+            workerEventInfos.remove(workerInfo);
+          } else {
+            workerEventInfos.put(workerInfo, new WorkerEventInfo(eventType.getNumber(), eventTime));
+          }
+        }
+      }
+    }
+  }
+
   public void updatePartitionSize() {
     long oldEstimatedPartitionSize = estimatedPartitionSize;
     long tmpTotalWritten = partitionTotalWritten.sumThenReset();
@@ -376,4 +462,16 @@ public abstract class AbstractMetaManager implements IMetadataHandler {
                 !excludedWorkers.contains(worker) && !manuallyExcludedWorkers.contains(worker))
         .forEach(workerInfo -> workerInfo.updateDiskMaxSlots(estimatedPartitionSize));
   }
+
+  public boolean isWorkerAvailable(WorkerInfo workerInfo) {
+    return !excludedWorkers.contains(workerInfo)
+        && !shutdownWorkers.contains(workerInfo)
+        && !manuallyExcludedWorkers.contains(workerInfo)
+        && (!workerEventInfos.containsKey(workerInfo)
+            && workerInfo.getWorkerStatus().getState() == PbWorkerStatus.State.Normal);
+  }
+
+  public void updateApplicationMeta(ApplicationMeta applicationMeta) {
+    applicationMetas.putIfAbsent(applicationMeta.appId(), applicationMeta);
+  }
 }
