@@ -19,6 +19,7 @@ package org.apache.celeborn.service.deploy.master.clustermeta.ha;
 
 import java.io.File;
 import java.io.IOException;
+import java.net.InetAddress;
 import java.net.InetSocketAddress;
 import java.nio.charset.StandardCharsets;
 import java.util.*;
@@ -27,14 +28,19 @@ import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicLong;
 import java.util.concurrent.locks.ReentrantReadWriteLock;
 
+import javax.net.ssl.KeyManager;
+import javax.net.ssl.TrustManager;
+
 import scala.Tuple2;
 
 import com.google.common.annotations.VisibleForTesting;
 import com.google.protobuf.InvalidProtocolBufferException;
 import org.apache.ratis.RaftConfigKeys;
 import org.apache.ratis.client.RaftClientConfigKeys;
+import org.apache.ratis.conf.Parameters;
 import org.apache.ratis.conf.RaftProperties;
 import org.apache.ratis.grpc.GrpcConfigKeys;
+import org.apache.ratis.grpc.GrpcTlsConfig;
 import org.apache.ratis.netty.NettyConfigKeys;
 import org.apache.ratis.proto.RaftProtos;
 import org.apache.ratis.protocol.*;
@@ -44,6 +50,7 @@ import org.apache.ratis.rpc.RpcType;
 import org.apache.ratis.rpc.SupportedRpcType;
 import org.apache.ratis.server.RaftServer;
 import org.apache.ratis.server.RaftServerConfigKeys;
+import org.apache.ratis.server.storage.RaftStorage.StartupOption;
 import org.apache.ratis.thirdparty.com.google.protobuf.ByteString;
 import org.apache.ratis.util.LifeCycle;
 import org.apache.ratis.util.SizeInBytes;
@@ -54,6 +61,8 @@ import org.slf4j.LoggerFactory;
 import org.apache.celeborn.common.CelebornConf;
 import org.apache.celeborn.common.client.MasterClient;
 import org.apache.celeborn.common.exception.CelebornRuntimeException;
+import org.apache.celeborn.common.network.ssl.SSLFactory;
+import org.apache.celeborn.common.protocol.TransportModuleConstants;
 import org.apache.celeborn.common.util.ThreadUtils;
 import org.apache.celeborn.common.util.Utils;
 import org.apache.celeborn.service.deploy.master.clustermeta.ResourceProtos;
@@ -76,6 +85,7 @@ public class HARaftServer {
   private final MasterNode localNode;
   private final InetSocketAddress ratisAddr;
   private final String rpcEndpoint;
+  private final String internalRpcEndpoint;
   private final RaftServer server;
   private final RaftGroup raftGroup;
   private final RaftPeerId raftPeerId;
@@ -117,18 +127,26 @@ public class HARaftServer {
     this.localNode = localNode;
     this.ratisAddr = localNode.ratisAddr();
     this.rpcEndpoint = localNode.rpcEndpoint();
+    this.internalRpcEndpoint = localNode.internalRpcEndpoint();
     this.raftPeerId = localRaftPeerId;
     this.raftGroup = RaftGroup.valueOf(RAFT_GROUP_ID, raftPeers);
     this.masterStateMachine = getStateMachine();
     this.conf = conf;
-    RaftProperties serverProperties = newRaftProperties(conf);
+
+    final RpcType rpc = SupportedRpcType.valueOfIgnoreCase(conf.haMasterRatisRpcType());
+    RaftProperties serverProperties = newRaftProperties(conf, rpc);
+    Parameters sslParameters =
+        localNode.sslEnabled() ? configureSsl(conf, serverProperties, rpc) : null;
     setDeadlineTime(Integer.MAX_VALUE, Integer.MAX_VALUE); // for default
     this.server =
         RaftServer.newBuilder()
             .setServerId(this.raftPeerId)
             .setGroup(this.raftGroup)
             .setProperties(serverProperties)
+            .setParameters(sslParameters)
             .setStateMachine(masterStateMachine)
+            // RATIS-1677. Do not auto format RaftStorage in RECOVER.
+            .setOption(StartupOption.valueOf(conf.haMasterRatisStorageStartupOption()))
             .build();
 
     StringBuilder raftPeersStr = new StringBuilder();
@@ -166,6 +184,8 @@ public class HARaftServer {
             .setId(localRaftPeerId)
             .setAddress(ratisAddr)
             .setClientAddress(localNode.rpcEndpoint())
+            // We use admin address to host the internal rpc address
+            .setAdminAddress(localNode.internalRpcEndpoint())
             .build();
     List<RaftPeer> raftPeers = new ArrayList<>();
     // Add this Ratis server to the Ratis ring
@@ -181,6 +201,8 @@ public class HARaftServer {
                     .setId(raftPeerId)
                     .setAddress(peer.ratisEndpoint())
                     .setClientAddress(peer.rpcEndpoint())
+                    // We use admin address to host the internal rpc address
+                    .setAdminAddress(peer.internalRpcEndpoint())
                     .build();
           } else {
             InetSocketAddress peerRatisAddr = peer.ratisAddr();
@@ -189,6 +211,8 @@ public class HARaftServer {
                     .setId(raftPeerId)
                     .setAddress(peerRatisAddr)
                     .setClientAddress(peer.rpcEndpoint())
+                    // We use admin address to host the internal rpc address
+                    .setAdminAddress(peer.internalRpcEndpoint())
                     .build();
           }
 
@@ -262,11 +286,9 @@ public class HARaftServer {
     }
   }
 
-  private RaftProperties newRaftProperties(CelebornConf conf) {
+  private RaftProperties newRaftProperties(CelebornConf conf, RpcType rpc) {
     final RaftProperties properties = new RaftProperties();
     // Set RPC type
-    final String rpcType = conf.haMasterRatisRpcType();
-    final RpcType rpc = SupportedRpcType.valueOfIgnoreCase(rpcType);
     RaftConfigKeys.Rpc.setType(properties, rpc);
 
     // Set the ratis port number
@@ -287,8 +309,22 @@ public class HARaftServer {
 
     // Set RAFT segment pre-allocated size
     long raftSegmentPreallocatedSize = conf.haMasterRatisLogPreallocatedSize();
+    long raftSegmentWriteBufferSize = conf.haMasterRatisLogWriteBufferSize();
     int logAppenderQueueNumElements = conf.haMasterRatisLogAppenderQueueNumElements();
     long logAppenderQueueByteLimit = conf.haMasterRatisLogAppenderQueueBytesLimit();
+    // RATIS-589. Eliminate buffer copying in SegmentedRaftLogOutputStream.
+    // 4 bytes (serialized size) + logAppenderQueueByteLimit + 4 bytes (checksum)
+    if (raftSegmentWriteBufferSize < logAppenderQueueByteLimit + 8) {
+      throw new IllegalArgumentException(
+          CelebornConf.HA_MASTER_RATIS_LOG_WRITE_BUFFER_SIZE().key()
+              + " (= "
+              + raftSegmentWriteBufferSize
+              + ") is less than "
+              + CelebornConf.HA_MASTER_RATIS_LOG_APPENDER_QUEUE_BYTE_LIMIT().key()
+              + " + 8 (= "
+              + (logAppenderQueueByteLimit + 8)
+              + ")");
+    }
     boolean shouldInstallSnapshot = conf.haMasterRatisLogInstallSnapshotEnabled();
     RaftServerConfigKeys.Log.Appender.setBufferElementLimit(
         properties, logAppenderQueueNumElements);
@@ -296,6 +332,8 @@ public class HARaftServer {
         properties, SizeInBytes.valueOf(logAppenderQueueByteLimit));
     RaftServerConfigKeys.Log.setPreallocatedSize(
         properties, SizeInBytes.valueOf(raftSegmentPreallocatedSize));
+    RaftServerConfigKeys.Log.setWriteBufferSize(
+        properties, SizeInBytes.valueOf(raftSegmentWriteBufferSize));
     RaftServerConfigKeys.Log.Appender.setInstallSnapshotEnabled(properties, shouldInstallSnapshot);
     int logPurgeGap = conf.haMasterRatisLogPurgeGap();
     RaftServerConfigKeys.Log.setPurgeGap(properties, logPurgeGap);
@@ -367,6 +405,37 @@ public class HARaftServer {
     return properties;
   }
 
+  private Parameters configureSsl(CelebornConf conf, RaftProperties properties, RpcType rpc) {
+
+    if (rpc != SupportedRpcType.GRPC) {
+      LOG.error(
+          "SSL has been disabled for Raft communication between masters. "
+              + "This is only supported when ratis is configured with GRPC");
+      return null;
+    }
+
+    // This is used only for querying state after initialization - not actual SSL
+    // also why nThreads does not matter
+    SSLFactory factory =
+        SSLFactory.createSslFactory(
+            Utils.fromCelebornConf(conf, TransportModuleConstants.RPC_SERVICE_MODULE, 1));
+
+    assert (null != factory);
+    assert (factory.hasKeyManagers());
+    assert (!factory.getTrustManagers().isEmpty());
+
+    TrustManager trustManager = factory.getTrustManagers().get(0);
+    KeyManager keyManager = factory.getKeyManagers().get(0);
+
+    Parameters params = new Parameters();
+    GrpcConfigKeys.TLS.setEnabled(properties, true);
+    GrpcConfigKeys.TLS.setConf(params, new GrpcTlsConfig(keyManager, trustManager, true));
+
+    LOG.info("SSL enabled for ratis communication between masters");
+
+    return params;
+  }
+
   private StateMachine getStateMachine() {
     StateMachine stateMachine = new StateMachine(this);
     stateMachine.setRaftGroupId(RAFT_GROUP_ID);
@@ -444,10 +513,14 @@ public class HARaftServer {
       RaftProtos.RoleInfoProto roleInfoProto = groupInfo.getRoleInfoProto();
       RaftProtos.RaftPeerRole thisNodeRole = roleInfoProto.getRole();
       Tuple2<String, String> leaderPeerRpcEndpoint = null;
+      Tuple2<String, String> leaderPeerInternalRpcEndpoint = null;
       if (thisNodeRole.equals(RaftProtos.RaftPeerRole.LEADER)) {
         // Current Node always uses original rpcEndpoint/internalRpcEndpoint, as if something wrong
         // they would never return to client.
-        setServerRole(thisNodeRole, Tuple2.apply(this.rpcEndpoint, this.rpcEndpoint));
+        setServerRole(
+            thisNodeRole,
+            Tuple2.apply(this.rpcEndpoint, this.rpcEndpoint),
+            Tuple2.apply(this.internalRpcEndpoint, this.internalRpcEndpoint));
       } else if (thisNodeRole.equals(RaftProtos.RaftPeerRole.FOLLOWER)) {
         ByteString leaderNodeId = roleInfoProto.getFollowerInfo().getLeaderInfo().getId().getId();
         // There may be a chance, here we get leaderNodeId as null. For
@@ -457,12 +530,20 @@ public class HARaftServer {
           String clientAddress =
               roleInfoProto.getFollowerInfo().getLeaderInfo().getId().getClientAddress();
           leaderPeerRpcEndpoint = Utils.addressToIpHostAddressPair(clientAddress);
+          // We use admin address to host the internal rpc address
+          if (conf.internalPortEnabled()) {
+            String adminAddress =
+                roleInfoProto.getFollowerInfo().getLeaderInfo().getId().getAdminAddress();
+            leaderPeerInternalRpcEndpoint = Utils.addressToIpHostAddressPair(adminAddress);
+          } else {
+            leaderPeerInternalRpcEndpoint = leaderPeerRpcEndpoint;
+          }
         }
 
-        setServerRole(thisNodeRole, leaderPeerRpcEndpoint);
+        setServerRole(thisNodeRole, leaderPeerRpcEndpoint, leaderPeerInternalRpcEndpoint);
 
       } else {
-        setServerRole(thisNodeRole, null);
+        setServerRole(thisNodeRole, null, null);
       }
     } catch (IOException e) {
       LOG.error(
@@ -470,13 +551,15 @@ public class HARaftServer {
               + "{} and resetting leader info.",
           RaftProtos.RaftPeerRole.UNRECOGNIZED,
           e);
-      setServerRole(null, null);
+      setServerRole(null, null, null);
     }
   }
 
   /** Set the current server role and the leader peer rpc endpoint. */
   private void setServerRole(
-      RaftProtos.RaftPeerRole currentRole, Tuple2<String, String> leaderPeerRpcEndpoint) {
+      RaftProtos.RaftPeerRole currentRole,
+      Tuple2<String, String> leaderPeerRpcEndpoint,
+      Tuple2<String, String> leaderPeerInternalRpcEndpoint) {
     this.roleCheckLock.writeLock().lock();
     try {
       boolean leaderChanged = false;
@@ -498,7 +581,8 @@ public class HARaftServer {
       this.cachedPeerRole = Optional.ofNullable(currentRole);
       if (null != leaderPeerRpcEndpoint) {
         this.cachedLeaderPeerRpcEndpoints =
-            Optional.of(new LeaderPeerEndpoints(leaderPeerRpcEndpoint));
+            Optional.of(
+                new LeaderPeerEndpoints(leaderPeerRpcEndpoint, leaderPeerInternalRpcEndpoint));
       } else {
         this.cachedLeaderPeerRpcEndpoints = Optional.empty();
       }
@@ -513,6 +597,11 @@ public class HARaftServer {
     return server.getGroupInfo(groupInfoRequest);
   }
 
+  // Exposed for testing
+  public InetAddress getRaftAddress() {
+    return this.ratisAddr.getAddress();
+  }
+
   public int getRaftPort() {
     return this.ratisAddr.getPort();
   }
@@ -521,6 +610,10 @@ public class HARaftServer {
     return this.rpcEndpoint;
   }
 
+  public String getInternalRpcEndpoint() {
+    return this.internalRpcEndpoint;
+  }
+
   void stepDown() {
     try {
       TransferLeadershipRequest request =
@@ -564,8 +657,13 @@ public class HARaftServer {
     // the rpcEndpoints Tuple2 (ip:port, host:port)
     public final Tuple2<String, String> rpcEndpoints;
 
-    public LeaderPeerEndpoints(Tuple2<String, String> rpcEndpoints) {
+    // the rpcInternalEndpoints Tuple2 (ip:port, host:port)
+    public final Tuple2<String, String> rpcInternalEndpoints;
+
+    public LeaderPeerEndpoints(
+        Tuple2<String, String> rpcEndpoints, Tuple2<String, String> rpcInternalEndpoints) {
       this.rpcEndpoints = rpcEndpoints;
+      this.rpcInternalEndpoints = rpcInternalEndpoints;
     }
   }
 }
