@@ -20,7 +20,8 @@ package org.apache.celeborn.service.deploy.master
 import java.io.IOException
 import java.net.BindException
 import java.util
-import java.util.concurrent.{ConcurrentHashMap, ScheduledFuture, TimeUnit}
+import java.util.concurrent.{ConcurrentHashMap, ExecutorService, ScheduledFuture, TimeUnit}
+import java.util.concurrent.atomic.AtomicBoolean
 import java.util.function.ToLongFunction
 
 import scala.collection.JavaConverters._
@@ -32,20 +33,25 @@ import org.apache.ratis.proto.RaftProtos.RaftPeerRole
 
 import org.apache.celeborn.common.CelebornConf
 import org.apache.celeborn.common.client.MasterClient
+import org.apache.celeborn.common.exception.CelebornException
 import org.apache.celeborn.common.identity.UserIdentifier
 import org.apache.celeborn.common.internal.Logging
-import org.apache.celeborn.common.meta.{DiskInfo, WorkerInfo}
+import org.apache.celeborn.common.meta.{DiskInfo, WorkerInfo, WorkerStatus}
 import org.apache.celeborn.common.metrics.MetricsSystem
 import org.apache.celeborn.common.metrics.source.{JVMCPUSource, JVMSource, ResourceConsumptionSource, SystemMiscSource, ThreadPoolSource}
+import org.apache.celeborn.common.network.CelebornRackResolver
+import org.apache.celeborn.common.network.protocol.TransportMessage
 import org.apache.celeborn.common.protocol._
 import org.apache.celeborn.common.protocol.message.{ControlMessages, StatusCode}
 import org.apache.celeborn.common.protocol.message.ControlMessages._
-import org.apache.celeborn.common.quota.{QuotaManager, ResourceConsumption}
+import org.apache.celeborn.common.quota.ResourceConsumption
 import org.apache.celeborn.common.rpc._
-import org.apache.celeborn.common.util.{CelebornHadoopUtils, CollectionUtils, JavaUtils, PbSerDeUtils, ThreadUtils, Utils}
+import org.apache.celeborn.common.rpc.{RpcSecurityContextBuilder, ServerSaslContextBuilder}
+import org.apache.celeborn.common.util.{CelebornHadoopUtils, CollectionUtils, JavaUtils, PbSerDeUtils, SignalUtils, ThreadUtils, Utils}
 import org.apache.celeborn.server.common.{HttpService, Service}
 import org.apache.celeborn.service.deploy.master.clustermeta.SingleMasterMetaManager
 import org.apache.celeborn.service.deploy.master.clustermeta.ha.{HAHelper, HAMasterMetaManager, MetaHandler}
+import org.apache.celeborn.service.deploy.master.quota.QuotaManager
 
 private[celeborn] class Master(
     override val conf: CelebornConf,
@@ -71,18 +77,66 @@ private[celeborn] class Master(
   metricsSystem.registerSource(new SystemMiscSource(conf, MetricsSystem.ROLE_MASTER))
 
   private val bindPreferIP: Boolean = conf.bindPreferIP
+  private val authEnabled = conf.authEnabled
+  private val secretRegistry = new MasterSecretRegistryImpl()
+  private val sendApplicationMetaThreads = conf.masterSendApplicationMetaThreads
+  // Send ApplicationMeta to workers
+  private var sendApplicationMetaExecutor: ExecutorService = _
 
-  override val rpcEnv: RpcEnv = RpcEnv.create(
-    RpcNameConstants.MASTER_SYS,
-    masterArgs.host,
-    masterArgs.host,
-    masterArgs.port,
-    conf,
-    Math.max(64, Runtime.getRuntime.availableProcessors()))
+  if (conf.logCelebornConfEnabled) {
+    logInfo(getConf)
+  }
 
+  override val rpcEnv: RpcEnv =
+    if (!authEnabled) {
+      RpcEnv.create(
+        RpcNameConstants.MASTER_SYS,
+        TransportModuleConstants.RPC_SERVICE_MODULE,
+        masterArgs.host,
+        masterArgs.host,
+        masterArgs.port,
+        conf,
+        Math.max(64, Runtime.getRuntime.availableProcessors()))
+    } else {
+      val externalSecurityContext = new RpcSecurityContextBuilder()
+        .withServerSaslContext(
+          new ServerSaslContextBuilder()
+            .withAddRegistrationBootstrap(true)
+            .withSecretRegistry(secretRegistry).build()).build()
+      logInfo(
+        s"Secure port enabled ${masterArgs.port} for secured RPC.")
+      RpcEnv.create(
+        RpcNameConstants.MASTER_SYS,
+        TransportModuleConstants.RPC_SERVICE_MODULE,
+        masterArgs.host,
+        masterArgs.host,
+        masterArgs.port,
+        conf,
+        Math.max(64, Runtime.getRuntime.availableProcessors()),
+        Some(externalSecurityContext))
+    }
+
+  // Visible for testing
+  private[master] var internalRpcEnvInUse: RpcEnv =
+    if (!conf.internalPortEnabled) {
+      rpcEnv
+    } else {
+      logInfo(
+        s"Internal port enabled, using internal port ${masterArgs.internalPort} for internal RPC.")
+      RpcEnv.create(
+        RpcNameConstants.MASTER_INTERNAL_SYS,
+        TransportModuleConstants.RPC_SERVICE_MODULE,
+        masterArgs.host,
+        masterArgs.host,
+        masterArgs.internalPort,
+        conf,
+        Math.max(64, Runtime.getRuntime.availableProcessors()))
+    }
+
+  private val rackResolver = new CelebornRackResolver(conf)
   private val statusSystem =
     if (conf.haEnabled) {
-      val sys = new HAMasterMetaManager(rpcEnv, conf)
+      val sys = new HAMasterMetaManager(internalRpcEnvInUse, conf, rackResolver)
       val handler = new MetaHandler(sys)
       try {
         handler.setUpMasterRatisServer(conf, masterArgs.masterClusterInfo.get)
@@ -100,8 +154,9 @@ private[celeborn] class Master(
       }
       sys
     } else {
-      new SingleMasterMetaManager(rpcEnv, conf)
+      new SingleMasterMetaManager(internalRpcEnvInUse, conf, rackResolver)
     }
+  secretRegistry.setMetadataHandler(statusSystem)
 
   // Threads
   private val forwardMessageThread =
@@ -120,19 +175,11 @@ private[celeborn] class Master(
   private val hdfsExpireDirsTimeoutMS = conf.hdfsExpireDirsTimeoutMS
   private val hasHDFSStorage = conf.hasHDFSStorage
 
-  private val quotaManager = QuotaManager.instantiate(conf)
+  private val quotaManager = new QuotaManager(conf, configService)
   private val masterResourceConsumptionInterval = conf.masterResourceConsumptionInterval
   private val userResourceConsumptions =
     JavaUtils.newConcurrentHashMap[UserIdentifier, (ResourceConsumption, Long)]()
 
-  // States
-  private def workersSnapShot: util.List[WorkerInfo] =
-    statusSystem.workers.synchronized(new util.ArrayList[WorkerInfo](statusSystem.workers))
-  private def lostWorkersSnapshot: ConcurrentHashMap[WorkerInfo, java.lang.Long] =
-    statusSystem.workers.synchronized(JavaUtils.newConcurrentHashMap(statusSystem.lostWorkers))
-  private def shutdownWorkerSnapshot: util.List[WorkerInfo] =
-    statusSystem.workers.synchronized(new util.ArrayList[WorkerInfo](statusSystem.shutdownWorkers))
-
   private def diskReserveSize = conf.workerDiskReserveSize
   private def diskReserveRatio = conf.workerDiskReserveRatio
 
@@ -168,11 +215,14 @@ private[celeborn] class Master(
   masterSource.addGauge(MasterSource.REGISTERED_SHUFFLE_COUNT) { () =>
     statusSystem.registeredShuffle.size
   }
+  masterSource.addGauge(MasterSource.WORKER_COUNT) { () => statusSystem.workers.size }
+  masterSource.addGauge(MasterSource.LOST_WORKER_COUNT) { () => statusSystem.lostWorkers.size }
   masterSource.addGauge(MasterSource.EXCLUDED_WORKER_COUNT) { () =>
     statusSystem.excludedWorkers.size + statusSystem.manuallyExcludedWorkers.size
   }
-  masterSource.addGauge(MasterSource.WORKER_COUNT) { () => statusSystem.workers.size }
-  masterSource.addGauge(MasterSource.LOST_WORKER_COUNT) { () => statusSystem.lostWorkers.size }
+  masterSource.addGauge(MasterSource.SHUTDOWN_WORKER_COUNT) { () =>
+    statusSystem.shutdownWorkers.size
+  }
   masterSource.addGauge(MasterSource.RUNNING_APPLICATION_COUNT) { () =>
     statusSystem.appHeartbeatTime.size
   }
@@ -197,12 +247,44 @@ private[celeborn] class Master(
             }).sum()
       }).sum()
   }
+
+  masterSource.addGauge(MasterSource.DEVICE_CELEBORN_TOTAL_CAPACITY) { () =>
+    statusSystem.workers.asScala.map(_.totalSpace()).sum
+  }
+
+  masterSource.addGauge(MasterSource.DEVICE_CELEBORN_FREE_CAPACITY) { () =>
+    statusSystem.workers.asScala.map(_.totalActualUsableSpace()).sum
+  }
+
   masterSource.addGauge(MasterSource.IS_ACTIVE_MASTER) { () => isMasterActive }
 
+  private val threadsStarted: AtomicBoolean = new AtomicBoolean(false)
   rpcEnv.setupEndpoint(RpcNameConstants.MASTER_EP, this)
+  // Visible for testing
+  private[master] var internalRpcEndpoint: RpcEndpoint = _
+  private var internalRpcEndpointRef: RpcEndpointRef = _
+  if (conf.internalPortEnabled) {
+    internalRpcEndpoint = new InternalRpcEndpoint(this, internalRpcEnvInUse, conf)
+    internalRpcEndpointRef = internalRpcEnvInUse.setupEndpoint(
+      RpcNameConstants.MASTER_INTERNAL_EP,
+      internalRpcEndpoint)
+  }
+
+  // Maintains the mapping for the workers assigned to each application
+  private val workersAssignedToApp
+      : util.concurrent.ConcurrentHashMap[String, util.Set[WorkerInfo]] =
+    JavaUtils.newConcurrentHashMap[String, util.Set[WorkerInfo]]()
 
   // start threads to check timeout for workers and applications
   override def onStart(): Unit = {
+    if (!threadsStarted.compareAndSet(false, true)) {
+      return
+    }
+    if (authEnabled) {
+      sendApplicationMetaExecutor = ThreadUtils.newDaemonFixedThreadPool(
+        sendApplicationMetaThreads,
+        "send-application-meta")
+    }
     checkForWorkerTimeOutTask = forwardMessageThread.scheduleWithFixedDelay(
       new Runnable {
         override def run(): Unit = Utils.tryLogNonFatalError {
@@ -248,6 +330,9 @@ private[celeborn] class Master(
   }
 
   override def onStop(): Unit = {
+    if (!threadsStarted.compareAndSet(true, false)) {
+      return
+    }
     logInfo("Stopping Celeborn Master.")
     if (checkForWorkerTimeOutTask != null) {
       checkForWorkerTimeOutTask.cancel(true)
@@ -262,6 +347,10 @@ private[celeborn] class Master(
       checkForHDFSRemnantDirsTimeOutTask.cancel(true)
     }
     forwardMessageThread.shutdownNow()
+    rackResolver.stop()
+    if (authEnabled) {
+      sendApplicationMetaExecutor.shutdownNow()
+    }
     logInfo("Celeborn Master is stopped.")
   }
 
@@ -317,6 +406,7 @@ private[celeborn] class Master(
           requestId,
           shouldResponse) =>
       logDebug(s"Received heartbeat from app $appId")
+      checkAuth(context, appId)
       executeWithLeaderChecker(
         context,
         handleHeartbeatFromApplication(
@@ -335,6 +425,8 @@ private[celeborn] class Master(
       val pushPort = pbRegisterWorker.getPushPort
       val fetchPort = pbRegisterWorker.getFetchPort
       val replicatePort = pbRegisterWorker.getReplicatePort
+      val internalPort = pbRegisterWorker.getInternalPort
+      val networkLocation = pbRegisterWorker.getNetworkLocation
       val disks = pbRegisterWorker.getDisksList.asScala
         .map { pbDiskInfo => pbDiskInfo.getMountPoint -> PbSerDeUtils.fromPbDiskInfo(pbDiskInfo) }
         .toMap.asJava
@@ -352,6 +444,8 @@ private[celeborn] class Master(
           pushPort,
           fetchPort,
           replicatePort,
+          internalPort,
+          networkLocation,
           disks,
           userResourceConsumption,
           requestId))
@@ -360,8 +454,9 @@ private[celeborn] class Master(
       // keep it for compatible reason
       context.reply(ReleaseSlotsResponse(StatusCode.SUCCESS))
 
-    case requestSlots @ RequestSlots(_, _, _, _, _, _, _, _, _, _, _) =>
+    case requestSlots @ RequestSlots(applicationId, _, _, _, _, _, _, _, _, _, _, _) =>
       logTrace(s"Received RequestSlots request $requestSlots.")
+      checkAuth(context, applicationId)
       executeWithLeaderChecker(context, handleRequestSlots(context, requestSlots))
 
     case pb: PbUnregisterShuffle =>
@@ -369,6 +464,7 @@ private[celeborn] class Master(
       val shuffleId = pb.getShuffleId
       val requestId = pb.getRequestId
       logDebug(s"Received UnregisterShuffle request $requestId, $applicationId, $shuffleId")
+      checkAuth(context, applicationId)
       executeWithLeaderChecker(
         context,
         handleUnregisterShuffle(context, applicationId, shuffleId, requestId))
@@ -376,6 +472,7 @@ private[celeborn] class Master(
     case ApplicationLost(appId, requestId) =>
       logDebug(
         s"Received ApplicationLost request $requestId, $appId from ${context.senderAddress}.")
+      checkAuth(context, appId)
       executeWithLeaderChecker(context, handleApplicationLost(context, appId, requestId))
 
     case HeartbeatFromWorker(
@@ -389,6 +486,7 @@ private[celeborn] class Master(
           activeShuffleKey,
           estimatedAppDiskUsage,
           highWorkload,
+          workerStatus,
           requestId) =>
       logDebug(s"Received heartbeat from" +
         s" worker $host:$rpcPort:$pushPort:$fetchPort:$replicatePort with $disks.")
@@ -406,6 +504,7 @@ private[celeborn] class Master(
           activeShuffleKey,
           estimatedAppDiskUsage,
           highWorkload,
+          workerStatus,
           requestId))
 
     case ReportWorkerUnavailable(failedWorkers: util.List[WorkerInfo], requestId: String) =>
@@ -443,6 +542,21 @@ private[celeborn] class Master(
 
     case _: PbCheckWorkersAvailable =>
       executeWithLeaderChecker(context, handleCheckWorkersAvailable(context))
+
+    case pb: PbWorkerEventRequest =>
+      val workers = new util.ArrayList[WorkerInfo](pb.getWorkersList
+        .asScala.map(PbSerDeUtils.fromPbWorkerInfo).toList.asJava)
+      executeWithLeaderChecker(
+        context,
+        handleWorkerEvent(
+          pb.getRequestId,
+          pb.getWorkerEventType.getNumber,
+          workers,
+          context))
+
+    case pb: PbApplicationMetaRequest =>
+      // This request is from a worker
+      executeWithLeaderChecker(context, handleRequestForApplicationMeta(context, pb))
   }
 
   private def timeoutDeadWorkers(): Unit = {
@@ -452,7 +566,7 @@ private[celeborn] class Master(
       return
     }
 
-    workersSnapShot.asScala.foreach { worker =>
+    statusSystem.workers.asScala.foreach { worker =>
       if (worker.lastHeartbeat < currentTime - workerHeartbeatTimeoutMs
         && !statusSystem.workerLostEvents.contains(worker)) {
         logWarning(s"Worker ${worker.readableAddress()} timeout! Trigger WorkerLost event.")
@@ -475,7 +589,7 @@ private[celeborn] class Master(
       return
     }
 
-    val unavailableInfoTimeoutWorkers = lostWorkersSnapshot.asScala.filter {
+    val unavailableInfoTimeoutWorkers = statusSystem.lostWorkers.asScala.filter {
       case (_, lostTime) => currentTime - lostTime > workerUnavailableInfoExpireTimeoutMs
     }.keySet.toList.asJava
 
@@ -522,9 +636,10 @@ private[celeborn] class Master(
       activeShuffleKeys: util.Set[String],
       estimatedAppDiskUsage: util.HashMap[String, java.lang.Long],
       highWorkload: Boolean,
+      workerStatus: WorkerStatus,
       requestId: String): Unit = {
     val targetWorker = new WorkerInfo(host, rpcPort, pushPort, fetchPort, replicatePort)
-    val registered = workersSnapShot.asScala.contains(targetWorker)
+    val registered = statusSystem.workers.asScala.contains(targetWorker)
     if (!registered) {
       logWarning(s"Received heartbeat from unknown worker " +
         s"$host:$rpcPort:$pushPort:$fetchPort:$replicatePort.")
@@ -540,18 +655,30 @@ private[celeborn] class Master(
         estimatedAppDiskUsage,
         System.currentTimeMillis(),
         highWorkload,
+        workerStatus,
         requestId)
     }
 
     val expiredShuffleKeys = new util.HashSet[String]
     activeShuffleKeys.asScala.foreach { shuffleKey =>
       if (!statusSystem.registeredShuffle.contains(shuffleKey)) {
-        logWarning(
-          s"Shuffle $shuffleKey expired on $host:$rpcPort:$pushPort:$fetchPort:$replicatePort.")
         expiredShuffleKeys.add(shuffleKey)
       }
     }
-    context.reply(HeartbeatFromWorkerResponse(expiredShuffleKeys, registered))
+    logDebug(
+      s"Shuffle ${expiredShuffleKeys.asScala.mkString("[", " ,", "]")} expired on ${targetWorker.toUniqueId()}.")
+
+    val workerEventInfo = statusSystem.workerEventInfos.get(targetWorker)
+    if (workerEventInfo == null) {
+      context.reply(HeartbeatFromWorkerResponse(
+        expiredShuffleKeys,
+        registered))
+    } else {
+      context.reply(HeartbeatFromWorkerResponse(
+        expiredShuffleKeys,
+        registered,
+        workerEventInfo.getEventType))
+    }
   }
 
   private def handleWorkerExclude(
@@ -579,9 +706,10 @@ private[celeborn] class Master(
       pushPort,
       fetchPort,
       replicatePort,
+      -1,
       new util.HashMap[String, DiskInfo](),
       JavaUtils.newConcurrentHashMap[UserIdentifier, ResourceConsumption]())
-    val worker: WorkerInfo = workersSnapShot
+    val worker: WorkerInfo = statusSystem.workers
       .asScala
       .find(_ == targetWorker)
       .orNull
@@ -603,6 +731,8 @@ private[celeborn] class Master(
       pushPort: Int,
       fetchPort: Int,
       replicatePort: Int,
+      internalPort: Int,
+      networkLocation: String,
       disks: util.Map[String, DiskInfo],
       userResourceConsumption: util.Map[UserIdentifier, ResourceConsumption],
       requestId: String): Unit = {
@@ -613,9 +743,10 @@ private[celeborn] class Master(
         pushPort,
         fetchPort,
         replicatePort,
+        internalPort,
         disks,
         userResourceConsumption)
-    if (workersSnapShot.contains(workerToRegister)) {
+    if (statusSystem.workers.contains(workerToRegister)) {
       logWarning(s"Receive RegisterWorker while worker" +
         s" ${workerToRegister.toString()} already exists, re-register.")
       // TODO: remove `WorkerRemove` because we have improve register logic to cover `WorkerRemove`
@@ -627,6 +758,8 @@ private[celeborn] class Master(
         pushPort,
         fetchPort,
         replicatePort,
+        internalPort,
+        networkLocation,
         disks,
         userResourceConsumption,
         newRequestId)
@@ -641,6 +774,8 @@ private[celeborn] class Master(
         pushPort,
         fetchPort,
         replicatePort,
+        internalPort,
+        networkLocation,
         disks,
         userResourceConsumption,
         requestId)
@@ -652,6 +787,8 @@ private[celeborn] class Master(
         pushPort,
         fetchPort,
         replicatePort,
+        internalPort,
+        networkLocation,
         disks,
         userResourceConsumption,
         requestId)
@@ -669,7 +806,8 @@ private[celeborn] class Master(
 
     if (numAvailableWorkers == 0) {
       logError(s"Offer slots for $shuffleKey failed due to all workers are excluded!")
-      context.reply(RequestSlotsResponse(StatusCode.WORKER_EXCLUDED, new WorkerResource()))
+      context.reply(
+        RequestSlotsResponse(StatusCode.WORKER_EXCLUDED, new WorkerResource(), requestSlots.packed))
     }
 
     val numWorkers = Math.min(
@@ -726,7 +864,10 @@ private[celeborn] class Master(
     // reply false if offer slots failed
     if (slots == null || slots.isEmpty) {
       logError(s"Offer slots for $numReducers reducers of $shuffleKey failed!")
-      context.reply(RequestSlotsResponse(StatusCode.SLOT_NOT_AVAILABLE, new WorkerResource()))
+      context.reply(RequestSlotsResponse(
+        StatusCode.SLOT_NOT_AVAILABLE,
+        new WorkerResource(),
+        requestSlots.packed))
       return
     }
 
@@ -754,7 +895,47 @@ private[celeborn] class Master(
       logInfo(s"Offered extra $offerSlotsExtraSize slots for $shuffleKey")
     }
 
-    context.reply(RequestSlotsResponse(StatusCode.SUCCESS, slots.asInstanceOf[WorkerResource]))
+    if (authEnabled) {
+      pushApplicationMetaToWorkers(requestSlots, slots)
+    }
+    context.reply(RequestSlotsResponse(
+      StatusCode.SUCCESS,
+      slots.asInstanceOf[WorkerResource],
+      requestSlots.packed))
+  }
+
+  def pushApplicationMetaToWorkers(
+      requestSlots: RequestSlots,
+      slots: util.Map[WorkerInfo, (util.List[PartitionLocation], util.List[PartitionLocation])])
+      : Unit = {
+    // Pass application registration information to the workers
+    val pbApplicationMeta = PbApplicationMeta.newBuilder()
+      .setAppId(requestSlots.applicationId)
+      .setSecret(secretRegistry.getSecretKey(requestSlots.applicationId))
+      .build()
+    val transportMessage =
+      new TransportMessage(MessageType.APPLICATION_META, pbApplicationMeta.toByteArray)
+    val workerSet = workersAssignedToApp.computeIfAbsent(
+      requestSlots.applicationId,
+      new util.function.Function[String, util.Set[WorkerInfo]] {
+        override def apply(key: String): util.Set[WorkerInfo] =
+          util.Collections.newSetFromMap(JavaUtils.newConcurrentHashMap[
+            WorkerInfo,
+            java.lang.Boolean]())
+      })
+    slots.keySet().asScala.foreach { worker =>
+      // The app meta info is send to a Worker only if it wasn't previously sent.
+      if (workerSet.add(worker)) {
+        sendApplicationMetaExecutor.submit(new Runnable {
+          override def run(): Unit = {
+            logDebug(s"Sending app registration info to ${worker.host}:${worker.internalPort}")
+            internalRpcEnvInUse.setupEndpointRef(
+              RpcAddress.apply(worker.host, worker.internalPort),
+              RpcNameConstants.WORKER_INTERNAL_EP).send(transportMessage)
+          }
+        })
+      }
+    }
   }
 
   def handleUnregisterShuffle(
@@ -781,6 +962,7 @@ private[celeborn] class Master(
   def handleApplicationLost(context: RpcCallContext, appId: String, requestId: String): Unit = {
     nonEagerHandler.submit(new Runnable {
       override def run(): Unit = {
+        workersAssignedToApp.remove(appId)
         statusSystem.handleAppLost(appId, requestId)
         logInfo(s"Removed application $appId")
         if (hasHDFSStorage) {
@@ -834,14 +1016,16 @@ private[celeborn] class Master(
       System.currentTimeMillis(),
       requestId)
     // unknown workers will retain in needCheckedWorkerList
-    needCheckedWorkerList.removeAll(workersSnapShot)
+    needCheckedWorkerList.removeAll(statusSystem.workers)
     if (shouldResponse) {
+      // UserResourceConsumption and DiskInfo are eliminated from WorkerInfo
+      // during serialization of HeartbeatFromApplicationResponse
       context.reply(HeartbeatFromApplicationResponse(
         StatusCode.SUCCESS,
         new util.ArrayList(
           (statusSystem.excludedWorkers.asScala ++ statusSystem.manuallyExcludedWorkers.asScala).asJava),
         needCheckedWorkerList,
-        shutdownWorkerSnapshot))
+        new util.ArrayList[WorkerInfo](statusSystem.shutdownWorkers)))
     } else {
       context.reply(OneWayMessageResponse)
     }
@@ -915,26 +1099,57 @@ private[celeborn] class Master(
     resourceConsumption
   }
 
-  private def handleCheckQuota(
+  private[master] def handleCheckQuota(
       userIdentifier: UserIdentifier,
       context: RpcCallContext): Unit = {
     val userResourceConsumption = handleResourceConsumption(userIdentifier)
-    val quota = quotaManager.getQuota(userIdentifier)
-    val (isAvailable, reason) =
-      quota.checkQuotaSpaceAvailable(userIdentifier, userResourceConsumption)
-    context.reply(CheckQuotaResponse(isAvailable, reason))
+    if (conf.quotaEnabled) {
+      val (isAvailable, reason) =
+        quotaManager.checkQuotaSpaceAvailable(userIdentifier, userResourceConsumption)
+      context.reply(CheckQuotaResponse(isAvailable, reason))
+    } else {
+      context.reply(CheckQuotaResponse(true, ""))
+    }
   }
 
   private def handleCheckWorkersAvailable(context: RpcCallContext): Unit = {
     context.reply(CheckWorkersAvailableResponse(!workersAvailable().isEmpty))
   }
 
+  private def handleWorkerEvent(
+      requestId: String,
+      workerEventTypeValue: Int,
+      workers: util.List[WorkerInfo],
+      context: RpcCallContext): Unit = {
+    statusSystem.handleWorkerEvent(workerEventTypeValue, workers, requestId)
+    context.reply(PbWorkerEventResponse.newBuilder().setSuccess(true).build())
+  }
+
   private def workersAvailable(
       tmpExcludedWorkerList: Set[WorkerInfo] = Set.empty): util.List[WorkerInfo] = {
-    workersSnapShot.asScala.filter { w =>
-      !statusSystem.excludedWorkers.contains(w) && !statusSystem.manuallyExcludedWorkers.contains(
-        w) && !statusSystem.shutdownWorkers.contains(w) && !tmpExcludedWorkerList.contains(w)
-    }.asJava
+    statusSystem.workers.asScala.filter { w =>
+      statusSystem.isWorkerAvailable(w) && !tmpExcludedWorkerList.contains(w)
+    }.toList.asJava
+  }
+
+  private def handleRequestForApplicationMeta(
+      context: RpcCallContext,
+      pb: PbApplicationMetaRequest): Unit = {
+    val appId = pb.getAppId
+    logDebug(s"Handling request for application meta info $appId.")
+    val secret = secretRegistry.getSecretKey(appId)
+    if (secret == null) {
+      logWarning(s"Could not find the application meta of $appId.")
+      context.sendFailure(new CelebornException(s"$appId is not registered."))
+    } else {
+      val pbApplicationMeta = PbApplicationMeta.newBuilder()
+        .setAppId(appId)
+        .setSecret(secret)
+        .build()
+      val transportMessage =
+        new TransportMessage(MessageType.APPLICATION_META, pbApplicationMeta.toByteArray)
+      context.reply(transportMessage)
+    }
   }
 
   override def getMasterGroupInfo: String = {
@@ -945,7 +1160,36 @@ private[celeborn] class Master(
   }
 
   private def getWorkers: String = {
-    workersSnapShot.asScala.mkString("\n")
+    statusSystem.workers.asScala.mkString("\n")
+  }
+
+  override def handleWorkerEvent(workerEventType: String, workers: String): String = {
+    val sb = new StringBuilder
+    if (workerEventType.isEmpty || workers.isEmpty) {
+      return sb.append(
+        s"handle eventType failed as eventType: $workerEventType or workers: $workers has empty value").toString()
+    }
+
+    sb.append("============================ Handle Worker Event =============================\n")
+    val workerArray = workers.split(",").filter(_.nonEmpty)
+    try {
+      val workerEventResponse = self.askSync[PbWorkerEventResponse](WorkerEventRequest(
+        workerArray.map(WorkerInfo.fromUniqueId).toList.asJava,
+        workerEventType,
+        MasterClient.genRequestId()))
+      if (workerEventResponse.getSuccess) {
+        sb.append(s"handle $workerEventType for ${workerArray.mkString(",")} successfully")
+      } else {
+        sb.append(s"handle $workerEventType for ${workerArray.mkString(",")} failed")
+      }
+    } catch {
+      case e: Throwable =>
+        val message =
+          s"handle $workerEventType for ${workerArray.mkString(",")} failed, message: ${e.getMessage}"
+        logError(message, e)
+        sb.append(message)
+    }
+    sb.append("\n").toString()
   }
 
   override def getWorkerInfo: String = {
@@ -958,7 +1202,7 @@ private[celeborn] class Master(
   override def getLostWorkers: String = {
     val sb = new StringBuilder
     sb.append("======================= Lost Workers in Master ========================\n")
-    lostWorkersSnapshot.asScala.toSeq.sortBy(_._2).foreach { case (worker, time) =>
+    statusSystem.lostWorkers.asScala.toSeq.sortBy(_._2).foreach { case (worker, time) =>
       sb.append(s"${worker.toUniqueId().padTo(50, " ").mkString}${Utils.formatTimestamp(time)}\n")
     }
     sb.toString()
@@ -967,7 +1211,7 @@ private[celeborn] class Master(
   override def getShutdownWorkers: String = {
     val sb = new StringBuilder
     sb.append("===================== Shutdown Workers in Master ======================\n")
-    shutdownWorkerSnapshot.asScala.foreach { worker =>
+    statusSystem.shutdownWorkers.asScala.foreach { worker =>
       sb.append(s"${worker.toUniqueId()}\n")
     }
     sb.toString()
@@ -982,13 +1226,6 @@ private[celeborn] class Master(
     sb.toString()
   }
 
-  override def getThreadDump: String = {
-    val sb = new StringBuilder
-    sb.append("========================= Master ThreadDump ==========================\n")
-    sb.append(Utils.getThreadDump()).append("\n")
-    sb.toString()
-  }
-
   override def getHostnameList: String = {
     val sb = new StringBuilder
     sb.append("================= LifecycleManager Hostname List ======================\n")
@@ -1026,12 +1263,11 @@ private[celeborn] class Master(
   override def exclude(addWorkers: String, removeWorkers: String): String = {
     val sb = new StringBuilder
     sb.append("============================ Add/Remove Excluded Workers  Manually =============================\n")
-    val workersToAdd = addWorkers.split(",").filter(_.nonEmpty).map(WorkerInfo.fromUniqueId).toList
-    val workersToRemove =
-      removeWorkers.split(",").filter(_.nonEmpty).map(WorkerInfo.fromUniqueId).toList
+    val workersToAdd = addWorkers.split(",").filter(_.nonEmpty)
+    val workersToRemove = removeWorkers.split(",").filter(_.nonEmpty)
     val workerExcludeResponse = self.askSync[PbWorkerExcludeResponse](WorkerExclude(
-      workersToAdd.asJava,
-      workersToRemove.asJava,
+      workersToAdd.map(WorkerInfo.fromUniqueId).toList.asJava,
+      workersToRemove.map(WorkerInfo.fromUniqueId).toList.asJava,
       MasterClient.genRequestId()))
     if (workerExcludeResponse.getSuccess) {
       sb.append(
@@ -1041,7 +1277,7 @@ private[celeborn] class Master(
         s"Failed to Exclude workers add ${workersToAdd.mkString(",")} and remove ${workersToRemove.mkString(",")}.\n")
     }
     val unknownExcludedWorkers =
-      (workersToAdd ++ workersToRemove).filter(!workersSnapShot.contains(_))
+      (workersToAdd ++ workersToRemove).filter(!statusSystem.workers.contains(_))
     if (unknownExcludedWorkers.nonEmpty) {
       sb.append(
         s"Unknown worker ${unknownExcludedWorkers.mkString(",")}. Workers in Master:\n$getWorkers.")
@@ -1098,16 +1334,31 @@ private[celeborn] class Master(
     }
   }
 
+  override def getWorkerEventInfo(): String = {
+    val sb = new StringBuilder
+    sb.append("======================= Workers Event in Master ========================\n")
+    statusSystem.workerEventInfos.asScala.foreach { case (worker, workerEventInfo) =>
+      sb.append(s"${worker.toUniqueId().padTo(50, " ").mkString}$workerEventInfo\n")
+    }
+    sb.toString()
+  }
+
   override def initialize(): Unit = {
     super.initialize()
     logInfo("Master started.")
     rpcEnv.awaitTermination()
+    if (conf.internalPortEnabled) {
+      internalRpcEnvInUse.awaitTermination()
+    }
   }
 
   override def stop(exitKind: Int): Unit = synchronized {
     if (!stopped) {
       logInfo("Stopping Master")
       rpcEnv.stop(self)
+      if (conf.internalPortEnabled) {
+        internalRpcEnvInUse.stop(internalRpcEndpointRef)
+      }
       super.stop(exitKind)
       logInfo("Master stopped.")
       stopped = true
@@ -1117,6 +1368,7 @@ private[celeborn] class Master(
 
 private[deploy] object Master extends Logging {
   def main(args: Array[String]): Unit = {
+    SignalUtils.registerLogger(log)
     val conf = new CelebornConf()
     val masterArgs = new MasterArguments(args, conf)
     try {
