@@ -38,6 +38,7 @@ object Dependencies {
   val lz4JavaVersion = sparkClientProjects.map(_.lz4JavaVersion).getOrElse("1.8.0")
 
   // Dependent library versions
+  val apLoaderVersion = "3.0-8"
   val commonsCompressVersion = "1.4.1"
   val commonsCryptoVersion = "1.0.0"
   val commonsIoVersion = "2.13.0"
@@ -46,18 +47,17 @@ object Dependencies {
   val findbugsVersion = "1.3.9"
   val guavaVersion = "33.1.0-jre"
   val hadoopVersion = "3.3.6"
-  val javaxServletVersion = "3.1.0"
   val junitInterfaceVersion = "0.13.3"
   // don't forget update `junitInterfaceVersion` when we upgrade junit
   val junitVersion = "4.13.2"
   val leveldbJniVersion = "1.8"
   val log4j2Version = "2.17.2"
   val jdkToolsVersion = "0.1"
-  val metricsVersion = "3.2.6"
+  val metricsVersion = "4.2.25"
   val mockitoVersion = "4.11.0"
-  val nettyVersion = "4.1.101.Final"
-  val ratisVersion = "2.5.1"
-  val roaringBitmapVersion = "0.9.32"
+  val nettyVersion = "4.1.109.Final"
+  val ratisVersion = "3.0.1"
+  val roaringBitmapVersion = "1.0.6"
   val rocksdbJniVersion = "8.11.3"
   val jacksonVersion = "2.15.3"
   val scalatestMockitoVersion = "1.17.14"
@@ -65,11 +65,23 @@ object Dependencies {
   val slf4jVersion = "1.7.36"
   val snakeyamlVersion = "2.2"
   val snappyVersion = "1.1.10.5"
+  val mybatisVersion = "3.5.15"
+  val hikaricpVersion = "4.0.3"
+  val h2Version = "2.2.224"
+  val swaggerVersion = "2.2.1"
+  val swaggerUiVersion = "4.9.1"
+  val jerseyVersion = "2.39.1"
+  val jettyVersion = "9.4.52.v20230823"
+  val jakartaServeletApiVersion = "4.0.4"
+
+  // For SSL support
+  val bouncycastleVersion = "1.77"
 
   // Versions for proto
   val protocVersion = "3.21.7"
   val protoVersion = "3.21.7"
 
+  val apLoader = "me.bechberger" % "ap-loader-all" % apLoaderVersion
   val commonsCompress = "org.apache.commons" % "commons-compress" % commonsCompressVersion
   val commonsCrypto = "org.apache.commons" % "commons-crypto" % commonsCryptoVersion excludeAll(
     ExclusionRule("net.java.dev.jna", "jna"))
@@ -95,11 +107,11 @@ object Dependencies {
     ExclusionRule("log4j", "log4j"),
     ExclusionRule("org.slf4j", "slf4j-log4j12"))
   val ioDropwizardMetricsCore = "io.dropwizard.metrics" % "metrics-core" % metricsVersion
-  val ioDropwizardMetricsGraphite = "io.dropwizard.metrics" % "metrics-graphite" % metricsVersion
+  val ioDropwizardMetricsGraphite = "io.dropwizard.metrics" % "metrics-graphite" % metricsVersion excludeAll (
+    ExclusionRule("com.rabbitmq", "amqp-client"))
   val ioDropwizardMetricsJvm = "io.dropwizard.metrics" % "metrics-jvm" % metricsVersion
   val ioNetty = "io.netty" % "netty-all" % nettyVersion excludeAll(
     ExclusionRule("io.netty", "netty-handler-ssl-ocsp"))
-  val javaxServletApi = "javax.servlet" % "javax.servlet-api" % javaxServletVersion
   val leveldbJniGroup = if (System.getProperty("os.name").startsWith("Linux")
     && System.getProperty("os.arch").equals("aarch64")) {
     // use org.openlabtesting.leveldbjni on aarch64 platform except MacOS
@@ -116,6 +128,7 @@ object Dependencies {
   val ratisClient = "org.apache.ratis" % "ratis-client" % ratisVersion
   val ratisCommon = "org.apache.ratis" % "ratis-common" % ratisVersion
   val ratisGrpc = "org.apache.ratis" % "ratis-grpc" % ratisVersion
+  val ratisMetricsDefault = "org.apache.ratis" % "ratis-metrics-default" % ratisVersion
   val ratisNetty = "org.apache.ratis" % "ratis-netty" % ratisVersion
   val ratisServer = "org.apache.ratis" % "ratis-server" % ratisVersion
   val ratisShell = "org.apache.ratis" % "ratis-shell" % ratisVersion excludeAll(
@@ -133,6 +146,25 @@ object Dependencies {
   val snakeyaml = "org.yaml" % "snakeyaml" % snakeyamlVersion
   val snappyJava = "org.xerial.snappy" % "snappy-java" % snappyVersion
   val zstdJni = "com.github.luben" % "zstd-jni" % zstdJniVersion
+  val mybatis = "org.mybatis" % "mybatis" % mybatisVersion
+  val hikaricp = "com.zaxxer" % "HikariCP" % hikaricpVersion
+  val jettyServer = "org.eclipse.jetty" % "jetty-server" % jettyVersion excludeAll(
+    ExclusionRule("javax.servlet", "javax.servlet-api"))
+  val jettyServlet = "org.eclipse.jetty" % "jetty-servlet" % jettyVersion excludeAll(
+    ExclusionRule("javax.servlet", "javax.servlet-api"))
+  val jettyProxy = "org.eclipse.jetty" % "jetty-proxy" % jettyVersion
+  val jakartaServletApi = "jakarta.servlet" % "jakarta.servlet-api" % jakartaServeletApiVersion
+  val jerseyServer = "org.glassfish.jersey.core" % "jersey-server" % jerseyVersion excludeAll(
+    ExclusionRule("jakarta.xml.bind", "jakarta.xml.bind-api"))
+  val jerseyContainerServletCore = "org.glassfish.jersey.containers" % "jersey-container-servlet-core" % jerseyVersion
+  val jerseyHk2 = "org.glassfish.jersey.inject" % "jersey-hk2" % jerseyVersion
+  val jerseyMediaJsonJackson = "org.glassfish.jersey.media" % "jersey-media-json-jackson" % jerseyVersion
+  val jerseyMediaMultipart = "org.glassfish.jersey.media" % "jersey-media-multipart" % jerseyVersion
+  val swaggerJaxrs2 = "io.swagger.core.v3" % "swagger-jaxrs2" %swaggerVersion excludeAll(
+    ExclusionRule("com.sun.activation", "jakarta.activation"),
+    ExclusionRule("org.javassist", "javassist"),
+    ExclusionRule("jakarta.activation", "jakarta.activation-api"))
+  val swaggerUi = "org.webjars" % "swagger-ui" % swaggerUiVersion
 
   // Test dependencies
   // https://www.scala-sbt.org/1.x/docs/Testing.html
@@ -142,6 +174,15 @@ object Dependencies {
   val mockitoInline = "org.mockito" % "mockito-inline" % mockitoVersion
   val scalatestMockito = "org.mockito" %% "mockito-scala-scalatest" % scalatestMockitoVersion
   val scalatest = "org.scalatest" %% "scalatest" % scalatestVersion
+  val h2 = "com.h2database" % "h2" % h2Version
+  val jerseyTestFrameworkCore = "org.glassfish.jersey.test-framework" % "jersey-test-framework-core" % jerseyVersion
+  val jerseyTestFrameworkProviderJetty = "org.glassfish.jersey.test-framework.providers" % "jersey-test-framework-provider-jetty" % jerseyVersion excludeAll(
+    ExclusionRule("org.eclipse.jetty", "jetty-util"),
+    ExclusionRule("org.eclipse.jetty", "jetty-continuation"))
+
+  // SSL support
+  val bouncycastleBcprovJdk18on = "org.bouncycastle" % "bcprov-jdk18on" % bouncycastleVersion % "test"
+  val bouncycastleBcpkixJdk18on = "org.bouncycastle" % "bcpkix-jdk18on" % bouncycastleVersion % "test"
 }
 
 object CelebornCommonSettings {
@@ -155,7 +196,7 @@ object CelebornCommonSettings {
   val scala213 = "2.13.5"
   val ALL_SCALA_VERSIONS = Seq(SCALA_2_11_12, SCALA_2_12_10, SCALA_2_12_15, SCALA_2_12_17, SCALA_2_12_18, scala213)
 
-  val DEFAULT_SCALA_VERSION = SCALA_2_12_15
+  val DEFAULT_SCALA_VERSION = SCALA_2_12_18
 
   val projectScalaVersion = defaultScalaVersion()
 
@@ -178,7 +219,6 @@ object CelebornCommonSettings {
     Compile / compile / javacOptions ++= Seq("-target", "1.8"),
 
     dependencyOverrides := Seq(
-      Dependencies.commonsCompress,
       Dependencies.commonsLogging,
       Dependencies.findbugsJsr305,
       Dependencies.slf4jApi),
@@ -260,12 +300,12 @@ object CelebornCommonSettings {
       }
     },
     licenses += ("Apache-2.0", url("https://www.apache.org/licenses/LICENSE-2.0")),
-    scmInfo := Some(
-      ScmInfo(
-        url("https://celeborn.apache.org/"),
-          "scm:git:https://github.com/apache/celeborn.git",
-          "scm:git:git@github.com:apache/celeborn.git"))
-
+      pomExtra :=
+        <url>https://celeborn.apache.org/</url>
+        <scm>
+          <url>git@github.com:apache/celeborn.git</url>
+          <connection>scm:git:git@github.com:apache/celeborn.git</connection>
+        </scm>
   )
 
   lazy val protoSettings = Seq(
@@ -273,7 +313,7 @@ object CelebornCommonSettings {
     PB.protocVersion := Dependencies.protocVersion,
     // set proto sources path
     Compile / PB.protoSources := Seq(sourceDirectory.value / "main" / "proto"),
-    Compile / PB.targets := Seq(PB.gens.java(Dependencies.protocVersion) -> (Compile / sourceManaged).value)
+    Compile / PB.targets := Seq(PB.gens.java -> (Compile / sourceManaged).value)
   )
 
   lazy val commonUnitTestDependencies = Seq(
@@ -341,6 +381,7 @@ object Utils {
     case Some("flink-1.15") => Some(Flink115)
     case Some("flink-1.17") => Some(Flink117)
     case Some("flink-1.18") => Some(Flink118)
+    case Some("flink-1.19") => Some(Flink119)
     case _ => None
   }
 
@@ -398,8 +439,6 @@ object CelebornCommon {
         Dependencies.hadoopClientApi,
         Dependencies.hadoopClientRuntime,
         Dependencies.jdkTools,
-        Dependencies.ratisClient,
-        Dependencies.ratisCommon,
         Dependencies.leveldbJniAll,
         Dependencies.roaringBitmap,
         Dependencies.scalaReflect,
@@ -413,7 +452,10 @@ object CelebornCommon {
         Dependencies.jacksonDatabind,
         Dependencies.jacksonAnnotations,
         Dependencies.log4jSlf4jImpl % "test",
-        Dependencies.log4j12Api % "test"
+        Dependencies.log4j12Api % "test",
+        // SSL support
+        Dependencies.bouncycastleBcprovJdk18on,
+        Dependencies.bouncycastleBcpkixJdk18on
       ) ++ commonUnitTestDependencies,
 
       Compile / sourceGenerators += Def.task {
@@ -467,18 +509,35 @@ object CelebornService {
         Dependencies.findbugsJsr305,
         Dependencies.commonsIo,
         Dependencies.ioNetty,
-        Dependencies.javaxServletApi,
         Dependencies.commonsCrypto,
         Dependencies.slf4jApi,
+        Dependencies.mybatis,
+        Dependencies.hikaricp,
+        Dependencies.swaggerJaxrs2,
+        Dependencies.swaggerUi,
+        Dependencies.jakartaServletApi,
+        Dependencies.jerseyServer,
+        Dependencies.jerseyContainerServletCore,
+        Dependencies.jerseyHk2,
+        Dependencies.jerseyMediaJsonJackson,
+        Dependencies.jerseyMediaMultipart,
+        Dependencies.jettyServer,
+        Dependencies.jettyServlet,
+        Dependencies.jettyProxy,
         Dependencies.log4jSlf4jImpl % "test",
-        Dependencies.log4j12Api % "test"
+        Dependencies.log4j12Api % "test",
+        Dependencies.h2 % "test",
+        Dependencies.jerseyTestFrameworkCore % "test",
+        Dependencies.jerseyTestFrameworkProviderJetty % "test"
       ) ++ commonUnitTestDependencies
     )
 }
 
 object CelebornMaster {
   lazy val master = Project("celeborn-master", file("master"))
-    .dependsOn(CelebornCommon.common, CelebornService.service)
+    .dependsOn(CelebornCommon.common)
+    .dependsOn(CelebornCommon.common % "test->test;compile->compile")
+    .dependsOn(CelebornService.service % "test->test;compile->compile")
     .settings (
       commonSettings,
       protoSettings,
@@ -492,6 +551,7 @@ object CelebornMaster {
         Dependencies.ratisClient,
         Dependencies.ratisCommon,
         Dependencies.ratisGrpc,
+        Dependencies.ratisMetricsDefault,
         Dependencies.ratisNetty,
         Dependencies.ratisServer,
         Dependencies.ratisShell
@@ -503,16 +563,13 @@ object CelebornWorker {
   lazy val worker = Project("celeborn-worker", file("worker"))
     .dependsOn(CelebornService.service)
     .dependsOn(CelebornCommon.common % "test->test;compile->compile")
+    .dependsOn(CelebornService.service % "test->test;compile->compile")
     .dependsOn(CelebornClient.client % "test->compile")
     .dependsOn(CelebornMaster.master % "test->compile")
     .settings (
       commonSettings,
-      excludeDependencies ++= Seq(
-        // ratis-common/ratis-client are the transitive dependencies from celeborn-common
-        ExclusionRule("org.apache.ratis", "ratis-common"),
-        ExclusionRule("org.apache.ratis", "ratis-client")
-      ),
       libraryDependencies ++= Seq(
+        Dependencies.apLoader,
         Dependencies.guava,
         Dependencies.commonsIo,
         Dependencies.ioNetty,
@@ -521,7 +578,9 @@ object CelebornWorker {
         Dependencies.leveldbJniAll,
         Dependencies.roaringBitmap,
         Dependencies.rocksdbJni,
-        Dependencies.scalatestMockito % "test"
+        Dependencies.scalatestMockito % "test",
+        Dependencies.jerseyTestFrameworkCore % "test",
+        Dependencies.jerseyTestFrameworkProviderJetty % "test"
       ) ++ commonUnitTestDependencies
     )
 }
@@ -608,7 +667,7 @@ object Spark33 extends SparkClientProjects {
   val sparkProjectScalaVersion = "2.12.15"
   // scalaBinaryVersion
   // val scalaBinaryVersion = "2.12"
-  val sparkVersion = "3.3.3"
+  val sparkVersion = "3.3.4"
   val zstdJniVersion = "1.5.2-1"
 
   override val includeColumnarShuffle: Boolean = true
@@ -624,7 +683,7 @@ object Spark34 extends SparkClientProjects {
   val lz4JavaVersion = "1.8.0"
   val sparkProjectScalaVersion = "2.12.17"
 
-  val sparkVersion = "3.4.2"
+  val sparkVersion = "3.4.3"
   val zstdJniVersion = "1.5.2-5"
 
   override val includeColumnarShuffle: Boolean = true
@@ -640,7 +699,7 @@ object Spark35 extends SparkClientProjects {
   val lz4JavaVersion = "1.8.0"
   val sparkProjectScalaVersion = "2.12.18"
 
-  val sparkVersion = "3.5.0"
+  val sparkVersion = "3.5.1"
   val zstdJniVersion = "1.5.5-4"
 }
 
@@ -729,7 +788,9 @@ trait SparkClientProjects {
         libraryDependencies ++= Seq(
           "org.apache.spark" %% "spark-core" % sparkVersion % "test",
           "org.apache.spark" %% "spark-sql" % sparkVersion % "test",
-          "org.apache.spark" %% "spark-core" % sparkVersion % "test" classifier "tests",
+          "org.apache.spark" %% "spark-core" % sparkVersion % "test" classifier "tests" excludeAll(
+            ExclusionRule("org.glassfish.jersey.inject", "*"),
+            ExclusionRule("org.glassfish.jersey.core", "*")),
           "org.apache.spark" %% "spark-sql" % sparkVersion % "test" classifier "tests"
         ) ++ commonUnitTestDependencies
       )
@@ -840,7 +901,7 @@ object Flink115 extends FlinkClientProjects {
 }
 
 object Flink117 extends FlinkClientProjects {
-  val flinkVersion = "1.17.0"
+  val flinkVersion = "1.17.2"
 
   // note that SBT does not allow using the period symbol (.) in project names.
   val flinkClientProjectPath = "client-flink/flink-1.17"
@@ -850,7 +911,7 @@ object Flink117 extends FlinkClientProjects {
 }
 
 object Flink118 extends FlinkClientProjects {
-  val flinkVersion = "1.18.0"
+  val flinkVersion = "1.18.1"
 
   // note that SBT does not allow using the period symbol (.) in project names.
   val flinkClientProjectPath = "client-flink/flink-1.18"
@@ -859,6 +920,16 @@ object Flink118 extends FlinkClientProjects {
   val flinkClientShadedProjectName: String = "celeborn-client-flink-1_18-shaded"
 }
 
+object Flink119 extends FlinkClientProjects {
+  val flinkVersion = "1.19.0"
+
+  // note that SBT does not allow using the period symbol (.) in project names.
+  val flinkClientProjectPath = "client-flink/flink-1.19"
+  val flinkClientProjectName = "celeborn-client-flink-1_19"
+  val flinkClientShadedProjectPath: String = "client-flink/flink-1.19-shaded"
+  val flinkClientShadedProjectName: String = "celeborn-client-flink-1_19-shaded"
+}
+
 trait FlinkClientProjects {
 
   val flinkVersion: String
@@ -880,7 +951,9 @@ trait FlinkClientProjects {
     .aggregate(flinkCommon, flinkClient, flinkIt)
 
   // get flink major version. e.g:
-  //   1.17.0 -> 1.17
+  //   1.19.0 -> 1.19
+  //   1.18.1 -> 1.18
+  //   1.17.2 -> 1.17
   //   1.15.4 -> 1.15
   //   1.14.6 -> 1.14
   lazy val flinkMajorVersion: String = flinkVersion.split("\\.").take(2).reduce(_ + "." + _)
@@ -1025,7 +1098,8 @@ object MRClientProjects {
           Dependencies.hadoopClientApi,
           Dependencies.hadoopClientRuntime,
           Dependencies.hadoopMapreduceClientApp
-        ) ++ commonUnitTestDependencies
+        ) ++ commonUnitTestDependencies,
+        dependencyOverrides += Dependencies.commonsCompress
       )
   }
 
