@@ -17,16 +17,23 @@
 
 package org.apache.celeborn.server.common.service.config;
 
+import java.io.IOException;
+
 import org.apache.celeborn.common.CelebornConf;
 
 public class DynamicConfigServiceFactory {
 
-  public static ConfigService getConfigService(CelebornConf celebornConf) {
-    String configStoreBackend = celebornConf.dynamicConfigStoreBackend();
-    if ("FS".equals(configStoreBackend)) {
-      return new FsConfigServiceImpl(celebornConf);
+  public static ConfigService getConfigService(CelebornConf celebornConf) throws IOException {
+    if (celebornConf.dynamicConfigStoreBackend().isEmpty()) {
+      return null;
+    } else {
+      String configStoreBackend = celebornConf.dynamicConfigStoreBackend().get();
+      // celeborn.dynamicConfig.store.backend checks value with FS, DB.
+      if ("FS".equals(configStoreBackend)) {
+        return new FsConfigServiceImpl(celebornConf);
+      } else {
+        return new DbConfigServiceImpl(celebornConf);
+      }
     }
-
-    return null;
   }
 }
