@@ -40,6 +40,8 @@ import java.util.concurrent.locks.ReentrantLock;
 import com.google.common.annotations.VisibleForTesting;
 import com.google.common.cache.Cache;
 import com.google.common.cache.CacheBuilder;
+import io.netty.buffer.ByteBuf;
+import io.netty.buffer.CompositeByteBuf;
 import org.apache.commons.io.IOUtils;
 import org.apache.hadoop.fs.FSDataInputStream;
 import org.apache.hadoop.fs.FSDataOutputStream;
@@ -49,7 +51,7 @@ import org.slf4j.LoggerFactory;
 
 import org.apache.celeborn.common.CelebornConf;
 import org.apache.celeborn.common.identity.UserIdentifier;
-import org.apache.celeborn.common.meta.FileInfo;
+import org.apache.celeborn.common.meta.*;
 import org.apache.celeborn.common.metrics.source.AbstractSource;
 import org.apache.celeborn.common.unsafe.Platform;
 import org.apache.celeborn.common.util.*;
@@ -85,29 +87,30 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
   protected final long sortTimeout;
   protected final long shuffleChunkSize;
   protected final double compactionFactor;
+  protected final boolean prefetchEnabled;
   protected final long reservedMemoryPerPartition;
-  private boolean gracefulShutdown;
-  private long partitionSorterShutdownAwaitTime;
+  private final long partitionSorterShutdownAwaitTime;
   private DB sortedFilesDb;
 
   protected final AbstractSource source;
 
   private final ExecutorService fileSorterExecutors;
   private final ExecutorService fileSorterSchedulerThread;
-  private final long indexCacheMaxWeight;
 
   public PartitionFilesSorter(
       MemoryManager memoryManager, CelebornConf conf, AbstractSource source) {
-    this.sortTimeout = conf.partitionSorterSortPartitionTimeout();
+    this.sortTimeout = conf.workerPartitionSorterSortPartitionTimeout();
     this.shuffleChunkSize = conf.shuffleChunkSize();
-    this.reservedMemoryPerPartition = conf.partitionSorterReservedMemoryPerPartition();
     this.compactionFactor = conf.workerPartitionSorterShuffleBlockCompactionFactor();
+    this.prefetchEnabled = conf.workerPartitionSorterPrefetchEnabled();
+    this.reservedMemoryPerPartition = conf.workerPartitionSorterReservedMemoryPerPartition();
     this.partitionSorterShutdownAwaitTime =
         conf.workerGracefulShutdownPartitionSorterCloseAwaitTimeMs();
-    this.indexCacheMaxWeight = conf.partitionSorterIndexCacheMaxWeight();
+    long indexCacheMaxWeight = conf.workerPartitionSorterIndexCacheMaxWeight();
     this.source = source;
     this.cleaner = new PartitionFilesCleaner(this);
-    this.gracefulShutdown = conf.workerGracefulShutdown();
+    boolean gracefulShutdown = conf.workerGracefulShutdown();
+    // Assume a chunk won't be larger than 2GB
     // ShuffleClient can fetch shuffle data from a restarted worker only
     // when the worker's fetching port is stable and enables graceful shutdown.
     if (gracefulShutdown) {
@@ -129,12 +132,12 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
 
     fileSorterExecutors =
         ThreadUtils.newDaemonCachedThreadPool(
-            "worker-file-sorter-executor", conf.partitionSorterThreads(), 120);
+            "worker-file-sorter-executor", conf.workerPartitionSorterThreads(), 120);
 
     indexCache =
         CacheBuilder.newBuilder()
-            .concurrencyLevel(conf.partitionSorterThreads())
-            .expireAfterAccess(conf.partitionSorterIndexExpire(), TimeUnit.MILLISECONDS)
+            .concurrencyLevel(conf.workerPartitionSorterThreads())
+            .expireAfterAccess(conf.workerPartitionSorterIndexExpire(), TimeUnit.MILLISECONDS)
             .maximumWeight(indexCacheMaxWeight)
             .weigher(
                 (key, cache) ->
@@ -149,18 +152,20 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
           try {
             while (!shutdown) {
               FileSorter task = shuffleSortTaskDeque.take();
-              memoryManager.reserveSortMemory(reservedMemoryPerPartition);
-              while (!memoryManager.sortMemoryReady()) {
-                Thread.sleep(20);
+              if (task.isPrefetch) {
+                memoryManager.reserveSortMemory(reservedMemoryPerPartition);
+                while (!memoryManager.sortMemoryReady()) {
+                  Thread.sleep(20);
+                }
               }
               fileSorterExecutors.submit(
                   () -> {
                     try {
                       task.sort();
-                    } catch (InterruptedException e) {
-                      logger.warn("File sorter thread was interrupted.");
                     } finally {
-                      memoryManager.releaseSortMemory(reservedMemoryPerPartition);
+                      if (task.isPrefetch) {
+                        memoryManager.releaseSortMemory(reservedMemoryPerPartition);
+                      }
                     }
                   });
             }
@@ -188,77 +193,166 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
   // 2. If the FileSorter task is already in the sorting queue but the sorted file has not been
   //    generated, it awaits until a timeout occurs (default 220 seconds).
   // 3. If the sorted file is generated, it returns the sorted FileInfo.
+  // This method will generate temporary file info for this shuffle read
   public FileInfo getSortedFileInfo(
       String shuffleKey, String fileName, FileInfo fileInfo, int startMapIndex, int endMapIndex)
       throws IOException {
-    String fileId = shuffleKey + "-" + fileName;
-    UserIdentifier userIdentifier = fileInfo.getUserIdentifier();
-    Set<String> sorted =
-        sortedShuffleFiles.computeIfAbsent(shuffleKey, v -> ConcurrentHashMap.newKeySet());
-    Set<String> sorting =
-        sortingShuffleFiles.computeIfAbsent(shuffleKey, v -> ConcurrentHashMap.newKeySet());
+    if (fileInfo instanceof MemoryFileInfo) {
+      MemoryFileInfo memoryFileInfo = ((MemoryFileInfo) fileInfo);
+      Map<Integer, List<ShuffleBlockInfo>> indexesMap;
+      sortMemoryShuffleFile(memoryFileInfo);
+      indexesMap = memoryFileInfo.getSortedIndexes();
 
-    String sortedFilePath = Utils.getSortedFilePath(fileInfo.getFilePath());
-    String indexFilePath = Utils.getIndexFilePath(fileInfo.getFilePath());
-    synchronized (sorting) {
-      if (sorted.contains(fileId)) {
-        return resolve(
-            shuffleKey,
-            fileId,
-            userIdentifier,
-            sortedFilePath,
-            indexFilePath,
-            startMapIndex,
-            endMapIndex);
-      }
-      if (!sorting.contains(fileId)) {
-        try {
-          FileSorter fileSorter = new FileSorter(fileInfo, fileId, shuffleKey);
-          sorting.add(fileId);
-          logger.debug(
-              "Adding sorter to sort queue shuffle key {}, file name {}", shuffleKey, fileName);
-          shuffleSortTaskDeque.put(fileSorter);
-        } catch (InterruptedException e) {
-          logger.error("Sorter scheduler thread is interrupted means worker is shutting down.", e);
-          throw new IOException(
-              "Sort scheduler thread is interrupted means worker is shutting down.", e);
-        } catch (IOException e) {
-          logger.error("File sorter access HDFS failed.", e);
-          throw new IOException("File sorter access HDFS failed.", e);
-        }
-      }
-    }
+      ReduceFileMeta reduceFileMeta =
+          new ReduceFileMeta(
+              ShuffleBlockInfoUtils.getChunkOffsetsFromShuffleBlockInfos(
+                  startMapIndex, endMapIndex, shuffleChunkSize, indexesMap, true),
+              shuffleChunkSize);
+      CompositeByteBuf targetBuffer =
+          MemoryManager.instance()
+              .getStoragePooledByteBufAllocator()
+              .compositeBuffer(Integer.MAX_VALUE);
+      ShuffleBlockInfoUtils.sliceSortedBufferByMapRange(
+          startMapIndex, endMapIndex, indexesMap, memoryFileInfo.getSortedBuffer(), targetBuffer);
+      return new MemoryFileInfo(
+          memoryFileInfo.getUserIdentifier(),
+          memoryFileInfo.isPartitionSplitEnabled(),
+          reduceFileMeta,
+          targetBuffer);
+    } else {
+      DiskFileInfo diskFileInfo = ((DiskFileInfo) fileInfo);
+      String fileId = shuffleKey + "-" + fileName;
+      UserIdentifier userIdentifier = diskFileInfo.getUserIdentifier();
+      Set<String> sorted =
+          sortedShuffleFiles.computeIfAbsent(shuffleKey, v -> ConcurrentHashMap.newKeySet());
+      Set<String> sorting =
+          sortingShuffleFiles.computeIfAbsent(shuffleKey, v -> ConcurrentHashMap.newKeySet());
 
-    long sortStartTime = System.currentTimeMillis();
-    while (!sorted.contains(fileId)) {
-      if (sorting.contains(fileId)) {
-        try {
-          Thread.sleep(50);
-          if (System.currentTimeMillis() - sortStartTime > sortTimeout) {
-            logger.error("Sorting file {} timeout after {}ms", fileId, sortTimeout);
+      String sortedFilePath = Utils.getSortedFilePath(diskFileInfo.getFilePath());
+      String indexFilePath = Utils.getIndexFilePath(diskFileInfo.getFilePath());
+      boolean fileSorting = true;
+      synchronized (sorting) {
+        if (sorted.contains(fileId)) {
+          fileSorting = false;
+        } else if (!sorting.contains(fileId)) {
+          try {
+            FileSorter fileSorter = new FileSorter(diskFileInfo, fileId, shuffleKey);
+            sorting.add(fileId);
+            logger.debug(
+                "Adding sorter to sort queue shuffle key {}, file name {}", shuffleKey, fileName);
+            shuffleSortTaskDeque.put(fileSorter);
+          } catch (InterruptedException e) {
+            logger.error(
+                "Sorter scheduler thread is interrupted means worker is shutting down.", e);
             throw new IOException(
-                "Sort file " + fileInfo.getFilePath() + " timeout after " + sortTimeout);
+                "Sort scheduler thread is interrupted means worker is shutting down.", e);
+          } catch (IOException e) {
+            logger.error("File sorter access HDFS failed.", e);
+            throw new IOException("File sorter access HDFS failed.", e);
           }
-        } catch (InterruptedException e) {
-          logger.error("Sorter scheduler thread is interrupted means worker is shutting down.", e);
-          throw new IOException(
-              "Sorter scheduler thread is interrupted means worker is shutting down.", e);
         }
-      } else {
-        logger.debug("Sorting shuffle file for {} {} failed.", shuffleKey, fileInfo.getFilePath());
-        throw new IOException(
-            "Sorting shuffle file for " + shuffleKey + " " + fileInfo.getFilePath() + " failed.");
+      }
+
+      if (fileSorting) {
+        long sortStartTime = System.currentTimeMillis();
+        while (!sorted.contains(fileId)) {
+          if (sorting.contains(fileId)) {
+            try {
+              Thread.sleep(50);
+              if (System.currentTimeMillis() - sortStartTime > sortTimeout) {
+                logger.error("Sorting file {} timeout after {}ms", fileId, sortTimeout);
+                throw new IOException(
+                    "Sort file " + diskFileInfo.getFilePath() + " timeout after " + sortTimeout);
+              }
+            } catch (InterruptedException e) {
+              logger.error(
+                  "Sorter scheduler thread is interrupted means worker is shutting down.", e);
+              throw new IOException(
+                  "Sorter scheduler thread is interrupted means worker is shutting down.", e);
+            }
+          } else {
+            logger.debug(
+                "Sorting shuffle file for {} {} failed.", shuffleKey, diskFileInfo.getFilePath());
+            throw new IOException(
+                "Sorting shuffle file for "
+                    + shuffleKey
+                    + " "
+                    + diskFileInfo.getFilePath()
+                    + " failed.");
+          }
+        }
+      }
+
+      return resolve(
+          shuffleKey,
+          fileId,
+          userIdentifier,
+          sortedFilePath,
+          indexFilePath,
+          startMapIndex,
+          endMapIndex);
+    }
+  }
+
+  public static void sortMemoryShuffleFile(MemoryFileInfo memoryFileInfo) {
+    ReduceFileMeta reduceFileMeta = ((ReduceFileMeta) memoryFileInfo.getFileMeta());
+    synchronized (reduceFileMeta.getSorted()) {
+      if (!reduceFileMeta.getSorted().get()) {
+        CompositeByteBuf originBuffer = memoryFileInfo.getBuffer();
+        Map<Integer, List<ShuffleBlockInfo>> blocksMap = new TreeMap<>();
+        int originReaderIndex = originBuffer.readerIndex();
+        int originWriterIndex = originBuffer.writerIndex();
+        int bufLength = originBuffer.readableBytes();
+        int index = 0;
+        ByteBuffer headerBuf = ByteBuffer.allocate(16);
+
+        while (index != bufLength) {
+          headerBuf.rewind();
+          originBuffer.readerIndex(index);
+          originBuffer.readBytes(headerBuf);
+          byte[] batchHeader = headerBuf.array();
+          int mapId = Platform.getInt(batchHeader, Platform.BYTE_ARRAY_OFFSET);
+          int compressedSize = Platform.getInt(batchHeader, Platform.BYTE_ARRAY_OFFSET + 12);
+          ShuffleBlockInfo shuffleBlockInfo = new ShuffleBlockInfo();
+          shuffleBlockInfo.offset = index;
+          shuffleBlockInfo.length = 16 + compressedSize;
+          List<ShuffleBlockInfo> singleMapIdShuffleBlockList =
+              blocksMap.computeIfAbsent(mapId, v -> new ArrayList<>());
+          singleMapIdShuffleBlockList.add(shuffleBlockInfo);
+          index += 16 + compressedSize;
+        }
+        originBuffer.setIndex(originReaderIndex, originWriterIndex);
+
+        // sorted buffer should not consolidate
+        // because this will affect origin buffer's reference count
+        CompositeByteBuf sortedBuffer =
+            MemoryManager.instance()
+                .getStoragePooledByteBufAllocator()
+                .compositeBuffer(Integer.MAX_VALUE - 1);
+        Map<Integer, List<ShuffleBlockInfo>> sortedBlocks = new TreeMap<>();
+        int sortedBufferIndex = 0;
+        for (Map.Entry<Integer, List<ShuffleBlockInfo>> entry : blocksMap.entrySet()) {
+          int mapId = entry.getKey();
+          List<ShuffleBlockInfo> blockInfos = entry.getValue();
+          List<ShuffleBlockInfo> sortedMapBlocks = new ArrayList<>(blockInfos.size());
+          sortedBlocks.put(mapId, sortedMapBlocks);
+          for (ShuffleBlockInfo blockInfo : blockInfos) {
+            int offset = (int) blockInfo.offset;
+            int length = (int) blockInfo.length;
+            ByteBuf slice = originBuffer.slice(offset, length);
+            sortedBuffer.addComponent(true, slice);
+            ShuffleBlockInfo shuffleBlockInfo = new ShuffleBlockInfo();
+            shuffleBlockInfo.offset = sortedBufferIndex;
+            shuffleBlockInfo.length = blockInfo.length;
+            sortedBufferIndex += (int) blockInfo.length;
+            sortedMapBlocks.add(shuffleBlockInfo);
+          }
+        }
+        memoryFileInfo.setSortedBuffer(sortedBuffer);
+        memoryFileInfo.setSortedIndexes(sortedBlocks);
+        reduceFileMeta.setSorted();
       }
     }
-
-    return resolve(
-        shuffleKey,
-        fileId,
-        userIdentifier,
-        sortedFilePath,
-        indexFilePath,
-        startMapIndex,
-        endMapIndex);
   }
 
   public void cleanup(HashSet<String> expiredShuffleKeys) {
@@ -298,7 +392,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
         }
       }
       long end = System.currentTimeMillis();
-      logger.info("Await partition sorter executor complete cost " + (end - start) + "ms");
+      logger.info("Await partition sorter executor complete cost {}ms", end - start);
     } else {
       fileSorterSchedulerThread.shutdownNow();
       fileSorterExecutors.shutdownNow();
@@ -352,8 +446,8 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
   }
 
   @VisibleForTesting
-  public Set<String> initSortedShuffleFiles(String shuffleKey) {
-    return sortedShuffleFiles.computeIfAbsent(shuffleKey, v -> ConcurrentHashMap.newKeySet());
+  public void initSortedShuffleFiles(String shuffleKey) {
+    sortedShuffleFiles.computeIfAbsent(shuffleKey, v -> ConcurrentHashMap.newKeySet());
   }
 
   @VisibleForTesting
@@ -473,7 +567,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
     return transferredSize;
   }
 
-  public FileInfo resolve(
+  public DiskFileInfo resolve(
       String shuffleKey,
       String fileId,
       UserIdentifier userIdentifier,
@@ -491,7 +585,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
                 FileChannel indexChannel = null;
                 FSDataInputStream hdfsIndexStream = null;
                 boolean isHdfs = Utils.isHdfsPath(indexFilePath);
-                int indexSize = 0;
+                int indexSize;
                 try {
                   if (isHdfs) {
                     hdfsIndexStream = StorageManager.hadoopFs().open(new Path(indexFilePath));
@@ -530,11 +624,12 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
     } catch (ExecutionException e) {
       throw new IOException("Read sorted shuffle file index failed.", e);
     }
-    return new FileInfo(
-        sortedFilePath,
-        ShuffleBlockInfoUtils.getChunkOffsetsFromShuffleBlockInfos(
-            startMapIndex, endMapIndex, shuffleChunkSize, indexMap),
-        userIdentifier);
+    ReduceFileMeta reduceFileMeta =
+        new ReduceFileMeta(
+            ShuffleBlockInfoUtils.getChunkOffsetsFromShuffleBlockInfos(
+                startMapIndex, endMapIndex, shuffleChunkSize, indexMap, false),
+            shuffleChunkSize);
+    return new DiskFileInfo(userIdentifier, reduceFileMeta, sortedFilePath);
   }
 
   class FileSorter {
@@ -545,6 +640,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
     private final String fileId;
     private final String shuffleKey;
     private final boolean isHdfs;
+    private final boolean isPrefetch;
     private final FileInfo originFileInfo;
 
     private FSDataInputStream hdfsOriginInput = null;
@@ -552,11 +648,12 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
     private FileChannel originFileChannel = null;
     private FileChannel sortedFileChannel = null;
 
-    FileSorter(FileInfo fileInfo, String fileId, String shuffleKey) throws IOException {
+    FileSorter(DiskFileInfo fileInfo, String fileId, String shuffleKey) throws IOException {
       this.originFileInfo = fileInfo;
       this.originFilePath = fileInfo.getFilePath();
       this.sortedFilePath = Utils.getSortedFilePath(originFilePath);
       this.isHdfs = fileInfo.isHdfs();
+      this.isPrefetch = !isHdfs && prefetchEnabled;
       this.originFileLen = fileInfo.getFileLength();
       this.fileId = fileId;
       this.shuffleKey = shuffleKey;
@@ -580,7 +677,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
       }
     }
 
-    public void sort() throws InterruptedException {
+    public void sort() {
       source.startTimer(WorkerSource.SORT_TIME(), fileId);
 
       try {
@@ -590,9 +687,9 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
         Map<Integer, List<ShuffleBlockInfo>> sortedBlockInfoMap = new HashMap<>();
 
         int batchHeaderLen = 16;
-        int reserveMemory = (int) reservedMemoryPerPartition;
         ByteBuffer headerBuf = ByteBuffer.allocate(batchHeaderLen);
-        ByteBuffer paddingBuf = ByteBuffer.allocateDirect(reserveMemory);
+        ByteBuffer paddingBuf =
+            isPrefetch ? ByteBuffer.allocateDirect((int) reservedMemoryPerPartition) : null;
 
         long index = 0;
         while (index != originFileLen) {
@@ -612,7 +709,6 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
           singleMapIdShuffleBlockList.add(blockInfo);
 
           index += batchHeaderLen + compressedSize;
-          paddingBuf.clear();
           readBufferBySize(paddingBuf, compressedSize);
         }
 
@@ -649,7 +745,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
 
         writeIndex(sortedBlockInfoMap, indexFilePath, isHdfs);
         updateSortedShuffleFiles(shuffleKey, fileId, originFileLen);
-        originFileInfo.setSorted();
+        originFileInfo.getReduceFileMeta().setSorted();
         cleaner.add(this);
         logger.debug("sort complete for {} {}", shuffleKey, originFilePath);
       } catch (Exception e) {
@@ -708,7 +804,7 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
     }
 
     public void deleteOriginFiles() throws IOException {
-      boolean deleteSuccess = false;
+      boolean deleteSuccess;
       if (isHdfs) {
         deleteSuccess = StorageManager.hadoopFs().delete(new Path(originFilePath), false);
       } else {
@@ -749,10 +845,13 @@ public class PartitionFilesSorter extends ShuffleRecoverHelper {
 
     private void readBufferBySize(ByteBuffer buffer, int toRead) throws IOException {
       if (isHdfs) {
-        // HDFS don't need warmup
+        // HDFS does not need to prefetch.
         hdfsOriginInput.seek(toRead + hdfsOriginInput.getPos());
-      } else {
+      } else if (prefetchEnabled) {
+        buffer.clear();
         readChannelBySize(originFileChannel, buffer, originFilePath, toRead);
+      } else {
+        originFileChannel.position(toRead + originFileChannel.position());
       }
     }
   }
@@ -795,7 +894,7 @@ class PartitionFilesCleaner {
                       logger.debug(
                           "Deleting the original files for shuffle key {}: {}",
                           sorter.getShuffleKey(),
-                          sorter.getOriginFileInfo().getFilePath());
+                          ((DiskFileInfo) sorter.getOriginFileInfo()).getFilePath());
                       sorter.deleteOriginFiles();
                       it.remove();
                     }
@@ -808,7 +907,7 @@ class PartitionFilesCleaner {
               }
             }
           } catch (InterruptedException e) {
-            logger.warn("partition file cleaner thread interrupted while wait new sorter.", e);
+            logger.warn("Partition file cleaner thread interrupted while waiting new sorter.", e);
           }
         });
   }
