@@ -20,11 +20,12 @@ package org.apache.celeborn.service.deploy.worker
 import java.io.File
 import java.lang.{Long => JLong}
 import java.util
-import java.util.{HashMap => JHashMap, HashSet => JHashSet, Locale, Map => JMap}
+import java.util.{HashMap => JHashMap, HashSet => JHashSet, Locale, Map => JMap, UUID}
 import java.util.concurrent._
 import java.util.concurrent.atomic.{AtomicBoolean, AtomicIntegerArray}
 
 import scala.collection.JavaConverters._
+import scala.util.Random
 
 import com.google.common.annotations.VisibleForTesting
 import io.netty.util.HashedWheelTimer
@@ -38,12 +39,17 @@ import org.apache.celeborn.common.internal.Logging
 import org.apache.celeborn.common.meta.{DiskInfo, WorkerInfo, WorkerPartitionLocationInfo}
 import org.apache.celeborn.common.metrics.MetricsSystem
 import org.apache.celeborn.common.metrics.source.{JVMCPUSource, JVMSource, ResourceConsumptionSource, SystemMiscSource, ThreadPoolSource}
-import org.apache.celeborn.common.network.TransportContext
-import org.apache.celeborn.common.protocol.{PartitionType, PbRegisterWorkerResponse, PbWorkerLostResponse, RpcNameConstants, TransportModuleConstants}
+import org.apache.celeborn.common.network.{CelebornRackResolver, TransportContext}
+import org.apache.celeborn.common.network.sasl.SaslServerBootstrap
+import org.apache.celeborn.common.network.server.TransportServerBootstrap
+import org.apache.celeborn.common.network.util.TransportConf
+import org.apache.celeborn.common.protocol.{PartitionType, PbRegisterWorkerResponse, PbWorkerLostResponse, RpcNameConstants, TransportModuleConstants, WorkerEventType}
+import org.apache.celeborn.common.protocol.PbWorkerStatus.State
 import org.apache.celeborn.common.protocol.message.ControlMessages._
 import org.apache.celeborn.common.quota.ResourceConsumption
 import org.apache.celeborn.common.rpc._
-import org.apache.celeborn.common.util.{CelebornExitKind, CollectionUtils, JavaUtils, ShutdownHookManager, ThreadUtils, Utils}
+import org.apache.celeborn.common.rpc.{RpcSecurityContextBuilder, ServerSaslContextBuilder}
+import org.apache.celeborn.common.util.{CelebornExitKind, CollectionUtils, JavaUtils, ShutdownHookManager, SignalUtils, ThreadUtils, Utils}
 // Can Remove this if celeborn don't support scala211 in future
 import org.apache.celeborn.common.util.FunctionConverter._
 import org.apache.celeborn.server.common.{HttpService, Service}
@@ -52,6 +58,7 @@ import org.apache.celeborn.service.deploy.worker.congestcontrol.CongestionContro
 import org.apache.celeborn.service.deploy.worker.memory.{ChannelsLimiter, MemoryManager}
 import org.apache.celeborn.service.deploy.worker.memory.MemoryManager.ServingState
 import org.apache.celeborn.service.deploy.worker.monitor.JVMQuake
+import org.apache.celeborn.service.deploy.worker.profiler.JVMProfiler
 import org.apache.celeborn.service.deploy.worker.storage.{PartitionFilesSorter, StorageManager}
 
 private[celeborn] class Worker(
@@ -76,33 +83,78 @@ private[celeborn] class Worker(
   metricsSystem.registerSource(new JVMCPUSource(conf, MetricsSystem.ROLE_WORKER))
   metricsSystem.registerSource(new SystemMiscSource(conf, MetricsSystem.ROLE_WORKER))
 
-  val rpcEnv: RpcEnv = RpcEnv.create(
-    RpcNameConstants.WORKER_SYS,
-    workerArgs.host,
-    workerArgs.host,
-    workerArgs.port,
-    conf,
-    Math.min(64, Math.max(4, Runtime.getRuntime.availableProcessors())))
+  val workerStatusManager = new WorkerStatusManager(conf)
+  private val authEnabled = conf.authEnabled
+  private val secretRegistry = new WorkerSecretRegistryImpl(conf.workerApplicationRegistryCacheSize)
+
+  if (conf.logCelebornConfEnabled) {
+    logInfo(getConf)
+  }
+
+  val rpcEnv: RpcEnv =
+    if (!authEnabled) {
+      RpcEnv.create(
+        RpcNameConstants.WORKER_SYS,
+        TransportModuleConstants.RPC_SERVICE_MODULE,
+        workerArgs.host,
+        workerArgs.host,
+        workerArgs.port,
+        conf,
+        Math.min(64, Math.max(4, Runtime.getRuntime.availableProcessors())))
+    } else {
+      val externalSecurityContext = new RpcSecurityContextBuilder()
+        .withServerSaslContext(
+          new ServerSaslContextBuilder()
+            .withAddRegistrationBootstrap(false)
+            .withSecretRegistry(secretRegistry).build()).build()
+      logInfo(
+        s"Secure port enabled ${workerArgs.port} for secured RPC.")
+      RpcEnv.create(
+        RpcNameConstants.WORKER_SYS,
+        TransportModuleConstants.RPC_SERVICE_MODULE,
+        workerArgs.host,
+        workerArgs.host,
+        workerArgs.port,
+        conf,
+        Math.max(64, Runtime.getRuntime.availableProcessors()),
+        Some(externalSecurityContext))
+    }
+
+  private[worker] var internalRpcEnvInUse =
+    if (!conf.internalPortEnabled) {
+      rpcEnv
+    } else {
+      RpcEnv.create(
+        RpcNameConstants.WORKER_INTERNAL_SYS,
+        TransportModuleConstants.RPC_SERVICE_MODULE,
+        workerArgs.host,
+        workerArgs.host,
+        workerArgs.internalPort,
+        conf,
+        Math.min(64, Math.max(4, Runtime.getRuntime.availableProcessors())))
+    }
 
   private val host = rpcEnv.address.host
   private val rpcPort = rpcEnv.address.port
+  private val internalPort = internalRpcEnvInUse.address.port
   Utils.checkHost(host)
 
   private val WORKER_SHUTDOWN_PRIORITY = 100
   val shutdown = new AtomicBoolean(false)
   private val gracefulShutdown = conf.workerGracefulShutdown
-  private var exitKind = CelebornExitKind.EXIT_IMMEDIATELY
   if (gracefulShutdown) {
-    val checkPortMap = Map(
+    var checkPortMap = Map(
       WORKER_RPC_PORT -> conf.workerRpcPort,
       WORKER_FETCH_PORT -> conf.workerFetchPort,
       WORKER_PUSH_PORT -> conf.workerPushPort,
       WORKER_REPLICATE_PORT -> conf.workerReplicatePort)
+    if (conf.internalPortEnabled) {
+      checkPortMap += (WORKER_INTERNAL_PORT -> conf.workerInternalPort)
+    }
     assert(
       !checkPortMap.values.exists(_ == 0),
       "If enable graceful shutdown, the worker should use non-zero port. " +
         s"${checkPortMap.map { case (k, v) => k.key + "=" + v }.mkString(", ")}")
-    exitKind = CelebornExitKind.WORKER_GRACEFUL_SHUTDOWN
     try {
       val recoverRoot = new File(conf.workerGracefulShutdownRecoverPath)
       if (!recoverRoot.exists()) {
@@ -118,7 +170,7 @@ private[celeborn] class Worker(
 
   val storageManager = new StorageManager(conf, workerSource)
 
-  val memoryManager: MemoryManager = MemoryManager.initialize(conf)
+  val memoryManager: MemoryManager = MemoryManager.initialize(conf, storageManager)
   memoryManager.registerMemoryListener(storageManager)
 
   val partitionsSorter = new PartitionFilesSorter(memoryManager, conf, workerSource)
@@ -141,8 +193,18 @@ private[celeborn] class Worker(
   var controller = new Controller(rpcEnv, conf, metricsSystem, workerSource)
   rpcEnv.setupEndpoint(RpcNameConstants.WORKER_EP, controller)
 
+  // Visible for testing
+  private[worker] var internalRpcEndpoint: RpcEndpoint = _
+  private var internalRpcEndpointRef: RpcEndpointRef = _
+  if (conf.internalPortEnabled) {
+    internalRpcEndpoint = new InternalRpcEndpoint(internalRpcEnvInUse, conf, secretRegistry)
+    internalRpcEndpointRef = internalRpcEnvInUse.setupEndpoint(
+      RpcNameConstants.WORKER_INTERNAL_EP,
+      internalRpcEndpoint)
+  }
+
   val pushDataHandler = new PushDataHandler(workerSource)
-  private val pushServer = {
+  private val (pushServerTransportContext, pushServer) = {
     val closeIdleConnections = conf.workerCloseIdleConnections
     val numThreads = conf.workerPushIoThreads.getOrElse(storageManager.totalFlusherThread)
     val transportConf =
@@ -156,11 +218,13 @@ private[celeborn] class Worker(
         pushServerLimiter,
         conf.workerPushHeartbeatEnabled,
         workerSource)
-    transportContext.createServer(conf.workerPushPort)
+    (
+      transportContext,
+      transportContext.createServer(conf.workerPushPort, getServerBootstraps(transportConf)))
   }
 
   val replicateHandler = new PushDataHandler(workerSource)
-  val (replicateServer, replicateClientFactory) = {
+  val (replicateTransportContext, replicateServer, replicateClientFactory) = {
     val closeIdleConnections = conf.workerCloseIdleConnections
     val numThreads =
       conf.workerReplicateIoThreads.getOrElse(storageManager.totalFlusherThread)
@@ -176,12 +240,13 @@ private[celeborn] class Worker(
         false,
         workerSource)
     (
+      transportContext,
       transportContext.createServer(conf.workerReplicatePort),
       transportContext.createClientFactory())
   }
 
   var fetchHandler: FetchHandler = _
-  private val fetchServer = {
+  private val (fetchServerTransportContext, fetchServer) = {
     val closeIdleConnections = conf.workerCloseIdleConnections
     val numThreads = conf.workerFetchIoThreads.getOrElse(storageManager.totalFlusherThread)
     val transportConf =
@@ -194,7 +259,9 @@ private[celeborn] class Worker(
         closeIdleConnections,
         conf.workerFetchHeartbeatEnabled,
         workerSource)
-    transportContext.createServer(conf.workerFetchPort)
+    (
+      transportContext,
+      transportContext.createServer(conf.workerFetchPort, getServerBootstraps(transportConf)))
   }
 
   private val pushPort = pushServer.getPort
@@ -221,6 +288,7 @@ private[celeborn] class Worker(
       pushPort,
       fetchPort,
       replicatePort,
+      internalPort,
       diskInfos,
       JavaUtils.newConcurrentHashMap[UserIdentifier, ResourceConsumption])
 
@@ -237,7 +305,11 @@ private[celeborn] class Worker(
   val shuffleCommitInfos: ConcurrentHashMap[String, ConcurrentHashMap[Long, CommitInfo]] =
     JavaUtils.newConcurrentHashMap[String, ConcurrentHashMap[Long, CommitInfo]]()
 
-  private val masterClient = new MasterClient(rpcEnv, conf)
+  private val masterClient = new MasterClient(internalRpcEnvInUse, conf, true)
+  secretRegistry.initialize(masterClient)
+
+  private val rackResolver = new CelebornRackResolver(conf)
+  private val networkLocation = rackResolver.resolve(host).getNetworkLocation
 
   // (workerInfo -> last connect timeout timestamp)
   val unavailablePeers: ConcurrentHashMap[WorkerInfo, Long] =
@@ -259,7 +331,7 @@ private[celeborn] class Worker(
       conf.workerCleanThreads)
   val asyncReplyPool: ScheduledExecutorService =
     ThreadUtils.newDaemonSingleThreadScheduledExecutor("worker-rpc-async-replier")
-  val timer = new HashedWheelTimer(ThreadUtils.namedSingleThreadFactory("worker-timer"))
+  val timer = new HashedWheelTimer()
 
   // Configs
   private val heartbeatInterval = conf.workerHeartbeatTimeout / 4
@@ -269,6 +341,12 @@ private[celeborn] class Worker(
   var cleaner: ExecutorService =
     ThreadUtils.newDaemonSingleThreadExecutor("worker-expired-shuffle-cleaner")
 
+  private var jvmProfiler: JVMProfiler = _
+  if (conf.workerJvmProfilerEnabled) {
+    jvmProfiler = new JVMProfiler(conf)
+    jvmProfiler.start()
+  }
+
   private var jvmQuake: JVMQuake = _
   if (conf.workerJvmQuakeEnabled) {
     jvmQuake = JVMQuake.create(conf, workerInfo.toUniqueId().replace(":", "-"))
@@ -308,11 +386,24 @@ private[celeborn] class Worker(
   workerSource.addGauge(WorkerSource.READ_BUFFER_ALLOCATED_COUNT) { () =>
     memoryManager.getAllocatedReadBuffers
   }
+  workerSource.addGauge(WorkerSource.MEMORY_FILE_STORAGE_SIZE) { () =>
+    memoryManager.getMemoryFileStorageCounter
+  }
+  workerSource.addGauge(WorkerSource.DIRECT_MEMORY_USAGE_RATIO) { () =>
+    memoryManager.workerMemoryUsageRatio()
+  }
+  workerSource.addGauge(WorkerSource.EVICTED_FILE_COUNT) { () =>
+    storageManager.evictedFileCount.get()
+  }
+  workerSource.addGauge(WorkerSource.MEMORY_STORAGE_FILE_COUNT) { () =>
+    storageManager.memoryWriters.size()
+  }
+
   workerSource.addGauge(WorkerSource.ACTIVE_SHUFFLE_SIZE) { () =>
-    storageManager.getActiveShuffleSize()
+    storageManager.getActiveShuffleSize
   }
   workerSource.addGauge(WorkerSource.ACTIVE_SHUFFLE_FILE_COUNT) { () =>
-    storageManager.getActiveShuffleFileCount()
+    storageManager.getActiveShuffleFileCount
   }
   workerSource.addGauge(WorkerSource.PAUSE_PUSH_DATA_TIME) { () =>
     memoryManager.getPausePushDataTime
@@ -326,6 +417,9 @@ private[celeborn] class Worker(
   workerSource.addGauge(WorkerSource.PAUSE_PUSH_DATA_AND_REPLICATE_COUNT) { () =>
     memoryManager.getPausePushDataAndReplicateCounter
   }
+  workerSource.addGauge(WorkerSource.ACTIVE_SLOTS_COUNT) { () =>
+    workerInfo.usedSlots()
+  }
 
   private def highWorkload: Boolean = {
     (memoryManager.currentServingState, conf.workerActiveConnectionMax) match {
@@ -350,6 +444,7 @@ private[celeborn] class Worker(
       workerInfo.updateThenGetDiskInfos(storageManager.disksSnapshot().map { disk =>
         disk.mountPoint -> disk
       }.toMap.asJava).values().asScala.toSeq ++ storageManager.hdfsDiskInfo
+    workerStatusManager.checkIfNeedTransitionStatus()
     val response = masterClient.askSync[HeartbeatFromWorkerResponse](
       HeartbeatFromWorker(
         host,
@@ -361,10 +456,14 @@ private[celeborn] class Worker(
         handleResourceConsumption(),
         activeShuffleKeys,
         estimatedAppDiskUsage,
-        highWorkload),
+        highWorkload,
+        workerStatusManager.currentWorkerStatus),
       classOf[HeartbeatFromWorkerResponse])
     response.expiredShuffleKeys.asScala.foreach(shuffleKey => workerInfo.releaseSlots(shuffleKey))
     cleanTaskQueue.put(response.expiredShuffleKeys)
+
+    val workerEvent = response.workerEvent
+    workerStatusManager.doTransition(workerEvent)
     if (!response.registered) {
       logError("Worker not registered in master, clean expired shuffle data and register again.")
       try {
@@ -425,15 +524,22 @@ private[celeborn] class Worker(
     replicateHandler.init(this)
     fetchHandler.init(this)
     controller.init(this)
+    workerStatusManager.init(this)
 
     logInfo("Worker started.")
     rpcEnv.awaitTermination()
+    if (conf.internalPortEnabled) {
+      internalRpcEnvInUse.awaitTermination()
+    }
   }
 
   override def stop(exitKind: Int): Unit = {
     if (!stopped) {
       logInfo("Stopping Worker.")
 
+      if (jvmProfiler != null) {
+        jvmProfiler.stop()
+      }
       if (jvmQuake != null) {
         jvmQuake.stop()
       }
@@ -474,8 +580,13 @@ private[celeborn] class Worker(
       replicateServer.shutdown(exitKind)
       fetchServer.shutdown(exitKind)
       pushServer.shutdown(exitKind)
+      replicateTransportContext.close()
+      fetchServerTransportContext.close()
+      pushServerTransportContext.close()
       metricsSystem.stop()
-
+      if (conf.internalPortEnabled) {
+        internalRpcEnvInUse.stop(internalRpcEndpointRef)
+      }
       super.stop(exitKind)
 
       logInfo("Worker is stopped.")
@@ -497,6 +608,8 @@ private[celeborn] class Worker(
               pushPort,
               fetchPort,
               replicatePort,
+              internalPort,
+              networkLocation,
               // Use WorkerInfo's diskInfo since re-register when heartbeat return not-registered,
               // StorageManager have update the disk info.
               workerInfo.diskInfos.asScala.toMap,
@@ -583,6 +696,7 @@ private[celeborn] class Worker(
   @VisibleForTesting
   def cleanup(expiredShuffleKeys: JHashSet[String], threadPool: ThreadPoolExecutor): Unit =
     synchronized {
+      val expiredApplicationIds = new JHashSet[String]()
       expiredShuffleKeys.asScala.foreach { shuffleKey =>
         partitionLocationInfo.removeShuffle(shuffleKey)
         shufflePartitionType.remove(shuffleKey)
@@ -592,46 +706,27 @@ private[celeborn] class Worker(
         workerInfo.releaseSlots(shuffleKey)
         val applicationId = Utils.splitShuffleKey(shuffleKey)._1
         if (!workerInfo.getApplicationIdSet.contains(applicationId)) {
-          // When the running applications does not contain the application corresponding to expired shuffle key,
-          // resource consumption source should remove lose application gauges.
-          removeAppResourceConsumption(applicationId)
-          removeAppActiveConnection(applicationId)
+          expiredApplicationIds.add(applicationId)
+          secretRegistry.unregister(applicationId)
         }
         logInfo(s"Cleaned up expired shuffle $shuffleKey")
       }
       partitionsSorter.cleanup(expiredShuffleKeys)
       fetchHandler.cleanupExpiredShuffleKey(expiredShuffleKeys)
       threadPool.execute(new Runnable {
-        override def run(): Unit = storageManager.cleanupExpiredShuffleKey(expiredShuffleKeys)
+        override def run(): Unit = {
+          removeAppActiveConnection(expiredApplicationIds)
+          workerSource.sample(
+            WorkerSource.CLEAN_EXPIRED_SHUFFLE_KEYS_TIME,
+            s"cleanExpiredShuffleKeys-${UUID.randomUUID()}") {
+            storageManager.cleanupExpiredShuffleKey(expiredShuffleKeys)
+          }
+        }
       })
     }
 
-  private def removeAppResourceConsumption(applicationId: String): Unit = {
-    removeResourceConsumptionGauge(
-      ResourceConsumptionSource.DISK_FILE_COUNT,
-      applicationId)
-    removeResourceConsumptionGauge(
-      ResourceConsumptionSource.DISK_BYTES_WRITTEN,
-      applicationId)
-    removeResourceConsumptionGauge(
-      ResourceConsumptionSource.HDFS_FILE_COUNT,
-      applicationId)
-    removeResourceConsumptionGauge(
-      ResourceConsumptionSource.HDFS_BYTES_WRITTEN,
-      applicationId)
-  }
-
-  private def removeResourceConsumptionGauge(
-      resourceConsumptionName: String,
-      applicationId: String): Unit = {
-    resourceConsumptionSource.removeGauge(
-      resourceConsumptionName,
-      resourceConsumptionSource.applicationLabel,
-      applicationId)
-  }
-
-  private def removeAppActiveConnection(applicationId: String): Unit = {
-    workerSource.removeAppActiveConnection(applicationId)
+  private def removeAppActiveConnection(applicationIds: JHashSet[String]): Unit = {
+    workerSource.removeAppActiveConnection(applicationIds)
   }
 
   override def getWorkerInfo: String = {
@@ -641,13 +736,6 @@ private[celeborn] class Worker(
     sb.toString()
   }
 
-  override def getThreadDump: String = {
-    val sb = new StringBuilder
-    sb.append("========================= Worker ThreadDump ==========================\n")
-    sb.append(Utils.getThreadDump()).append("\n")
-    sb.toString()
-  }
-
   override def getApplicationList: String = {
     val sb = new StringBuilder
     sb.append("================= LifecycleManager Application List ======================\n")
@@ -708,23 +796,17 @@ private[celeborn] class Worker(
   override def exit(exitType: String): String = {
     exitType.toUpperCase(Locale.ROOT) match {
       case "DECOMMISSION" =>
-        exitKind = CelebornExitKind.WORKER_DECOMMISSION
         ShutdownHookManager.get().updateTimeout(
           conf.workerDecommissionForceExitTimeout,
           TimeUnit.MILLISECONDS)
+        workerStatusManager.doTransition(WorkerEventType.Decommission)
       case "GRACEFUL" =>
-        exitKind = CelebornExitKind.WORKER_GRACEFUL_SHUTDOWN
+        workerStatusManager.doTransition(WorkerEventType.Graceful)
       case "IMMEDIATELY" =>
-        exitKind = CelebornExitKind.EXIT_IMMEDIATELY
-      case _ => // Use origin code
+        workerStatusManager.doTransition(WorkerEventType.Immediately)
+      case _ =>
+        workerStatusManager.doTransition(workerStatusManager.exitEventType)
     }
-    // Use the original EXIT_CODE
-    new Thread() {
-      override def run(): Unit = {
-        Thread.sleep(10000)
-        System.exit(0)
-      }
-    }.start()
     val sb = new StringBuilder
     sb.append("============================ Exit Worker =============================\n")
     sb.append(s"Exit worker by $exitType triggered: \n")
@@ -736,6 +818,8 @@ private[celeborn] class Worker(
     // During shutdown, to avoid allocate slots in this worker,
     // add this worker to master's excluded list. When restart, register worker will
     // make master remove this worker from excluded list.
+    logInfo("Worker start to shutdown gracefully")
+    workerStatusManager.transitionState(State.InGraceFul)
     try {
       masterClient.askSync(
         ReportWorkerUnavailable(List(workerInfo).asJava),
@@ -765,9 +849,11 @@ private[celeborn] class Worker(
       logWarning(s"Waiting for all PartitionLocation release cost ${waitTime}ms, " +
         s"unreleased PartitionLocation: \n$partitionLocationInfo")
     }
+
+    workerStatusManager.transitionState(State.Exit)
   }
 
-  def decommissionWorker(): Unit = {
+  def sendWorkerUnavailableToMaster(): Unit = {
     try {
       masterClient.askSync(
         ReportWorkerUnavailable(List(workerInfo).asJava),
@@ -779,6 +865,12 @@ private[celeborn] class Worker(
             s"\n${storageManager.shuffleKeySet().asScala.mkString("[", ", ", "]")}",
           e)
     }
+  }
+
+  def decommissionWorker(): Unit = {
+    logInfo("Worker start to decommission")
+    workerStatusManager.transitionState(State.InDecommission)
+    sendWorkerUnavailableToMaster()
     shutdown.set(true)
     val interval = conf.workerDecommissionCheckInterval
     val timeout = conf.workerDecommissionForceExitTimeout
@@ -796,12 +888,15 @@ private[celeborn] class Worker(
       logWarning(s"Waiting for all shuffle expired cost ${waitTime}ms, " +
         s"unreleased shuffle: \n${storageManager.shuffleKeySet().asScala.mkString("[", ", ", "]")}")
     }
+    workerStatusManager.transitionState(State.Exit)
   }
 
   def exitImmediately(): Unit = {
     // During shutdown, to avoid allocate slots in this worker,
     // add this worker to master's excluded list. When restart, register worker will
     // make master remove this worker from excluded list.
+    logInfo("Worker start to exit immediately")
+    workerStatusManager.transitionState(State.InExit)
     try {
       masterClient.askSync[PbWorkerLostResponse](
         WorkerLost(
@@ -819,34 +914,49 @@ private[celeborn] class Worker(
           e)
     }
     shutdown.set(true)
+    workerStatusManager.transitionState(State.Exit)
   }
 
   ShutdownHookManager.get().addShutdownHook(
     new Thread(new Runnable {
       override def run(): Unit = {
         logInfo("Shutdown hook called.")
-        exitKind match {
-          case CelebornExitKind.WORKER_GRACEFUL_SHUTDOWN =>
-            logInfo("Worker start to shutdown gracefully")
+        workerStatusManager.exitEventType match {
+          case WorkerEventType.Graceful =>
             shutdownGracefully()
-          case CelebornExitKind.WORKER_DECOMMISSION =>
-            logInfo("Worker start to decommission")
+          case WorkerEventType.Decommission =>
             decommissionWorker()
           case _ =>
-            logInfo("Worker start to exit immediately")
             exitImmediately()
         }
-        stop(exitKind)
+
+        if (workerStatusManager.exitEventType == WorkerEventType.Graceful) {
+          stop(CelebornExitKind.WORKER_GRACEFUL_SHUTDOWN)
+        } else {
+          stop(CelebornExitKind.EXIT_IMMEDIATELY)
+        }
       }
     }),
     WORKER_SHUTDOWN_PRIORITY)
 
   @VisibleForTesting
   def getPushFetchServerPort: (Int, Int) = (pushPort, fetchPort)
+
+  def getServerBootstraps(transportConf: TransportConf)
+      : java.util.List[TransportServerBootstrap] = {
+    val serverBootstraps = new java.util.ArrayList[TransportServerBootstrap]()
+    if (authEnabled) {
+      serverBootstraps.add(new SaslServerBootstrap(
+        transportConf,
+        secretRegistry))
+    }
+    serverBootstraps
+  }
 }
 
 private[deploy] object Worker extends Logging {
   def main(args: Array[String]): Unit = {
+    SignalUtils.registerLogger(log)
     val conf = new CelebornConf
     val workerArgs = new WorkerArguments(args, conf)
     // There are many entries for setting the master address, and we should unify the entries as
