@@ -28,7 +28,7 @@ import org.apache.celeborn.common.CelebornConf
 import org.apache.celeborn.common.metrics.MetricsSystem
 import org.apache.celeborn.common.metrics.source.AbstractSource
 import org.apache.celeborn.common.network.client.TransportClient
-import org.apache.celeborn.common.util.{CollectionUtils, JavaUtils, Utils}
+import org.apache.celeborn.common.util.{JavaUtils, Utils}
 
 class WorkerSource(conf: CelebornConf) extends AbstractSource(conf, MetricsSystem.ROLE_WORKER) {
   override val sourceName = "worker"
@@ -49,6 +49,7 @@ class WorkerSource(conf: CelebornConf) extends AbstractSource(conf, MetricsSyste
   addCounter(REPLICATE_DATA_WRITE_FAIL_COUNT)
   addCounter(REPLICATE_DATA_CREATE_CONNECTION_FAIL_COUNT)
   addCounter(REPLICATE_DATA_CONNECTION_EXCEPTION_COUNT)
+  addCounter(REPLICATE_DATA_FAIL_NON_CRITICAL_CAUSE_COUNT)
   addCounter(REPLICATE_DATA_TIMEOUT_COUNT)
 
   addCounter(PUSH_DATA_HANDSHAKE_FAIL_COUNT)
@@ -77,6 +78,8 @@ class WorkerSource(conf: CelebornConf) extends AbstractSource(conf, MetricsSyste
   addTimer(TAKE_BUFFER_TIME)
   addTimer(SORT_TIME)
 
+  addTimer(CLEAN_EXPIRED_SHUFFLE_KEYS_TIME)
+
   def getCounterCount(metricsName: String): Long = {
     val metricNameWithLabel = metricNameWithCustomizedLabels(metricsName, Map.empty)
     namedCounters.get(metricNameWithLabel).counter.getCount
@@ -90,30 +93,27 @@ class WorkerSource(conf: CelebornConf) extends AbstractSource(conf, MetricsSyste
   }
 
   def connectionInactive(client: TransportClient): Unit = {
-    appActiveConnections.remove(client.getChannel.id().asLongText())
+    val applicationIds = appActiveConnections.remove(client.getChannel.id().asLongText())
     incCounter(ACTIVE_CONNECTION_COUNT, -1)
+    applicationIds.asScala.foreach(applicationId =>
+      incCounter(ACTIVE_CONNECTION_COUNT, -1, Map(applicationLabel -> applicationId)))
   }
 
   def recordAppActiveConnection(client: TransportClient, shuffleKey: String): Unit = {
     val applicationIds = appActiveConnections.get(client.getChannel.id().asLongText())
     val applicationId = Utils.splitShuffleKey(shuffleKey)._1
     if (applicationIds != null && !applicationIds.contains(applicationId)) {
+      addCounter(ACTIVE_CONNECTION_COUNT, Map(applicationLabel -> applicationId))
+      incCounter(ACTIVE_CONNECTION_COUNT, 1, Map(applicationLabel -> applicationId))
       applicationIds.add(applicationId)
-      addGauge(ACTIVE_CONNECTION_COUNT, Map(applicationLabel -> applicationId)) { () =>
-        appActiveConnections.asScala.count { case (_, applicationIds) =>
-          applicationIds.contains(applicationId)
-        }
-      }
     }
   }
 
-  def removeAppActiveConnection(applicationId: String): Unit = {
-    appActiveConnections.asScala.foreach { case (_, applicationIds) =>
-      if (applicationIds.contains(applicationId)) {
-        applicationIds.remove(applicationId)
-        removeGauge(ACTIVE_CONNECTION_COUNT, Map(applicationLabel -> applicationId))
-      }
-    }
+  def removeAppActiveConnection(applicationIds: util.Set[String]): Unit = {
+    applicationIds.asScala.foreach(applicationId =>
+      removeCounter(ACTIVE_CONNECTION_COUNT, Map(applicationLabel -> applicationId)))
+    appActiveConnections.values().asScala.foreach(connectionAppIds =>
+      applicationIds.asScala.foreach(applicationId => connectionAppIds.remove(applicationId)))
   }
 
   // start cleaner thread
@@ -128,7 +128,7 @@ object WorkerSource {
   // fetch data
   val OPEN_STREAM_TIME = "OpenStreamTime"
   val FETCH_CHUNK_TIME = "FetchChunkTime"
-  val CHUNK_STREAM_COUNT = "ChunkStreamCount"
+  val ACTIVE_CHUNK_STREAM_COUNT = "ActiveChunkStreamCount"
   val OPEN_STREAM_SUCCESS_COUNT = "OpenStreamSuccessCount"
   val OPEN_STREAM_FAIL_COUNT = "OpenStreamFailCount"
   val FETCH_CHUNK_SUCCESS_COUNT = "FetchChunkSuccessCount"
@@ -144,6 +144,7 @@ object WorkerSource {
   val REPLICATE_DATA_WRITE_FAIL_COUNT = "ReplicateDataWriteFailCount"
   val REPLICATE_DATA_CREATE_CONNECTION_FAIL_COUNT = "ReplicateDataCreateConnectionFailCount"
   val REPLICATE_DATA_CONNECTION_EXCEPTION_COUNT = "ReplicateDataConnectionExceptionCount"
+  val REPLICATE_DATA_FAIL_NON_CRITICAL_CAUSE_COUNT = "ReplicateDataFailNonCriticalCauseCount"
   val REPLICATE_DATA_TIMEOUT_COUNT = "ReplicateDataTimeoutCount"
   val PUSH_DATA_HANDSHAKE_FAIL_COUNT = "PushDataHandshakeFailCount"
   val REGION_START_FAIL_COUNT = "RegionStartFailCount"
@@ -168,6 +169,7 @@ object WorkerSource {
 
   // slots
   val SLOTS_ALLOCATED = "SlotsAllocated"
+  val ACTIVE_SLOTS_COUNT = "ActiveSlotsCount"
   val RESERVE_SLOTS_TIME = "ReserveSlotsTime"
 
   // connection
@@ -184,9 +186,14 @@ object WorkerSource {
   val BUFFER_STREAM_READ_BUFFER = "BufferStreamReadBuffer"
   val READ_BUFFER_DISPATCHER_REQUESTS_LENGTH = "ReadBufferDispatcherRequestsLength"
   val READ_BUFFER_ALLOCATED_COUNT = "ReadBufferAllocatedCount"
+  val MEMORY_FILE_STORAGE_SIZE = "MemoryFileStorageSize"
+  val DIRECT_MEMORY_USAGE_RATIO = "DirectMemoryUsageRatio"
+  val EVICTED_FILE_COUNT = "EvictedFileCount"
+
+  val MEMORY_STORAGE_FILE_COUNT = "MemoryStorageFileCount"
 
   // credit
-  val CREDIT_STREAM_COUNT = "CreditStreamCount"
+  val ACTIVE_CREDIT_STREAM_COUNT = "ActiveCreditStreamCount"
   val ACTIVE_MAP_PARTITION_COUNT = "ActiveMapPartitionCount"
 
   // local device
@@ -203,4 +210,11 @@ object WorkerSource {
   // active shuffle
   val ACTIVE_SHUFFLE_SIZE = "ActiveShuffleSize"
   val ACTIVE_SHUFFLE_FILE_COUNT = "ActiveShuffleFileCount"
+
+  // decommission
+  val IS_DECOMMISSIONING_WORKER = "IsDecommissioningWorker"
+
+  // clean
+  val CLEAN_TASK_QUEUE_SIZE = "CleanTaskQueueSize"
+  val CLEAN_EXPIRED_SHUFFLE_KEYS_TIME = "CleanExpiredShuffleKeysTime"
 }
