@@ -22,12 +22,13 @@ import java.nio.charset.StandardCharsets
 import java.nio.file.{FileAlreadyExistsException, Files, Paths}
 import java.util
 import java.util.concurrent.{ConcurrentHashMap, ScheduledExecutorService, ThreadPoolExecutor, TimeUnit}
-import java.util.concurrent.atomic.AtomicInteger
+import java.util.concurrent.atomic.{AtomicInteger, AtomicLong}
 import java.util.function.{BiConsumer, IntUnaryOperator}
 
 import scala.collection.JavaConverters._
 import scala.concurrent.duration._
 
+import com.google.common.annotations.VisibleForTesting
 import io.netty.buffer.PooledByteBufAllocator
 import org.apache.commons.io.FileUtils
 import org.apache.hadoop.fs.{FileSystem, Path}
@@ -37,22 +38,32 @@ import org.apache.celeborn.common.CelebornConf
 import org.apache.celeborn.common.exception.CelebornException
 import org.apache.celeborn.common.identity.UserIdentifier
 import org.apache.celeborn.common.internal.Logging
-import org.apache.celeborn.common.meta.{DeviceInfo, DiskInfo, DiskStatus, FileInfo, TimeWindow}
+import org.apache.celeborn.common.meta.{DeviceInfo, DiskFileInfo, DiskInfo, DiskStatus, FileInfo, MapFileMeta, MemoryFileInfo, ReduceFileMeta, TimeWindow}
 import org.apache.celeborn.common.metrics.source.{AbstractSource, ThreadPoolSource}
 import org.apache.celeborn.common.network.util.{NettyUtils, TransportConf}
 import org.apache.celeborn.common.protocol.{PartitionLocation, PartitionSplitMode, PartitionType, StorageInfo}
 import org.apache.celeborn.common.quota.ResourceConsumption
 import org.apache.celeborn.common.util.{CelebornExitKind, CelebornHadoopUtils, JavaUtils, PbSerDeUtils, ThreadUtils, Utils}
 import org.apache.celeborn.service.deploy.worker._
+import org.apache.celeborn.service.deploy.worker.memory.MemoryManager
 import org.apache.celeborn.service.deploy.worker.memory.MemoryManager.MemoryPressureListener
 import org.apache.celeborn.service.deploy.worker.shuffledb.{DB, DBBackend, DBProvider}
 import org.apache.celeborn.service.deploy.worker.storage.StorageManager.hadoopFs
 
 final private[worker] class StorageManager(conf: CelebornConf, workerSource: AbstractSource)
   extends ShuffleRecoverHelper with DeviceObserver with Logging with MemoryPressureListener {
+  // fileInfos and partitionDataWriters are one to one mapping
   // mount point -> file writer
   val workingDirWriters =
-    JavaUtils.newConcurrentHashMap[File, ConcurrentHashMap[String, FileWriter]]()
+    JavaUtils.newConcurrentHashMap[File, ConcurrentHashMap[String, PartitionDataWriter]]()
+  val hdfsWriters = JavaUtils.newConcurrentHashMap[String, PartitionDataWriter]()
+  val memoryWriters = JavaUtils.newConcurrentHashMap[MemoryFileInfo, PartitionDataWriter]()
+  // (shuffleKey->(filename->DiskFileInfo))
+  private val diskFileInfos =
+    JavaUtils.newConcurrentHashMap[String, ConcurrentHashMap[String, DiskFileInfo]]()
+  // (shuffleKey->(filename->MemoryFileInfo))
+  val memoryFileInfos =
+    JavaUtils.newConcurrentHashMap[String, ConcurrentHashMap[String, MemoryFileInfo]]()
 
   val hasHDFSStorage = conf.hasHDFSStorage
 
@@ -107,21 +118,22 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
   private val deviceMonitor =
     DeviceMonitor.createDeviceMonitor(conf, this, deviceInfos, tmpDiskInfos, workerSource)
 
-  private val byteBufAllocator: PooledByteBufAllocator =
+  val storageBufferAllocator: PooledByteBufAllocator =
     NettyUtils.getPooledByteBufAllocator(new TransportConf("StorageManager", conf), null, true)
+
   // (mountPoint -> LocalFlusher)
   private val (
     localFlushers: ConcurrentHashMap[String, LocalFlusher],
     _totalLocalFlusherThread: Int) = {
     val flushers = JavaUtils.newConcurrentHashMap[String, LocalFlusher]()
-    var totalThread = 0;
+    var totalThread = 0
     disksSnapshot().foreach { diskInfo =>
       if (!flushers.containsKey(diskInfo.mountPoint)) {
         val flusher = new LocalFlusher(
           workerSource,
           deviceMonitor,
           diskInfo.threadCount,
-          byteBufAllocator,
+          storageBufferAllocator,
           conf.workerPushMaxComponents,
           diskInfo.mountPoint,
           diskInfo.storageType,
@@ -137,10 +149,9 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
 
   val hdfsDir = conf.hdfsDir
   val hdfsPermission = new FsPermission("755")
-  val hdfsWriters = JavaUtils.newConcurrentHashMap[String, FileWriter]()
   val (hdfsFlusher, _totalHdfsFlusherThread) =
     if (hasHDFSStorage) {
-      logInfo(s"Initialize HDFS support with path ${hdfsDir}")
+      logInfo(s"Initialize HDFS support with path $hdfsDir")
       try {
         StorageManager.hadoopFs = CelebornHadoopUtils.getHadoopFS(conf)
       } catch {
@@ -152,7 +163,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
         Some(new HdfsFlusher(
           workerSource,
           conf.workerHdfsFlusherThreads,
-          byteBufAllocator,
+          storageBufferAllocator,
           conf.workerPushMaxComponents)),
         conf.workerHdfsFlusherThreads)
     } else {
@@ -160,6 +171,12 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
     }
 
   def totalFlusherThread: Int = _totalLocalFlusherThread + _totalHdfsFlusherThread
+  val activeTypes = conf.availableStorageTypes
+
+  def localOrHdfsStorageAvailable(): Boolean = {
+    StorageInfo.HDFSAvailable(activeTypes) || StorageInfo.localDiskAvailable(
+      activeTypes) || hdfsDir.nonEmpty || !diskInfos.isEmpty
+  }
 
   override def notifyError(mountPoint: String, diskStatus: DiskStatus): Unit = this.synchronized {
     if (diskStatus == DiskStatus.CRITICAL_ERROR) {
@@ -193,18 +210,16 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
   }
 
   // shuffleKey -> (fileName -> file info)
-  private val fileInfos =
-    JavaUtils.newConcurrentHashMap[String, ConcurrentHashMap[String, FileInfo]]()
   private val RECOVERY_FILE_NAME_PREFIX = "recovery"
   private var RECOVERY_FILE_NAME = "recovery.ldb"
-  private var db: DB = null
+  private var db: DB = _
   private var saveCommittedFileInfosExecutor: ScheduledExecutorService = _
   private val saveCommittedFileInfoBySyncMethod =
     conf.workerGracefulShutdownSaveCommittedFileInfoSync
   private val saveCommittedFileInfoInterval =
     conf.workerGracefulShutdownSaveCommittedFileInfoInterval
-  private val committedFileInfos: ConcurrentHashMap[String, ConcurrentHashMap[String, FileInfo]] =
-    JavaUtils.newConcurrentHashMap[String, ConcurrentHashMap[String, FileInfo]]()
+  private val committedFileInfos =
+    JavaUtils.newConcurrentHashMap[String, ConcurrentHashMap[String, DiskFileInfo]]()
   // ShuffleClient can fetch data from a restarted worker only
   // when the worker's fetching port is stable.
   val workerGracefulShutdown = conf.workerGracefulShutdown
@@ -265,7 +280,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
           try {
             val files = PbSerDeUtils.fromPbFileInfoMap(entry.getValue, cache)
             logDebug(s"Reload DB: $shuffleKey -> $files")
-            fileInfos.put(shuffleKey, files)
+            diskFileInfos.put(shuffleKey, files)
             committedFileInfos.put(shuffleKey, files)
             db.delete(entry.getKey)
           } catch {
@@ -294,27 +309,48 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
         logDebug(s"Update FileInfos into DB: $shuffleKey -> $files")
       } catch {
         case exception: Exception =>
-          logError(s"Update FileInfos into DB: ${shuffleKey} failed.", exception)
+          logError(s"Update FileInfos into DB: $shuffleKey failed.", exception)
       }
     }
   }
 
-  private def getNextIndex() = counter.getAndUpdate(counterOperator)
+  def saveAllMemoryStorageFileInfosToDB(): Unit = {
+    for (writer <- memoryWriters.asScala) {
+      Utils.tryLogNonFatalError(writer._2.evict(false))
+    }
+  }
+
+  private def getNextIndex = counter.getAndUpdate(counterOperator)
 
   private val newMapFunc =
-    new java.util.function.Function[String, ConcurrentHashMap[String, FileInfo]]() {
-      override def apply(key: String): ConcurrentHashMap[String, FileInfo] =
+    new java.util.function.Function[String, ConcurrentHashMap[String, DiskFileInfo]]() {
+      override def apply(key: String): ConcurrentHashMap[String, DiskFileInfo] =
+        JavaUtils.newConcurrentHashMap()
+    }
+
+  private val diskFileInfoMapFunc =
+    new java.util.function.Function[String, ConcurrentHashMap[String, DiskFileInfo]]() {
+      override def apply(key: String): ConcurrentHashMap[String, DiskFileInfo] =
+        JavaUtils.newConcurrentHashMap()
+    }
+
+  private val memoryFileInfoMapFunc =
+    new java.util.function.Function[String, ConcurrentHashMap[String, MemoryFileInfo]]() {
+      override def apply(key: String): ConcurrentHashMap[String, MemoryFileInfo] =
         JavaUtils.newConcurrentHashMap()
     }
 
   private val workingDirWriterListFunc =
-    new java.util.function.Function[File, ConcurrentHashMap[String, FileWriter]]() {
-      override def apply(t: File): ConcurrentHashMap[String, FileWriter] =
-        JavaUtils.newConcurrentHashMap[String, FileWriter]()
+    new java.util.function.Function[File, ConcurrentHashMap[String, PartitionDataWriter]]() {
+      override def apply(t: File): ConcurrentHashMap[String, PartitionDataWriter] =
+        JavaUtils.newConcurrentHashMap[String, PartitionDataWriter]()
     }
 
+  @VisibleForTesting
+  val evictedFileCount = new AtomicLong
+
   @throws[IOException]
-  def createWriter(
+  def createPartitionDataWriter(
       appId: String,
       shuffleId: Int,
       location: PartitionLocation,
@@ -322,8 +358,8 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       splitMode: PartitionSplitMode,
       partitionType: PartitionType,
       rangeReadFilter: Boolean,
-      userIdentifier: UserIdentifier): FileWriter = {
-    createWriter(
+      userIdentifier: UserIdentifier): PartitionDataWriter = {
+    createPartitionDataWriter(
       appId,
       shuffleId,
       location,
@@ -336,7 +372,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
   }
 
   @throws[IOException]
-  def createWriter(
+  def createPartitionDataWriter(
       appId: String,
       shuffleId: Int,
       location: PartitionLocation,
@@ -345,156 +381,97 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       partitionType: PartitionType,
       rangeReadFilter: Boolean,
       userIdentifier: UserIdentifier,
-      partitionSplitEnabled: Boolean): FileWriter = {
+      partitionSplitEnabled: Boolean): PartitionDataWriter = {
     if (healthyWorkingDirs().size <= 0 && !hasHDFSStorage) {
       throw new IOException("No available working dirs!")
     }
+    val partitionDataWriterContext = new PartitionDataWriterContext(
+      splitThreshold,
+      splitMode,
+      rangeReadFilter,
+      location,
+      appId,
+      shuffleId,
+      userIdentifier,
+      partitionType,
+      partitionSplitEnabled)
 
-    val fileName = location.getFileName
-    var retryCount = 0
-    var exception: IOException = null
-    val suggestedMountPoint = location.getStorageInfo.getMountPoint
-    while (retryCount < conf.workerCreateWriterMaxAttempts) {
-      val diskInfo = diskInfos.get(suggestedMountPoint)
-      val dirs =
-        if (diskInfo != null && diskInfo.status.equals(DiskStatus.HEALTHY)) {
-          diskInfo.dirs
-        } else {
-          logDebug(s"Disk unavailable for $suggestedMountPoint, return all healthy" +
-            s" working dirs. diskInfo $diskInfo")
-          healthyWorkingDirs()
-        }
-      if (dirs.isEmpty && hdfsFlusher.isEmpty) {
-        throw new IOException(s"No available disks! suggested mountPoint $suggestedMountPoint")
-      }
-      val shuffleKey = Utils.makeShuffleKey(appId, shuffleId)
-      if ((dirs.isEmpty && location.getStorageInfo.HDFSAvailable()) || location.getStorageInfo.HDFSOnly()) {
-        val shuffleDir =
-          new Path(new Path(hdfsDir, conf.workerWorkingDir), s"$appId/$shuffleId")
-        val fileInfo =
-          new FileInfo(
-            new Path(shuffleDir, fileName).toString,
-            userIdentifier,
-            partitionType,
-            partitionSplitEnabled)
-        fileInfos.computeIfAbsent(shuffleKey, newMapFunc).put(fileName, fileInfo)
-        FileSystem.mkdirs(StorageManager.hadoopFs, shuffleDir, hdfsPermission)
-        val hdfsWriter = partitionType match {
-          case PartitionType.MAP => new MapPartitionFileWriter(
-              fileInfo,
-              hdfsFlusher.get,
+    val writer =
+      try {
+        partitionType match {
+          case PartitionType.MAP => new MapPartitionDataWriter(
+              this,
               workerSource,
               conf,
               deviceMonitor,
-              splitThreshold,
-              splitMode,
-              rangeReadFilter)
-          case PartitionType.REDUCE => new ReducePartitionFileWriter(
-              fileInfo,
-              hdfsFlusher.get,
+              partitionDataWriterContext)
+          case PartitionType.REDUCE => new ReducePartitionDataWriter(
+              this,
               workerSource,
               conf,
               deviceMonitor,
-              splitThreshold,
-              splitMode,
-              rangeReadFilter)
+              partitionDataWriterContext)
           case _ => throw new UnsupportedOperationException(s"Not support $partitionType yet")
         }
-        if (workerGracefulShutdown) {
-          hdfsWriter.setStorageManager(this)
-          hdfsWriter.setShuffleKey(shuffleKey)
-        }
-        hdfsWriters.put(fileInfo.getFilePath, hdfsWriter)
-        return hdfsWriter
-      } else if (dirs.nonEmpty && location.getStorageInfo.localDiskAvailable()) {
-        val dir = dirs(getNextIndex() % dirs.size)
-        val mountPoint = DeviceInfo.getMountPoint(dir.getAbsolutePath, mountPoints)
-        val shuffleDir = new File(dir, s"$appId/$shuffleId")
-        val file = new File(shuffleDir, fileName)
-        try {
-          val fileInfo =
-            new FileInfo(
-              file.getAbsolutePath,
-              userIdentifier,
-              partitionType,
-              partitionSplitEnabled)
-          fileInfo.setMountPoint(mountPoint)
-          fileInfos.computeIfAbsent(shuffleKey, newMapFunc).put(fileName, fileInfo)
-          shuffleDir.mkdirs()
-          if (file.exists()) {
-            throw new FileAlreadyExistsException(
-              s"Shuffle data file ${file.getAbsolutePath} already exists.")
-          } else {
-            val createFileSuccess = file.createNewFile()
-            if (!createFileSuccess) {
-              throw new CelebornException(
-                s"Create shuffle data file ${file.getAbsolutePath} failed!")
-            }
-          }
-          val fileWriter = partitionType match {
-            case PartitionType.MAP => new MapPartitionFileWriter(
-                fileInfo,
-                localFlushers.get(mountPoint),
-                workerSource,
-                conf,
-                deviceMonitor,
-                splitThreshold,
-                splitMode,
-                rangeReadFilter)
-            case PartitionType.REDUCE => new ReducePartitionFileWriter(
-                fileInfo,
-                localFlushers.get(mountPoint),
-                workerSource,
-                conf,
-                deviceMonitor,
-                splitThreshold,
-                splitMode,
-                rangeReadFilter)
-            case _ => throw new UnsupportedOperationException(s"Not support $partitionType yet")
-          }
-          if (workerGracefulShutdown) {
-            fileWriter.setStorageManager(this)
-            fileWriter.setShuffleKey(shuffleKey)
-          }
-          deviceMonitor.registerFileWriter(fileWriter)
-          val map = workingDirWriters.computeIfAbsent(dir, workingDirWriterListFunc)
-          map.put(fileInfo.getFilePath, fileWriter)
-          location.getStorageInfo.setMountPoint(mountPoint)
-          logDebug(s"location $location set disk hint to ${location.getStorageInfo} ")
-          return fileWriter
-        } catch {
-          case fe: FileAlreadyExistsException =>
-            logError("Failed to create fileWriter because of existed file", fe)
-            throw fe
-          case t: Throwable =>
-            logError(
-              s"Create FileWriter for ${file.getAbsolutePath} of mount $mountPoint " +
-                s"failed, report to DeviceMonitor",
-              t)
-            exception = new IOException(t)
-            deviceMonitor.reportNonCriticalError(
-              mountPoint,
-              exception,
-              DiskStatus.READ_OR_WRITE_FAILURE)
-        }
-      } else {
-        exception = new IOException("No storage available for location:" + location.toString)
+      } catch {
+        case e: Exception =>
+          logError("Create partition data writer failed", e)
+          throw e
       }
-      retryCount += 1
-    }
-
-    throw exception
+    writer
   }
 
-  def getFileInfo(shuffleKey: String, fileName: String): FileInfo = {
-    val shuffleMap = fileInfos.get(shuffleKey)
-    if (shuffleMap ne null) {
-      shuffleMap.get(fileName)
-    } else {
-      null
+  def registerMemoryPartitionWriter(writer: PartitionDataWriter, fileInfo: MemoryFileInfo): Unit = {
+    memoryWriters.put(fileInfo, writer)
+  }
+
+  def unregisterMemoryPartitionWriterAndFileInfo(
+      fileInfo: MemoryFileInfo,
+      shuffleKey: String,
+      fileName: String): Unit = {
+    memoryWriters.remove(fileInfo)
+    val map = memoryFileInfos.get(shuffleKey)
+    if (map != null) {
+      map.remove(fileName)
     }
   }
 
+  def registerDiskFilePartitionWriter(
+      writer: PartitionDataWriter,
+      workingDir: File,
+      fileInfo: DiskFileInfo): Unit = {
+    if (writer.getDiskFileInfo.isHdfs) {
+      hdfsWriters.put(fileInfo.getFilePath, writer)
+      return
+    }
+    deviceMonitor.registerFileWriter(writer)
+    workingDirWriters.computeIfAbsent(workingDir, workingDirWriterListFunc).put(
+      fileInfo.getFilePath,
+      writer)
+  }
+
+  def getFileInfo(
+      shuffleKey: String,
+      fileName: String): FileInfo = {
+    val memoryShuffleMap = memoryFileInfos.get(shuffleKey)
+    if (memoryShuffleMap != null) {
+      val memoryFileInfo = memoryShuffleMap.get(fileName)
+      if (memoryFileInfo != null) {
+        return memoryFileInfo
+      }
+    }
+
+    val diskShuffleMap = diskFileInfos.get(shuffleKey)
+    if (diskShuffleMap != null) {
+      val diskFileInfo = diskShuffleMap.get(fileName)
+      if (diskFileInfo != null) {
+        return diskFileInfo
+      }
+    }
+
+    null
+  }
+
   def getFetchTimeMetric(file: File): TimeWindow = {
     if (diskInfos != null) {
       val diskInfo = diskInfos.get(DeviceInfo.getMountPoint(file.getAbsolutePath, diskInfos))
@@ -506,51 +483,57 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
 
   def shuffleKeySet(): util.HashSet[String] = {
     val hashSet = new util.HashSet[String]()
-    hashSet.addAll(fileInfos.keySet())
+    hashSet.addAll(diskFileInfos.keySet())
     hashSet
   }
 
   def topAppDiskUsage: util.Map[String, Long] = {
-    fileInfos.asScala.map { keyedWriters =>
+    diskFileInfos.asScala.map { keyedWriters =>
       {
         keyedWriters._1 -> keyedWriters._2.values().asScala.map(_.getFileLength).sum
       }
     }.toList.map { case (shuffleKey, usage) =>
-      Utils.splitShuffleKey(shuffleKey)._1 -> usage
+      shuffleKey.split("-")(0) -> usage
     }.groupBy(_._1).map { case (key, values) =>
       key -> values.map(_._2).sum
     }.toSeq.sortBy(_._2).reverse.take(conf.metricsAppTopDiskUsageCount * 2).toMap.asJava
   }
 
   def cleanFile(shuffleKey: String, fileName: String): Unit = {
-    val fileInfo = getFileInfo(shuffleKey, fileName)
-    if (fileInfo != null) {
-      cleanFileInternal(shuffleKey, fileInfo)
-    }
+    cleanFileInternal(shuffleKey, getFileInfo(shuffleKey, fileName))
   }
 
   def cleanFileInternal(shuffleKey: String, fileInfo: FileInfo): Boolean = {
+    if (fileInfo == null) return false
     var isHdfsExpired = false
-    if (fileInfo.isHdfs) {
-      isHdfsExpired = true
-      val hdfsFileWriter = hdfsWriters.get(fileInfo.getFilePath)
-      if (hdfsFileWriter != null) {
-        hdfsFileWriter.destroy(new IOException(
-          s"Destroy FileWriter $hdfsFileWriter caused by shuffle $shuffleKey expired."))
-        hdfsWriters.remove(fileInfo.getFilePath)
-      }
-    } else {
-      val workingDir =
-        fileInfo.getFile.getParentFile.getParentFile.getParentFile
-      val writers = workingDirWriters.get(workingDir)
-      if (writers != null) {
-        val fileWriter = writers.get(fileInfo.getFilePath)
-        if (fileWriter != null) {
-          fileWriter.destroy(new IOException(
-            s"Destroy FileWriter $fileWriter caused by shuffle $shuffleKey expired."))
-          writers.remove(fileInfo.getFilePath)
+    fileInfo match {
+      case info: DiskFileInfo =>
+        if (info.isHdfs) {
+          isHdfsExpired = true
+          val hdfsFileWriter = hdfsWriters.get(info.getFilePath)
+          if (hdfsFileWriter != null) {
+            hdfsFileWriter.destroy(new IOException(
+              s"Destroy FileWriter $hdfsFileWriter caused by shuffle $shuffleKey expired."))
+            hdfsWriters.remove(info.getFilePath)
+          }
+        } else {
+          val workingDir =
+            info.getFile.getParentFile.getParentFile.getParentFile
+          val writers = workingDirWriters.get(workingDir)
+          if (writers != null) {
+            val fileWriter = writers.get(info.getFilePath)
+            if (fileWriter != null) {
+              fileWriter.destroy(new IOException(
+                s"Destroy FileWriter $fileWriter caused by shuffle $shuffleKey expired."))
+              writers.remove(info.getFilePath)
+            }
+          }
         }
-      }
+      case mInfo: MemoryFileInfo =>
+        val memoryWriter = memoryWriters.remove(mInfo)
+        memoryWriter.destroy(new IOException(
+          s"Destroy FileWriter $memoryWriter caused by shuffle $shuffleKey expired."))
+      case _ =>
     }
 
     isHdfsExpired
@@ -561,8 +544,8 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
       cleanDB: Boolean = true): Unit = {
     expiredShuffleKeys.asScala.foreach { shuffleKey =>
       logInfo(s"Cleanup expired shuffle $shuffleKey.")
-      if (fileInfos.containsKey(shuffleKey)) {
-        val removedFileInfos = fileInfos.remove(shuffleKey)
+      if (diskFileInfos.containsKey(shuffleKey)) {
+        val removedFileInfos = diskFileInfos.remove(shuffleKey)
         var isHdfsExpired = false
         if (removedFileInfos != null) {
           removedFileInfos.asScala.foreach {
@@ -597,6 +580,13 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
           }
         }
       }
+      if (memoryFileInfos.containsKey(shuffleKey)) {
+        val memoryFileMaps = memoryFileInfos.remove(shuffleKey)
+        memoryFileMaps.asScala.foreach(u => {
+          cleanFileInternal(shuffleKey, u._2)
+          MemoryManager.instance().releaseMemoryFileStorage(u._2.releaseMemoryBuffers())
+        })
+      }
     }
   }
 
@@ -710,6 +700,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
     if (db != null) {
       if (exitKind == CelebornExitKind.WORKER_GRACEFUL_SHUTDOWN) {
         try {
+          saveAllMemoryStorageFileInfosToDB()
           saveAllCommittedFileInfosToDB()
           db.close()
         } catch {
@@ -743,10 +734,12 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
   }
 
   private def flushFileWriters(): Unit = {
-    workingDirWriters.forEach(new BiConsumer[File, ConcurrentHashMap[String, FileWriter]] {
-      override def accept(t: File, writers: ConcurrentHashMap[String, FileWriter]): Unit = {
-        writers.forEach(new BiConsumer[String, FileWriter] {
-          override def accept(file: String, writer: FileWriter): Unit = {
+    workingDirWriters.forEach(new BiConsumer[File, ConcurrentHashMap[String, PartitionDataWriter]] {
+      override def accept(
+          t: File,
+          writers: ConcurrentHashMap[String, PartitionDataWriter]): Unit = {
+        writers.forEach(new BiConsumer[String, PartitionDataWriter] {
+          override def accept(file: String, writer: PartitionDataWriter): Unit = {
             if (writer.getException == null) {
               try {
                 writer.flushOnMemoryPressure()
@@ -764,8 +757,8 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
         })
       }
     })
-    hdfsWriters.forEach(new BiConsumer[String, FileWriter] {
-      override def accept(t: String, u: FileWriter): Unit = {
+    hdfsWriters.forEach(new BiConsumer[String, PartitionDataWriter] {
+      override def accept(t: String, u: PartitionDataWriter): Unit = {
         u.flushOnMemoryPressure()
       }
     })
@@ -792,7 +785,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
         val writers = workingDirWriters.get(dir)
         if (writers != null) {
           writers.synchronized {
-            writers.values.asScala.map(_.getFileInfo.getFileLength).sum
+            writers.values.asScala.map(_.getDiskFileInfo.getFileLength).sum
           }
         } else {
           0
@@ -804,6 +797,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
         Math.min(diskInfo.configuredUsableSpace - totalUsage, fileSystemReportedUsableSpace)
       logDebug(s"updateDiskInfos  workingDirUsableSpace:$workingDirUsableSpace filemeta:$fileSystemReportedUsableSpace conf:${diskInfo.configuredUsableSpace} totalUsage:$totalUsage")
       diskInfo.setUsableSpace(workingDirUsableSpace)
+      diskInfo.setTotalSpace(totalUsage + workingDirUsableSpace)
       diskInfo.updateFlushTime()
       diskInfo.updateFetchTime()
     }
@@ -811,9 +805,9 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
   }
 
   def userResourceConsumptionSnapshot(): Map[UserIdentifier, ResourceConsumption] = {
-    fileInfos.synchronized {
+    diskFileInfos.synchronized {
       // shuffleId -> (fileName -> fileInfo)
-      fileInfos
+      diskFileInfos
         .asScala
         .toList
         .flatMap { case (shuffleKey, fileInfoMaps) =>
@@ -840,7 +834,7 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
   }
 
   def resourceConsumption(
-      fileInfos: List[FileInfo],
+      fileInfos: List[DiskFileInfo],
       subResourceConsumptions: util.Map[String, ResourceConsumption] = null)
       : ResourceConsumption = {
     val diskFileInfos = fileInfos.filter(!_.isHdfs)
@@ -856,16 +850,188 @@ final private[worker] class StorageManager(conf: CelebornConf, workerSource: Abs
   def notifyFileInfoCommitted(
       shuffleKey: String,
       fileName: String,
-      fileInfo: FileInfo): Unit = {
+      fileInfo: DiskFileInfo): Unit = {
     committedFileInfos.computeIfAbsent(shuffleKey, newMapFunc).put(fileName, fileInfo)
   }
 
-  def getActiveShuffleSize(): Long = {
-    fileInfos.values().asScala.map(_.values().asScala.map(_.getBytesFlushed).sum).sum
+  def getActiveShuffleSize: Long = {
+    diskFileInfos.values().asScala.map(_.values().asScala.map(_.getFileLength).sum).sum
   }
 
-  def getActiveShuffleFileCount(): Long = {
-    fileInfos.asScala.values.map(_.size()).sum
+  def getActiveShuffleFileCount: Long = {
+    diskFileInfos.asScala.values.map(_.size()).sum
+  }
+
+  def createFile(
+      partitionDataWriterContext: PartitionDataWriterContext,
+      useMemoryShuffle: Boolean): (MemoryFileInfo, Flusher, DiskFileInfo, File) = {
+    logDebug(
+      s"create file for ${partitionDataWriterContext.getShuffleKey} ${partitionDataWriterContext.getPartitionLocation.getFileName}")
+    val location = partitionDataWriterContext.getPartitionLocation
+    if (useMemoryShuffle
+      && location.getStorageInfo.memoryAvailable()
+      && MemoryManager.instance().memoryFileStorageAvailable()) {
+      logDebug(s"Create memory file for ${partitionDataWriterContext.getShuffleKey} ${partitionDataWriterContext.getPartitionLocation.getFileName}")
+      (
+        createMemoryFile(
+          partitionDataWriterContext.getAppId,
+          partitionDataWriterContext.getShuffleId,
+          location.getFileName,
+          partitionDataWriterContext.getUserIdentifier,
+          partitionDataWriterContext.getPartitionType,
+          partitionDataWriterContext.isPartitionSplitEnabled),
+        null,
+        null,
+        null)
+    } else if (location.getStorageInfo.localDiskAvailable() || location.getStorageInfo.HDFSAvailable()) {
+      logDebug(s"create non-memory file for ${partitionDataWriterContext.getShuffleKey} ${partitionDataWriterContext.getPartitionLocation.getFileName}")
+      val createDiskFileResult = createDiskFile(
+        location,
+        partitionDataWriterContext.getAppId,
+        partitionDataWriterContext.getShuffleId,
+        location.getFileName,
+        partitionDataWriterContext.getUserIdentifier,
+        partitionDataWriterContext.getPartitionType,
+        partitionDataWriterContext.isPartitionSplitEnabled)
+      (null, createDiskFileResult._1, createDiskFileResult._2, createDiskFileResult._3)
+    } else {
+      (null, null, null, null)
+    }
+  }
+
+  def createMemoryFile(
+      appId: String,
+      shuffleId: Int,
+      fileName: String,
+      userIdentifier: UserIdentifier,
+      partitionType: PartitionType,
+      partitionSplitEnabled: Boolean): MemoryFileInfo = {
+    val fileMeta = partitionType match {
+      case PartitionType.REDUCE =>
+        new ReduceFileMeta(conf.shuffleChunkSize)
+      case PartitionType.MAP =>
+        new MapFileMeta()
+      case PartitionType.MAPGROUP =>
+        throw new NotImplementedError("Map group is not implemented")
+    }
+    val shuffleKey = Utils.makeShuffleKey(appId, shuffleId)
+    val memoryFileInfo =
+      new MemoryFileInfo(
+        userIdentifier,
+        partitionSplitEnabled,
+        fileMeta)
+    logDebug(s"create memory file for ${shuffleKey} ${fileName} and put it int memoryFileInfos")
+    memoryFileInfos.computeIfAbsent(shuffleKey, memoryFileInfoMapFunc).put(
+      fileName,
+      memoryFileInfo)
+    memoryFileInfo
+  }
+
+  /**
+   * @return (Flusher,DiskFileInfo,workingDir)
+   */
+  def createDiskFile(
+      location: PartitionLocation,
+      appId: String,
+      shuffleId: Int,
+      fileName: String,
+      userIdentifier: UserIdentifier,
+      partitionType: PartitionType,
+      partitionSplitEnabled: Boolean): (Flusher, DiskFileInfo, File) = {
+    val suggestedMountPoint = location.getStorageInfo.getMountPoint
+    var retryCount = 0
+    var exception: IOException = null
+    val shuffleKey = Utils.makeShuffleKey(appId, shuffleId)
+    while (retryCount < conf.workerCreateWriterMaxAttempts) {
+      val diskInfo = diskInfos.get(suggestedMountPoint)
+      val dirs =
+        if (diskInfo != null && diskInfo.status.equals(DiskStatus.HEALTHY)) {
+          diskInfo.dirs
+        } else {
+          logDebug(s"Disk unavailable for $suggestedMountPoint, return all healthy" +
+            s" working dirs. diskInfo $diskInfo")
+          healthyWorkingDirs()
+        }
+      if (dirs.isEmpty && hdfsFlusher.isEmpty) {
+        throw new IOException(s"No available disks! suggested mountPoint $suggestedMountPoint")
+      }
+
+      if (dirs.isEmpty && location.getStorageInfo.HDFSAvailable()) {
+        val shuffleDir =
+          new Path(new Path(hdfsDir, conf.workerWorkingDir), s"$appId/$shuffleId")
+        FileSystem.mkdirs(StorageManager.hadoopFs, shuffleDir, hdfsPermission)
+        val hdfsFilePath = new Path(shuffleDir, fileName).toString
+        val hdfsFileInfo = new DiskFileInfo(
+          userIdentifier,
+          partitionSplitEnabled,
+          new ReduceFileMeta(conf.shuffleChunkSize),
+          hdfsFilePath,
+          StorageInfo.Type.HDFS)
+        diskFileInfos.computeIfAbsent(shuffleKey, diskFileInfoMapFunc).put(
+          fileName,
+          hdfsFileInfo)
+        return (hdfsFlusher.get, hdfsFileInfo, null)
+      } else if (dirs.nonEmpty && location.getStorageInfo.localDiskAvailable()) {
+        val dir = dirs(getNextIndex % dirs.size)
+        val mountPoint = DeviceInfo.getMountPoint(dir.getAbsolutePath, mountPoints)
+        val shuffleDir = new File(dir, s"$appId/$shuffleId")
+        shuffleDir.mkdirs()
+        val file = new File(shuffleDir, fileName)
+        try {
+          if (file.exists()) {
+            throw new FileAlreadyExistsException(
+              s"Shuffle data file ${file.getAbsolutePath} already exists.")
+          } else {
+            val createFileSuccess = file.createNewFile()
+            if (!createFileSuccess) {
+              throw new CelebornException(
+                s"Create shuffle data file ${file.getAbsolutePath} failed!")
+            }
+          }
+          val filePath = file.getAbsolutePath
+          val fileMeta = partitionType match {
+            case PartitionType.REDUCE =>
+              new ReduceFileMeta(conf.shuffleChunkSize)
+            case PartitionType.MAP =>
+              new MapFileMeta()
+            case PartitionType.MAPGROUP =>
+              throw new NotImplementedError("Map group is not implemented")
+          }
+          val diskFileInfo = new DiskFileInfo(
+            userIdentifier,
+            partitionSplitEnabled,
+            fileMeta,
+            filePath,
+            StorageInfo.Type.HDD)
+          logInfo(s"created file at $filePath")
+          diskFileInfos.computeIfAbsent(shuffleKey, diskFileInfoMapFunc).put(
+            fileName,
+            diskFileInfo)
+          return (
+            localFlushers.get(mountPoint),
+            diskFileInfo,
+            dir)
+        } catch {
+          case fe: FileAlreadyExistsException =>
+            logError("Failed to create fileWriter because of existed file", fe)
+            throw fe
+          case t: Throwable =>
+            logError(
+              s"Create FileWriter for ${file.getAbsolutePath} of mount $mountPoint " +
+                s"failed, report to DeviceMonitor",
+              t)
+            deviceMonitor.reportNonCriticalError(
+              mountPoint,
+              new IOException(t),
+              DiskStatus.READ_OR_WRITE_FAILURE)
+            throw t
+        }
+      } else {
+        exception = new IOException("No storage available for location:" + location.toString)
+      }
+      retryCount += 1
+    }
+    throw exception
   }
 }
 
